%Buy Example 3.16 Becker-Nanz TR->ESORICS'07
%Formalization using Situation Calculus (see Apt-Bezem 91 "Acyclic Programs")
%
%holds(P,S) :- property P holds in the situation S
%resulting from a sequence of events
%(latest events first)

holds(true,_S).
holds(and(A,B),S) :- holds(A,S), holds(B,S).
holds(not(A),S) :- \+ holds(A,S).
holds(bought,[buy|_]).
holds(played1,[play1|S]):-
	holds(bought,S), 
	holds(not(played1),S).
holds(played2,[play2|S]) :- 
	holds(played1,S), 
	holds(not(played2),S).
holds(F,[E|S]) :-		%"inertia" or "frame" axiom
	prop(F),
	event(E),
	\+ abnormal(F,E,S),
	holds(F,S).

%abnormal(P,E,S) :- event E in situation S causes P to stop
%being true in [E|S]

abnormal(played1,buy,_S).
abnormal(played2,play1,_Es):-
	holds(bought,S), 
	holds(not(played1),S).
abnormal(bought,play2,_Es) :- 
	holds(played1,S), 
	holds(not(played2),S).

event(buy).
event(play1).
event(play2).

prop(bought).
prop(played1).
prop(played2).


/* Query

| ?- holds(played2,Es).
Es = [play2,play1,buy|_A] ? 
yes
*/