
% ~~~ ARRAYS THEORY ~~~ %

in_interval(A,X,B) :- A<X, X<B.

empty([]).

m(Y,[X|Xs]) :- Y=X.
m(Y,[X|Xs]) :- m(Y,Xs).

sumlist([],Y) :- Y=0.0.
sumlist([X|L],Y) :- sumlist(L,Y-X).

haspositive([X|L]) :- X>0.
haspositive([X|L]) :- haspositive(L).

list([]).
list([X|Xs]) :- list(Xs).

ordered([]).
ordered([X|L]) :- ordered2(X,L).

ordered2(X,[]).
ordered2(X,[Y|L]) :- X=<Y, ordered2(Y,L).

leq_list([],[]).
leq_list([X|L],[Y|M]) :- X=<Y, leq_list(L,M).

elem(N,[X|Xs],Y) :- N=1.0, Y=X.
elem(N,[X|Xs],Y) :- N>=2.0, elem(N-1.0,Xs,Y).

zip([],[],[]).
zip([X|L],[Y|M],[Z|N]) :- Z=X+Y, zip(L,M,N).

prefix([],[]).
prefix([],[X|Y]).
prefix([X|Y],[W|Z]) :- X=W, prefix(Y,Z).

len([],A) :- A=0.0.
len([X|Xs],A) :- len(Xs,A-1.0).


% ~~~ PROPERTIES ~~~ %

% every list has an upper-bound value
prop01 :- a([L],e([M],a([Y],(m(Y,L) -> Y=<M)))).
% every list has a maximum
prop02 :- a([L],((\+ empty(L)) -> e([M],(m(M,L) and a([Y],(m(Y,L) -> Y=<M)))))).
prop02neg :- \+ a([L],((\+ empty(L)) -> e([M],(m(M,L) and a([Y],(m(Y,L) -> Y=<M)))))).
% if the sum of all the elements of a list is positive, then there is at least one positive element
prop03 :- a([L,Y],((sumlist(L,Y) and Y>0) -> haspositive(L))).
prop04 :- a([L,Y],((sumlist(L,Y) and Y>0) -> e([X], (m(X,L) and X>0)))).
% if two lists are ordered then the one-to-one-sum list is ordered
prop05 :- a([L,M,N],((ordered(L) and ordered(M) and zip(L,M,N)) -> ordered(N))).
% if two lists are smaller one-to-one on the elements, then the overall sum of the first is smaller than that of the second
prop06 :- a([L,M,X,Y],((leq_list(L,M) and sumlist(L,X) and sumlist(M,Y)) -> X=<Y)).
% given two distinct, ordered elements A,B of a list and any length, there exists a list of that length
%	whose elements are all cantained among A and B
prop07 :- a([L,A,B,N],((m(A,L) and m(B,L) and A=\=B and B>A) -> e([M],(len(M,N) and a([X], (m(X,M) -> (A<X and X<B))))))).
% there exists a list L such that all of its prefixes have the overall sum contained in the given list
prop08 :- e([L],a([S,N],((prefix(S,L) and sumlist(S,N)) -> e([X],(m(X,L) and X=N))))).
% given two lists one of whis is smaller than the other, then there exists an element of the
%	longer list whose position is greater of the length of the smaller list
prop09 :- a([A,N,M,L],((len(L,N) and N>=1 and len(A,M) and N=<M) -> e([E],(elem(N,A,E))))).
% existence of the reverse list
prop10 :- a([L],e([R],a([X,Y,N,M],((elem(X,L,N) and elem(Y,L,M) and N<M) -> e([P,Q],(elem(X,R,P) and elem(Y,L,Q) and P>Q)))))).
% existence of the multiplication
%	forall X,Y there exists a list L and a real Prod
%		whose length is Y and whose elements are all X and whose sumlist is Prod
% that is, for any two X,Y it is always defined the multiplication among these two
prop11 :- a([X,Y],e([L,Prod],(len(L,Y) and a([Z],((\+ e([W],(elem(W,L,Z)))) or elem(X,L,Z))) and sumlist(L,Prod)))).
% ~~~ TYPES ~~~ %

type(elem,3,[none,list,none]).
type(zip,3,[list,list,list]).
type(in_interval,3,[none,none,none]).
type(m,2,[none,list]).
type(ordered2,2,[none,list]).
type(len,2,[list,none]).
type(leq_list,2,[list,list]).
type(sublist,2,[list,list]).
type(sumlist,2,[list,none]).
type(ordered,1,[list]).
type(haspositive,1,[list]).
type(list,1,[list]).
type(empty,1,[list]).

