
% ------
% A program that (nondeterministically) compares the frontiers of two trees

frontier(tree(T1,T2),L) :- frontier(T1,L1), frontier(T2,L2), append(L1,L2,L).	% w(1)
frontier(tip(A),[A]).															% w(2)

eqtree(S,T) :- frontier(S,L), frontier(T,L).									% w(3)

% ------
% After Tupling Strategy

frontier1(TL,L) :- frontierlist(TL,LL), flatten(LL,L).							% w(4)

frontierlist([],[]).															% w(5)
frontierlist([T|TL],[L|LL]) :- frontier(T,L), frontierlist(TL,LL).				% w(6)

flatten([],[]).																	% w(7)
flatten([L|LL],N) :- flatten(LL,M), append(L,M,N).								% w(8)

eqtreelist(SL,TL) :- frontier1(SL,L), frontier1(TL,L).							% w(9)

% ------
% AUXILIARY
append([],L,L).% :- list(L).														% w(10)
append([H|T],L,[H|S]) :- append(T,L,S).											% w(11)

list([]).																		% w(12)
list([_|T]) :- list(T).															% w(13)


% ------
% LAWS
%
% semantics(law,variables,weight,constraint)
%
% NOTICE: 	- w(N) are weights for clauses in P
%		- u(N) are weights in the laws
%		- a(N) are auxiliary weights to be used only in law's constraints
%		  also, auxiliary weights CAN'T be shared across laws 


% LAW 1
new1(A,B,C) :- append(A,B,C).
new2(A,B,C) :- append(A,B,C), list(A).

% LAW 2
new3(A,B,C) :- append(A,B,C).
new4(A,B,C) :- append(A,B,C), list(C).

% LAW 3
new5(L1,L2,L3,L) :- append(L1,L2,T),append(T,L3,L).
new6(L1,L2,L3,L) :- append(L2,L3,M),append(L1,M,L).

% LAW 4
new7(S,L) :- frontier(S,L).
new8(S,L) :- frontier(S,L1),append(L1,[],L).

% LAW 5
new9(S,L) :- frontierlist([S],[L1]),flatten([L1],L).
new10(S,L) :- frontierlist([S],LL),flatten(LL,L).

% LAW 6
new11(L1,L) :- flatten([],[]),append(L1,[],L).
new12(L1,L) :- flatten([],M),append(L1,M,L).

% LAW 7
new13(T1,T2,TL,L) :- frontierlist([T1,T2|TL],[L1,L2|LL]),flatten([L1,L2|LL],L).
new14(T1,T2,TL,L) :- frontierlist([T1,T2|TL],M),flatten(M,L).

% LAW1
% liberal
% law(([frontier(S,L)],[frontier(S,L1),append(L1,[],L)]),(u(2)-u(1)),[(w(12)+w(10)+w(4)+u(2))>=u(1)]).
% strict
%%%% law(([frontier(S,L)],[frontier(S,L1),append(L1,[],L)]),(u(2)-u(1)),[(w(12)+w(10)+w(4)+u(2))>=u(1),w(1)>=1,(u(1)+w(2))>=1,(u(2)+w(12)+w(10)+w(11)+w(2))>=1]).
% OLD PROOF : u(1)>=(u(2)+w(11)+w(10)),(-u(1))>=(-w(10)-u(2))

% LAW2
% liberal
% law(([frontierlist([S],[L1]),flatten([L1],L)],[frontierlist([S],LL),flatten(LL,L)]),(u(4)-u(3)),[u(3)>=u(4)]).
% strict
%%%% law(([frontierlist([S],[L1]),flatten([L1],L)],[frontierlist([S],LL),flatten(LL,L)]),(u(4)-u(3)),[u(3)>=u(4),(u(3)+w(6)+w(5))>=1,(u(4)+w(6)+w(5))>=1]).

% LAW3 - associativity of append/3
% liberal
% law(([append(L1,L2,T),append(T,L3,L)],[append(L2,L3,M),append(L1,M,L)]),(u(6)-u(5)),[u(5)>=u(6)]).
% strict
%%%% law(([append(L1,L2,T),append(T,L3,L)],[append(L2,L3,M),append(L1,M,L)]),(u(6)-u(5)),[u(5)>=u(6),(u(1)+w(10))>=1,w(11)>=1,(u(2)+w(10))>=1]).

% LAW4 - fold the empty goal
%%%% law(([],[flatten([],[])]),(u(10)-u(9)),[u(9)>=(u(10)-w(5))]).

% LAW5 - generalize une term to a variable
% liberal
% law(([flatten([],[]),append(L1,[],L)],[flatten([],M),append(L1,M,L)]),(u(12)-u(11)),[u(11)>=u(12)]).
% strict
%%%% law(([flatten([],[]),append(L1,[],L)],[flatten([],M),append(L1,M,L)]),(u(12)-u(11)),[u(11)>=u(12),(u(11)+w(7))>=1,(u(12)+w(7))>=1]).

% LAW6 - fold the empty goal
%%%% law(([],[frontierlist([],[])]),(u(14)-u(13)),[u(13)>=u(14)]).

% LAW7 - fold the empty goal
% liberal
% law(([frontierlist([T1,T2|TL],[L1,L2|LL]),flatten([L1,L2|LL],L)],[frontierlist([T1,T2|TL],M),flatten(M,L)]),(u(16)-u(15)),[u(15)>=u(16)]).
% strict
%%%% law(([frontierlist([T1,T2|TL],[L1,L2|LL]),flatten([L1,L2|LL],L)],[frontierlist([T1,T2|TL],M),flatten(M,L)]),(u(16)-u(15)),[u(15)>=u(16),(u(15)+w(6)+w(6))>=1,(u(16)+w(6)+w(6))>=1]).

% OLD STUFF
%
% LAW4
% liberal
% law(([a(L1,L2,L3),a(L3,[],L)],[a(L1,[],L4),a(L2,[],L5),a(L4,L5,L)]),(u(8)-u(7)),[u(8)=:=(u(7)-w(10))]).
% strict
% law(([a(L1,L2,L3),a(L3,[],L)],[a(L1,[],L4),a(L2,[],L5),a(L4,L5,L)]),(u(8)-u(7)),[u(8)=:=(u(7)-w(10)),(u(1)+w(2))>=1,w(1)>=1,(u(2)+w(2)+w(11)+w(10))>=1,(w(1)-w(10)-u(2))>=1,w(11)>=1]).



% law(([append(A,B,C)],[append(A,B,C),list(A)]),w(15)-w(14),[rat(0,1)=rat(0,1),w(12)=w(14)-w(15)],symmetric).
% law(([append(A,B,C),append(C,D,E)],[append(B,D,F),append(A,F,E)]),w(19)-w(18),[w(11)>=rat(0,1),w(18)-w(19)>=rat(0,1)],normal).
law(([append(A,B,C),append(C,D,E)],[append(B,D,F),append(A,F,E)]),w(19)-w(18),[rat(0,1)=rat(0,1),w(18)=w(19)],symmetric).