
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                      %
% last modified - 27/09/2002           %
% by Valerio Senni                     %
% - mwgc_th.pl -                       %
%                                      %
% theory describing MWGC problem       %
% to be provided to MAP system for     %
% its solution                         %
%                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PROBLEM DESCRIPTION

% INITIAL/FINAL STATE AND SEQUENCE
initial([l,l,l,l]).
final([r,r,r,r]).

inEv([I|E]) :- initial(I).

finEv([F]) :- final(F).
finEv([X|E]) :- finEv(E).

% BAD STATES AND GOOD SEQUENCES
badst([M,W,G,C]) :- eq_side(W,G), dif_side(M,W).
badst([M,W,G,C]) :- eq_side(G,C), dif_side(M,G).

goodSeq([S]) :- \+ badst(S).
goodSeq([S|E]) :- \+ badst(S), goodSeq(E).

% TRANSITIONS AND TRANSITION SEQUENCES
 manLR([l,W,G,C],[r,W,G,C]).
 manRL([r,W,G,C],[l,W,G,C]).
 
wolfLR([l,l,G,C],[r,r,G,C]).
wolfRL([r,r,G,C],[l,l,G,C]).

goatLR([l,W,l,C],[r,W,r,C]).
goatRL([r,W,r,C],[l,W,l,C]).

 cabLR([l,W,G,l],[r,W,G,r]).
 cabRL([r,W,G,r],[l,W,G,l]).

t(S1,S2) :- manLR(S1,S2).
t(S1,S2) :- manRL(S1,S2).
t(S1,S2) :- wolfLR(S1,S2).
t(S1,S2) :- wolfRL(S1,S2).
t(S1,S2) :- goatLR(S1,S2).
t(S1,S2) :- goatRL(S1,S2).
t(S1,S2) :- cabLR(S1,S2).
t(S1,S2) :- cabRL(S1,S2).

trSeq([X,Y]) :- t(X,Y).
trSeq([X,Y|E]) :- t(X,Y), trSeq([Y|E]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% OPERATORS

leq(0,X).
leq(s(X),s(Y)) :- leq(X,Y).

less(0,s(X)).
less(s(X),s(Y)) :- less(X,Y).

eq_side(X,X).

dif_side(r,l).
dif_side(l,r).

length(0,[X]).
length(s(N),[X|L]) :- length(N,L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TYPES DEFINITIONS

side(l).
side(r).

actor(man).
actor(wolf).
actor(goat).
actor(cab).

nat(0).
nat(s(N)) :- nat(N).

state([M,W,G,C]) :- side(M),side(W),side(G),side(C).

stEvolution([]).
stEvolution([S|E]) :- state(S), stEvolution(E).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TYPES DECLARATION

% type(,,[]).

type(initial,1,[state]).
type(final,1,[state]).
type(inEv,1,[stEvolution]).
type(finEv,1,[stEvolution]).
type(badst,1,[state]).
type(t,2,[state,state]).
type(trSeq,1,[stEvolution]).
type(goodSeq,1,[stEvolution]).

type(manLR,2,[state,state]).
type(manRL,2,[state,state]).
type(wolfLR,2,[state,state]).
type(wolfRL,2,[state,state]).
type(goatLR,2,[state,state]).
type(goatRL,2,[state,state]).
type(cabLR,2,[state,state]).
type(cabRL,2,[state,state]).

% OPERATORS
type(less,2,[nat,nat]).
type(leq,2,[nat,nat]).
type(eq_side,2,[side,side]).
type(dif_side,2,[side,side]).
type(length,2,[nat,stEvolution]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
