%%%%%%%% input program: Progs/yamep.pl %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% maximum of a non-empty set
synth1 :-
synth(max_set(S,M), (m(M,S) and a([X],(m(X,S)->leq(X,M))))).
% 10ms

% minimum of a non-empty set
synth2 :-
synth(min_set(S,M),(m(M,S) and a([X],(m(X,S)->leq(M,X))))).
%20ms

% set union(S1,S2,S3) iff S3 = S1 U S2
synth3 :-
synth(set_union(S1,S2,S3),
(a([X],( m(X,S3)  <-> ( m(X,S1) or m(X,S2) ) )))).
%630ms

% singleton(S,X) iff  S = {X}
synth4 :-
synth(singleton(S,X),(m(X,S) and a([Y],(m(Y,S) -> eq_nat(X,Y))) )).
%10ms

% set_diff(S1,S2,S3) iff S3 = S1 - S2
synth5 :-
synth(set_diff(S1,S2,S3), a([X],( m(X,S3) <-> (m(X,S1) and \+ m(X,S2))))).
%470ms

%%% YAMEP specific

% create(W,U,W1,U1)
synth6 :-
synth(create(W,U,W1,U1), (eq_set(U,U1) and
e([Z], (set_union(W,U,Z) and ((emptyset(Z) and singleton(W1,0) ) or
( (\+ emptyset(Z)) and e([M,W2], ( (max_set(Z,M) and
singleton(W2,s(M))) and set_union(W,W2,W1))))))))).
%6200ms

% use(W,U,W1,U1)
synth7 :-
synth(use(W,U,W1,U1), e([X],(m(X,W) and e([WU], (set_union(W,U,WU) and
(min_set(WU,X) and e([SingX],(singleton(SingX,X) and (set_diff(W,SingX,W1)
and set_union(U,SingX,U1)) )) ))) ))).
%840ms

% release(W,U,W1,U1)
synth8:-
synth(release(W,U,W1,U1), (eq_set(W,W1) and e([X],(m(X,U) and
e([SingX],(singleton(SingX,X) and set_diff(U,SingX,U1) )))))).
% ok. (very fast)

% init(W,U)
synth9:-
synth(init(W,U),(emptyset(W) and emptyset(U))).
% ok. (very fast)

% unsafe(U)
synth10:-
synth(unsafe(U),e([X,Y], (m(X,U) and (m(Y,U) and (\+ eq_nat(X,Y)))))).
% ok. (very fast)


%%%% YAMEP derivation specific

% constraint of the new definition
% c(W,U)
synth11:-
synth(c(W,U),(e([X],(m(X,W) and e([WU], (set_union(W,U,WU) and
min_set(WU,X))))) and (\+ emptyset(U)) )).
% ok. (very fast)


% 1st iteration
% remove
prop1 :-
prove(\+ e([W,U],(init(W,U) and unsafe(U) ))).
% ok. (very fast)

% satisfiable
prop2 :-
prove(e([W,U,W1,U1],(create(W,U,W1,U1) and unsafe(U1) ))).
% not proved Fabio
% 2830ms

% folding possible
prop3 :-
prove(a([W,U,W1,U1],((create(W,U,W1,U1) and unsafe(U1)) -> unsafe(U)  ))).
% 7050ms

prop4 :-
prove(a([W,U,W1,U1],((release(W,U,W1,U1) and unsafe(U1)) -> unsafe(U)))).
% ok. (very fast)

% folding impossible
prop5 :-
prove( \+ a([W,U,W1,U1],((use(W,U,W1,U1) and unsafe(U1)) -> unsafe(U)
))).
% 1080ms

% satisfiable
prop6 :-
prove(e([W,U,W1,U1],(use(W,U,W1,U1) and unsafe(U1)))).
%390ms

% folding possible
prop7 :-
prove(a([W,U,W1,U1],((use(W,U,W1,U1) and unsafe(U1)) -> c(W,U)  ))).
% not proved (memory exhausted) Fabio
% true 3440ms

% 2nd iteration

% remove
prop8 :-
prove(\+ e([W,U],(init(W,U) and c(W,U) ))).
% ok. (very fast)

prop9 :-
prove(\+ e([W,U,W1,U1],(use(W,U,W1,U1) and c(W1,U1)))).
% ok. (very fast)

% satisfiable
prop10 :-
prove(e([W,U,W1,U1],(create(W,U,W1,U1) and c(W1,U1)))).
% not proved by Fabio
% true 4880ms

prop11 :-
prove(e([W,U,W1,U1],(release(W,U,W1,U1) and c(W1,U1)))).
% 430ms

% folding possible
prop12 :-
prove(a([W,U,W1,U1],( (create(W,U,W1,U1) and c(W1,U1)) -> c(W,U) ))).
% proved (takes about four minutes) Fabio
% 24760ms

prop12bis :-
prove(a([W,U],( e([W1,U1],(create(W,U,W1,U1) and c(W1,U1))) -> c(W,U) ))).
% 24870ms

prop13 :-
prove(a([W,U,W1,U1],( (release(W,U,W1,U1) and c(W1,U1)) -> unsafe(U) ))).
% proved (takes about two minutes) Fabio
% 550ms

