%Parameterized Semaphore - Pnueli 01

%holds(P,H) :- Property P holds after the sequence H of events
%latest events first

holds(true,_H).
holds(and(A,B),H) :- holds(A,H), holds(B,H).
holds(not(A),H) :- \+ holds(A,H).
holds(think(I),[]).
holds(free,[]).
holds(P,[E|Es]) :- elem(P), event(E), initiates(E,P), happens(E,Es).
holds(P,[E|Es]) :- elem(P), event(E), \+ terminates(E,P), happens(E,Es), holds(P,Es).


%happens(E,Es) :- E is an event that happens after the sequence Es of events

happens(req(I),Es) :- 
	holds(think(I),Es), 
	holds(free,Es).
happens(rel(I),Es) :- 
	holds(use(I),Es), 
	holds(busy,Es).


%elem(P) :- P is an elementary property

elem(use(I)).
elem(think(I)).
elem(free).
elem(busy).


%event(E) :- E is an event.

event(req(I)).
event(rel(I)).


% initiates(Event,Property) :- Property holds after Event

initiates(req(I),use(I)).
initiates(req(I),busy).
initiates(rel(I),think(I)).
initiates(rel(I),free).


% terminates(Event,Property) :- Property ceases to hold after Event

terminates(req(I),think(I)).
terminates(req(I),free).
terminates(rel(I),use(I)).
terminates(rel(I),busy).


unsafe :- holds(use(I),H), holds(use(J),H), \+ eq(I,J).

eq(I,I).

/* Queries

| ?- holds(played2,Es).
Es = [play2,play1,buy|_A] ? 
yes

| ?- holds(and(played1,not(bought)),[play2,play1,buy]).
yes
*/