% SIMPLIFY STRATEGY
% Delete subsumed clauses, delete clauses with false body,
% delete duplicate literals, % delete types for variable which do not occur
% in the body.
% The predicate simplify(Cs,Ds) is defined above


% simplify(Cs,Ds) :- simplify the clauses in the list Cs and derive
% the list Ds by:
% (i) Deleting subsumed clauses
% (ii) Deleting clauses with false body, and
% (iii) deleting duplicate literals.
% The predicate drop_useless_types which deletes type atoms for variables
% not occurring in the body will act as the identity when called
% from statement_into_clauses. However, it may not be the identity
% when called by apply_strat (see below).

% DYNAMICS
:- dynamic do_subs/0.
:- dynamic do_bcdet/0.
:- dynamic do_useless_elim/0.

% MODE: simplify_arrays(+Def,-NewDef)
% TYPE: simplify_arrays(def,def)
% simplify_arrays(A,A).
simplify_arrays((P,Clauses),(P,NewClauses)) :-
		presimplify(Clauses,SC1), !,
		
		(derivation_trace_on -> (write(' - Pre-Simplify :'),nl,show_def((P,SC1)),nl) ; true),
		
		% ( do_subs -> 
		simplify_subs(SC1,SC2), !,
		
		(derivation_trace_on -> (write(' - Simplify-Subsume :'),nl,show_def((P,SC2)),nl) ; true),
		
		% ; SC2 = SC1),
		( do_bcdet -> simplify_bcdet(SC2,NewClauses), ! ; NewClauses = SC2).

presimplify([],[]).
presimplify([(H,bd(C,Body))|Cs],Ss) :-
		false_body_arrays((H,bd(C,Body))),
		presimplify(Cs,Ss).
presimplify([(H,bd(C,B))|Cs],R) :-
		delete_duplicate_literals((H,bd(C,B)),(H,bd(C,B1))),
		drop_useless_types_constr((H,bd(C,B1)),(H,bd(C,B2))),
		linking_variables((H,B2),C,Vars),
		solve(clpr,C,Vars,NewC),
		% eliminate_ev(H,C,Constr),
		% simplify_c(H,Constr,C2),
		% eliminate_ev(H,C2,C1),

		% R = [(H,bd(C1,B1))|Ss],
		(do_useless_elim -> 
			(useless_and_true_preds(B2,B3) -> R = [(H,bd(NewC,B3))|Ss] ; R = Ss) ;
			R = [(H,bd(NewC,B2))|Ss]
		),
		
		presimplify(Cs,Ss).

false_body_arrays((H,bd(C,Body))) :-
	term_variables((H,Body),Vars),
	solve(clpr,C,Vars,[false]).
false_body_arrays((_,bd(_,Body))) :-
        member(L,Body),
        is_member((\+ L),Body).

% MODE: delete_duplicate_literals(+Cl,-NewCl)
% TYPE: delete_duplicate_literals(clause,clause)
% SEMANTICS: we are given a clause and we return the same
%   clause with no repeated atoms CONSIDERING EQUALITIES IN CONSTRAINTS
delete_duplicate_literals((H,bd(C,B)),(H,bd(C,B1))):-
		find_equalities(C,[]),
		delete_duplicates_noeq(bd([],B),bd(B1)).
delete_duplicate_literals((H,bd(C,B)),(H,bd(C,B2))):-
		find_equalities(C,EQs), EQs\=[],
		copy_term((B,EQs),(B1,EQs1)),
		unify_equalities(EQs1),
		delete_duplicates_eq(bd([],B),bd([],B1),bd(B2)).

% FUNCTIONAL !!
find_equalities([],[]).
find_equalities([C|Cs],[C|EQs]) :-
	is_pure_equality(C),
	find_equalities(Cs,EQs).
find_equalities([C|Cs],EQs) :-
	\+ is_pure_equality(C),
	find_equalities(Cs,EQs).

is_pure_equality(C) :- ((C = (A=B)) ; (C = (A=:=B))), var(A), var(B).

unify_equalities([]).
unify_equalities([C|Cs]) :-
	((C = (A=B)) ; (C = (A=:=B))), A=B,
	unify_equalities(Cs).

% MODE: delete_duplicates_noeq(+Body,-NewBody)
% TYPE: delete_duplicates_noeq(,)
% SEMANTICS: we are given a clause and we return the same
%   clause with no repeated atoms
delete_duplicates_noeq(bd(B,[]),bd(B)).
delete_duplicates_noeq(bd(B,[L|Ls]),bd(NewB)) :-
        is_member(L,B),
        !,
        delete_duplicates_noeq(bd(B,Ls),bd(NewB)).
delete_duplicates_noeq(bd(B,[L|Ls]),bd(NewB)) :-
        append(B,[L],BL),
        delete_duplicates_noeq(bd(BL,Ls),bd(NewB)).

% MODE: delete_duplicates_eq(+OkBody,+UnifiedBody,-NewBody)
% TYPE: delete_duplicates_eq(,)
delete_duplicates_eq(bd(B,[]),bd(_,[]),bd(B)).
delete_duplicates_eq(bd(B,[_|Ls]),bd(UB,[UL|ULs]),bd(NewB)) :-
        is_member(UL,UB),
        !,
        delete_duplicates_eq(bd(B,Ls),bd(UB,ULs),bd(NewB)).
delete_duplicates_eq(bd(B,[L|Ls]),bd(UB,[UL|ULs]),bd(NewB)) :-
        append(B,[L],BL),
	append(UB,[UL],UBL),
        delete_duplicates_eq(bd(BL,Ls),bd(UBL,ULs),bd(NewB)).


% MODE: useless_and_true_preds(+Body,-Body)
% TYPE: useless_and_true_preds(list(literal),list(literal)
% SEMANTICS: 
useless_and_true_preds([],[]).
useless_and_true_preds([L|B],NewB) :-
	subsumes(\+ A,L),
        functor(A,Pred,Arity),
	useless_list(UL),
	% true_list(TL),
	% (member(Pred/Arity,TL) -> (write('1 Clause eliminated -L'), nl, fail)
	%			;
	(member(Pred/Arity,UL) ->
		(NewB = NewBs %, write('1 -L'), nl)
		 ; NewB = [L|NewBs])
	),
	!,
	useless_and_true_preds(B,NewBs).
useless_and_true_preds([A|B],NewB) :-
        functor(A,Pred,Arity),
	useless_list(UL),
	% true_list(TL),
	(member(Pred/Arity,UL) ->
		 (fail) %write('1 Clause eliminated +L'), nl,
		; % (member(Pred/Arity,TL) -> (NewB = NewBs, write('1 +L'), nl) ; 
		 NewB = [A|NewBs]
		  % )
	),
	!,
	useless_and_true_preds(B,NewBs).

% MODE: simplify_subs(+Clauses,-SubsumedClauses
% TYPE: simplify_subs(list(clause),list(clause))
% SEMANTICS: 
simplify_subs([],[]).
simplify_subs([C1|Cs],Ds) :-
        del_elem(C2,Cs,Cs1),
		% (do_stats -> increment_num_try_subs ; true),
        % subsumes_constrained(C1,C2),
	subsumes_clause_chk_constr(C1,C2),
		% (do_stats -> increment_num_subs ; true),
		%% write('simplify_subs - eliminated by subsumption 1 on 2'),nl,
        %% write('The first clause subsumes the second :'), nl,
        %% show_all_constr([C1,C2]),nl,
        !,
        simplify_subs([C1|Cs1],Ds).
simplify_subs([C1|Cs],Ds) :-
        member(C2,Cs),
		% (do_stats -> increment_num_try_subs ; true),
        % subsumes_constrained(C2,C1),
	subsumes_clause_chk_constr(C2,C1),
		% (do_stats -> increment_num_subs ; true),
		%% write('simplify_subs - eliminated by subsumption 2 on 1'),nl,
        %% write('The first clause subsumes the second :'), nl,
        %% show_all_constr([C2,C1]),nl,
        !,
        simplify_subs(Cs,Ds).
simplify_subs([C|Cs],[C|Ds]) :-
        % drop_useless_types(C,D),
        simplify_subs(Cs,Ds).

% increment_num_subs :-
%	num_subs(N),
%	M is N + 1,
%	retractall(num_subs(_)),
%	assert(num_subs(M)).
% increment_num_try_subs :-
%	num_try_subs(N),
%	M is N + 1,
%	retractall(num_try_subs(_)),
%	assert(num_try_subs(M)).

% ==============================
% SUBSUMPTION OVER CONSTRAINTS

% C1:  H1 <- C1, B1.
% C2: (H1 <- C2, B1, B2.)T
%
% C1 subsumes C2 iff exists a substitution T such that C2
% has a form as shown above AND C2 -> C1.
% we translate C2 -> C1 into sat((\+ C2) OR C1)
%

subsumes_constrained(Cl1,Cl2) :-
        copy_term(Cl1,(H1,bd(C1,G1))),
        copy_term(Cl2,(H2,bd(C2,G2))),
        subsumes(H1,H2),
        subsumeB(G1,G2),
	entails(C2,C1),
        variant(Cl2,(H2,bd(C2,G2))). %!!

subsumeB([],_Us).
subsumeB([T|Ts],Us) :-
	member(U,Us),
	subsumes(T,U),
	subsumeB(Ts,Us).

% =========================================================
% new subsumption algorithm - Maurizio Proietti 24-05-2004
% subsumes_clause_chk_constr(Clause1,Clause2) :-
%       copy_term(Clause1,(H1,bd(_,B1))),
%	copy_term(Clause2,(_,bd(C2,_))),
%	ground_representation(Clause2,(H3,bd(_,B3))),
%       subsumes(H1,H3),
%       subsumes_body(B1,B3),
% 	entails(C2,Clause1).

subsumes_clause_chk_constr((H1,bd(C1,B1)),(H2,bd(C2,B2))) :-
        copy_term((H1,bd(C1,B1)),(Hc1,bd(_,Bc1))),
	% copy_term(Clause2,(_,bd(C2,_))),
	ground_representation((H2,bd(C2,B2)),(H3,bd(_,B3))),
        subsumes(Hc1,H3),
        subsumes_body(Bc1,B3),
	copy_term(((H1,C1),(H2,C2)),((H4,C4),(H5,C5))),
	unify_with_occurs_check(H4,H5),
	entails(C5,C4).

ground_representation(T,G) :-
	copy_term(T,G),
	term_variables(G,Vars),
	instantiate_vars(Vars,0).

instantiate_vars([],_).
instantiate_vars([X|Xs],N) :-
	X=var(N),
	N1 is N+1,
	instantiate_vars(Xs,N1).

subsumes_body([],_).
subsumes_body([L1|B1],B2) :-
        del_elem(L2,B2,B3),
        subsumes(L1,L2),
        subsumes_body(B1,B3).

% =========================================================

% MODE: drop_useless_types(+Clause,-NewClause)
% TYPE: drop_useless_types(clause,clause)
% SEMANTICS: 
drop_useless_types_constr((H,bd(C,[])),(H,bd(C,[]))).
drop_useless_types_constr((H,bd(C,[L|Ls])),(H,bd(C,Ms))) :-
	is_type_atom_arrays(L),
	\+ useful_type_atom(L,Ls),
	!,
	Ms=Ls1,
	drop_useless_types_constr((H,bd(C,Ls)),(H,bd(C,Ls1))).
drop_useless_types_constr((H,bd(C,[L|Ls])),(H,bd(C,Ms))) :-
        Ms=[L|Ls1],
        drop_useless_types_constr((H,bd(C,Ls)),(H,bd(C,Ls1))).

is_type_atom_arrays(A) :-
        functor(A,P,1),
        (P=list;P=nat).

