%:- module(basic_operations,[in/2,empty/2,add_clause/2,add_clauses/2,
%   remove_clause/2,remove_clauses/2,remove_all/1,select_clauses/3,
%   filter_clauses/4,set_to/2,show_all/1,prove_all/1,hd_pred/2,hd_args/2,
%   clause_hd/2,clause_bd/2,clause_bd_cns/2,clause_bd_goal/2,atom/2,
%   atom_pred/2,literal_pred/2,split_bd/2,is_a_constraint/1]).

:- op(950,xfx, [in]).

% Sets of clauses are represented as sets of facts where the
% predicate name is the set identifier.


% --- Operations on sets of clauses ---
(H,B) in Set :-
	G=..[Set,H,B],
	call(G).

empty(Set) :-
	G=..[Set,_,_],
	\+ G.

add_clause((H,B),Set) :-
	G=..[Set,H,B],
	assertz(G).

add_clauses(Cls,Set) :-
	prove_all((member(Cl,Cls),add_clause(Cl,Set))).

remove_clause((H,B),Set) :-
	G=..[Set,H,B],
	retract(G).

remove_clauses(Cls,Set) :-
	prove_all((member(Cl,Cls),remove_clause(Cl,Set))).

remove_all(Set) :-
	G=..[Set,_,_],
	retractall(G).

select_clauses(C in Set,G,Cls) :-
	findall(C, (C in Set, G), Cls).

filter_clauses(S1,_P,_Args,_S2) :- empty(S1), !.
filter_clauses(S1,P,Args,S2) :-
	remove_clause(C,S1),
	G=..[P,C,Args],
	(G -> (add_clause(C,S2), filter(S1,G,S2)) ; filter(S1,G,S2)).

set_to(S1,S2) :-
	remove_all(S1),
	prove_all((C in S2, add_clause(C,S1))).

show_all(Set) :-
	prove_all(((H,bd(Cns,Lits)) in Set, append(Cns,Lits,Bd), show_clause(0,(H,Bd)), nl)).

prove_all(G) :- findall(_,G,_).


% --- Operations on clauses,goals,atoms ---

hd_pred(Cl,P/N) :-
	Cl=(H,_),
	functor(H,P,N).

hd_args(Cl,Args) :-
	Cl=(H,_),
	H=..[_|Args].

clause_hd(Cl,H) :-
	Cl=(H,_).

clause_bd(Cl,B) :-
	Cl=(_,B).

clause_bd_cns(Cl,Cns) :-
	Cl=(_,bd(Cns,_)).

clause_bd_goal(Cl,G) :-
	Cl=(_,bd(_,G)).

atom(L,A) :-
	L=(\+ A1) -> A=A1 ; A=L.

atom_pred(A,P/N) :-
	functor(A,P,N).

literal_pred(L,P/N) :-
	atom(L,A),
	functor(A,P,N).

% Change body representation

split_bd([],bd([],[])).
split_bd([L|Ls],bd(Cns,Lits)) :-
	is_a_constraint(L) ->	
		(Cns=[L|Cns1], split_bd(Ls,bd(Cns1,Lits)));
		(Lits=[L|Lits1], split_bd(Ls,bd(Cns,Lits1))).

is_a_constraint(Cn) :-
	functor(Cn,F,_),
	(F=(=:=); F=(>=); F=(=<); F=(>); F=(<); F=(=\=)).

negative_literal(L) :-
	L=(\+ _).

positive_literal(L) :-
	\+ negative_literal(L).

cfact(Cl) :-
	clause_bd_goal(Cl,[]).

%is_a_literal(L) :-
%	L=..[sat|_].
%is_a_literal(L) :- 
%	L=..[\+,sat|_].
