%%%%%%%%%%%%%%%%%%%%%% CTL operators %%%%%%%%%%%%%%%%%%%%%%%

%:- module(ctl_operators, [ctl_clause/2]).

% First Order Logic

ctl_clause(prop(P),[\+ exists_state(and(initial,not(P)))]).
ctl_clause(exists_state(P),[sat(_X,P)]).


% Connectives

ctl_clause(sat(S,E),[elem(S,E)]).

ctl_clause(sat(X, or(F1,_F2)),[sat(X, F1)]).
ctl_clause(sat(X, or(_F1,F2)),[sat(X, F2)]).

ctl_clause(sat(X, and(F1, F2)),[sat(X, F1), sat(X, F2)]).

ctl_clause(sat(X, not(F)), [\+ sat(X, F)]).

ctl_clause(sat(X, imp(F1,_F2)), [\+ sat(X, F1)]).
ctl_clause(sat(X, imp(_F1,F2)), [sat(X, F2)]).


% CTL operators

ctl_clause(sat(S1, ex(F)), [ts(S1,Ss), mem(S2,Ss), sat(S2, F)]).

ctl_clause(sat(S, ef(F)), [sat(S,F)]).
ctl_clause(sat(S1, ef(F)), [t(S1,S2), sat(S2,ef(F))]).
%ctl_clause(sat(S1, ef(F)), [ts(S1,Ss), mem(S2,Ss), sat(S2,ef(F))]).

ctl_clause(sat(S, eu(_,F2)), [sat(S, F2)]).
ctl_clause(sat(S1, eu(F1,F2)),
	[sat(S1,F1), t(S1,S2), sat(S2,eu(F1,F2))]).

ctl_clause(sat(S, eg(F)), [sat(S,not(af(not(F))))]).

ctl_clause(sat(S, er(F1,F2)), [sat(S,not(au(not(F1),not(F2))))]).


ctl_clause(sat(S, ax(F)), [ts(S,Ss), sat_all(Ss, F)]).

ctl_clause(sat(S, af(F)), [sat(S,F)]).
ctl_clause(sat(S, af(F)), [ts(S,Ss), sat_all(Ss, af(F))]).
%ctl_clause(sat(S, af(F)), [\+ sat(S,not_af(F))]).
%ctl_clause(sat(S1,not_af(F)), [t(S1,S2), \+ sat(S2, af(F))]).
%ctl_clause(sat(S1,not_af(F)), [t(S1,S2), \+ sat(S2,F), sat(S2, notaf(F))]).

ctl_clause(sat(S, au(_,F2)), [sat(S, F2)]).
ctl_clause(sat(S1, au(F1,F2)),
	[sat(S1,F1), ts(S1,Ss), sat_all(Ss,au(F1,F2))]).

ctl_clause(sat(S, ag(F)), [sat(S,not(ef(not(F))))]).

ctl_clause(sat(S, ar(F1,F2)), [sat(S,not(eu(not(F1),not(F2))))]).


%%%%% Auxiliary predicates

ctl_clause(sat_all([], _), []).
ctl_clause(sat_all([S|Ss], F), [sat(S,F), sat_all(Ss, F)]).

ctl_clause(mem(X,[X|_Ys]), []).
ctl_clause(mem(X,[_Y|Ys]), [mem(X,Ys)]).



%%%%%%%%%%%%%% Witnesses %%%%%%%%%%%%

ctl_clause(witness(ef(P),W),[sat(X,initial),sat_w(X,ef(P),W)]).
ctl_clause(witness(eu(P1,P2),W),[sat(X,initial),sat_w(X,eu(P1,P2),W)]).
ctl_clause(witness(af(P),W),[sat(X,initial),sat_w(X,af(P),W)]).


% Connectives

ctl_clause(sat_w(S,E,[]),[elem(S,E)]).

ctl_clause(sat_w(X, or(F1,_F2), []),[sat(X, F1)]).
ctl_clause(sat_w(X, or(_F1,F2), []),[sat(X, F2)]).

ctl_clause(sat_w(X, and(F1, F2), []),[sat(X, F1), sat(X, F2)]).

ctl_clause(sat_w(X, not(F), []), [\+ sat(X, F)]).

ctl_clause(sat_w(X, imp(F1,_F2), []), [\+ sat(X, F1)]).
ctl_clause(sat_w(X, imp(_F1,F2), []), [sat(X, F2)]).


% CTL operators

ctl_clause(sat_w(S1, ex(F), []), [sat(S1, ex(F))]).

ctl_clause(sat_w(S, ef(F), []), [sat(S,F)]).
ctl_clause(sat_w(S1, ef(F), [T|W]), [t(S1,T,S2), sat_w(S2,ef(F),W)]).

ctl_clause(sat_w(S, eu(_,F2), []), [sat(S, F2)]).
ctl_clause(sat_w(S1, eu(F1,F2), [T|W]),
	[sat(S1,F1), t(S1,T,S2), sat_w(S2,eu(F1,F2),W)]).

ctl_clause(sat_w(S, eg(F), []), [sat(S,eg(F))]).

ctl_clause(sat_w(S, er(F1,F2), []), [sat(S, er(F1,F2))]).


ctl_clause(sat_w(S, ax(F), []), [sat(S, ax(F))]).

ctl_clause(sat_w(S, af(F), []), [sat(S,F)]).
ctl_clause(sat_w(S, af(F), W), [ts(S,Ss), sat_w_all(Ss, af(F), W)]).

ctl_clause(sat_w(S, au(_,F2), _W), [sat(S, au(_,F2))]).

ctl_clause(sat_w(S, ag(F), []), [sat(S, ag(F))]).

ctl_clause(sat_w(S, ar(F1,F2), []), [sat(S, ar(F1,F2))]).


%%%%% Auxiliary predicates

ctl_clause(sat_w_all([], _, []), []).
ctl_clause(sat_w_all([(T,S)|Ss], F, [(T,W)|Ws]), [sat_w(S,F,W), sat_w_all(Ss, F, Ws)]).