% definition rules.pl
%
% This file contains the predicates which specify the effects of applying the transformation
%rules associated with the predicates ts_define and gen_define which are defined in top_level.pl,
% after selection and checking of the input parameters.

:- use_module(library(terms), [ variant/2 ]).

/*---------
 ts_define(H,B): Introduce a new definition with head H and body B.
----------*/
apply_ts_define(Head,Body) :-
   current_prog(PN,CNs),
   ((label_arc(PN,(def,NewCN),PNSucc), NewCN=[N],
     clause3(N,(H,B),[def|_]), variant((Head,Body),(H,B))) ->
     (label_node(PNSucc,CNsSucc),
     update_current_prog((PN,CNs),(PNSucc,CNsSucc)))
     ; (add_clauses(NewCN,[(Head,Body)],[def|_]),
      (derivation_trace_on ->
        (nl, show_clauses(NewCN));true),
     update_programs((def,NewCN),PNSucc))).

/*---------
apply_gen_define(L): Introduce a new definition made out of clauses in L, where L is a list
of terms of the form (Head,Body) .
----------*/
apply_gen_define(L) :- apply_gen_define(L,_).
apply_gen_define(L2,NewCN) :- 
   add_clauses_gen_define(NewCN,L2),
   (derivation_trace_on -> (nl, show_clauses(NewCN));true),
   update_programs((gen_def,NewCN),_).

add_clauses_gen_define(L1,L2) :-
   max_clause(N), gen_X(L2,N,X), add_clauses(L1,L2, [gen_def|X]).
gen_X([],_,[]).
gen_X([_|L2],N,[N2|L3]):- N2 is N + 1, gen_X(L2,N2,L3).

