% determ_prog_util.pl
%
% This file contains utilitaries predicates used in determ_prog.pl

:- use_module(library(terms), [ variant/2 ]).

:-dynamic comment_history/2.

erase_environment2 :-
  erase_environment,
  retractall(max_depth(_)),
  retractall(comment_history(_,_)),
  retractall(current_depth(_)),
  retractall(initial_def(_,_,_)),
  retractall(init_def(_,_,_)),
  retractall(current_new_def_functor(_)),
  retractall(strat(_)),
  retractall(mode(_,_,_)),
  retractall(tr_session(_)),
  retractall(init_prog(_)).

assert_strat_parameters :- 
    strat_parameters(S), nl, write('Strategy parameters:'),nl,nl,
    assert_strat_parameters(S).
assert_strat_modes :- 
    needs_modes, !,
    strat_modes(S), nl, write('Strategy modes:'),nl,nl,
    assert_strat_modes(S).
assert_strat_modes.
/*-------------
trace_session/[0,1] 
no_trace_session/[0,1] 
      is an option which allow the user to interrupt temporarily
      the transformation process at some certain points in order
      to better follow what is going on and/or to insert
      some comments to be inserted in the history text.
      With no argument, the transformation will stop after the unfolding,
      case split and folding steps.
---------------*/
trace_session :-
  asserta(tr_session(unf)), asserta(tr_session(cs)), asserta(tr_session(fold)).

trace_session(Opt) :- asserta(tr_session(Opt)).
no_trace_session :- retractall(tr_session(_)).
no_trace_session(Opt) :- retractall(tr_session(Opt)).

show_strat_parameters :-
   write('Strategy parameters:'), nl, strat(S), write(S), nl, fail.
show_strat_parameters.

show_mode_predicates :-
    mode(_,_,_), !, write('Strategy modes:'), nl, show_mode_predicates2.
show_mode_predicates.

show_mode_predicates2 :-
    mode(F,N,Mode), write(mode(F,N,Mode)), nl, fail.
show_mode_predicates2.

assert_initial_def(NewCN) :-
    init_def(Pred,Arity,Initial_clauses),      
    (needs_modes -> assert_mode_initial_def(Initial_clauses) ; true),
    nl, write('Initial definition:'), nl,
    add_clauses_gen_define(NewCN,Initial_clauses),
    nl, show_clauses(NewCN),
    update_programs((gen_def,NewCN),_),
    assert(initial_def(Pred,Arity,NewCN)),
    assert(current_new_def_functor(1)).

retract_parameters_strat :-
      retractall(strat(_)),
      retractall(max_depth(_)),
      retractall(comment_history(_,_)),
      retractall(init_prog(_)),
      retractall(mode(_,_,_)),
      retractall(initial_def(_,_,_)),
      retractall(init_def(_,_,_)),
      retractall(tr_session(_)),
      retractall(current_new_def_functor(_)),
      retractall(current_depth(_)).

increment_depth :-
    retract(current_depth(X)), !, max_depth(Xmax),
    (X = Xmax -> msg_derivation_stopped, assert(max_depth(stop)) ; X2 is X + 1,
     assert(current_depth(X2))).
increment_depth :- assert(current_depth(1)).
reinitialize_depth :- retractall(current_depth(_)).

check_trace_session(Opt) :- tr_session(Opt),!, get_user_comment.
check_trace_session(_).

get_user_comment :-
    get_line(X), ( X = [] -> true ;
    (X = [112|_] -> pp, get_user_comment ;
    (X = [100|_] -> defs, get_user_comment ;
    (X = [111|_] -> no_trace_session ;
    (X = [99|_] -> insert_comment_history, get_user_comment ;
    (help_trace_session, get_user_comment)))))).

pred_in_initial_prog(Pred):- clause3(_,(Pred,_),prog).

non_instantiated_head(Hd):-
    functor(Hd,F,N), functor(Hd2,F,N), variant(Hd,Hd2).

definition_pred_used_for_folding([C|_],H) :- clause3(C,(H,_),_).

gen_new_name_functor(NewNumberDef) :-
   (current_new_def_functor(N) ->    retract(current_new_def_functor(N)) 
                                     ; N = 0),
   N2 is N + 1, assert(current_new_def_functor(N2)),
   name(N2,NameN2), name(NewNumberDef,[110,101,119|NameN2]).

exist_modes(P,N,X) :- clause(mode(P,N,X),true), !.

read_modes(P,N,X) :- clause(mode(P,N,X),true), !.
read_modes(P,N,X) :-
  functor(T,P,N), T =.. [_|X], msg_give_modes(P,N), give_modes(X,1), assert(mode(P,N,X)).

infer_modes_initial_def([V|L3],B1,['+'|Modes]):-
   occur_in_input(V,B1), !, infer_modes_initial_def(L3,B1,Modes).
infer_modes_initial_def([_|L3],B1,['-'|Modes]):-
   infer_modes_initial_def(L3,B1,Modes).
infer_modes_initial_def([],_,[]).

infer_modes([V|L3],B1,[_|Modes]):-
   occur_in_input(V,B1), !, infer_modes(L3,B1,Modes).
infer_modes([_|L3],B1,['-'|Modes]):- infer_modes(L3,B1,Modes).
infer_modes([],_,[]).

complete_input_args([]).
complete_input_args([+|L]) :- !, complete_input_args(L).
complete_input_args([_|L]) :- complete_input_args(L).

assert_modes_new_def([D|NewDefsN]) :- 
   infer_modes_new_def(D,X,N,ModeArgs),
   assert_modes_new_def(NewDefsN,X,N,ModeArgs).
assert_modes_new_def([],X,N,ModeArgs) :-
   complete_input_args(ModeArgs), assert(mode(X,N,ModeArgs)).
assert_modes_new_def([D|Defs],X,N,ModeArgs) :-
   infer_modes_new_def(D,X,N,ModeArgs),
   assert_modes_new_def(Defs,X,N,ModeArgs).

infer_modes_new_def(X,F,N,ModeArgs) :-
   clause_number_to_clause(X,(H,B)), predname_and_arity(H,F,N),
   H =..[_|L], infer_modes(L,B,ModeArgs).

anti_unify_list([T],T).
anti_unify_list([T1,T2|L], T) :-
   anti_unify(T1,T2,T3), anti_unify_list([T3|L],T).

all_input_modes(0,[]) :- !.
all_input_modes(N,['+'|Modes]):- N2 is N-1, all_input_modes(N2,Modes).
all_input_modes([]) :- !.
all_input_modes(['+'|Modes]):- all_input_modes(Modes).

assert_mode_initial_def([(Head, Body)|_]) :-
   Head=..[F|Args], infer_modes_initial_def(Args,Body,Modes),
   length(Args,N), assert(mode(F,N,Modes)).

occur_in_input(V,[B|_]):-
   functor(B,F,N), read_modes(F,N,Modes), B =.. [_|Args],
   occur_in_input(V,Args,Modes), !.
occur_in_input(V,[_|Bs]):- occur_in_input(V,Bs).
occur_in_input(V,[A|_],['+'|_]) :-
   free_variables(A,[],[],Vars), belongs(V,Vars).
occur_in_input(V,[_|Args],[_|Modes]) :-  occur_in_input(V,Args,Modes).

anti_unify_heads([N], CH, [(H,B)]):-
   (clause3(N, (H,B),_) ; N= (H,B)),
   copy_term((H,B),(CH,CB)), bind_freevars_in_head(CH,CB), !.

anti_unify_heads([N1,N2], H, [(H1,B1), (H2,B2)] ) :-
   (clause3(N1, (H1,B1),_) ; N1=(H1,B1)),
   (clause3(N2, (H2,B2),_) ; N2=(H2,B2)),
    copy_term((H1,B1),(CH1,CB1)), copy_term((H2,B2),(CH2,CB2)),
    bind_freevars_in_head(CH1,CB1), bind_freevars_in_head(CH2,CB2),
    anti_unify(H1,H2,H), !.

anti_unify_heads([N1,N2|L], H, [(H1,B1), (H2,B2)|L2] ) :-
   (clause3(N1, (H1,B1),_) ; N1=(H1,B1)),
   (clause3(N2, (H2,B2),_) ; N2=(H2,B2)),
    copy_term((H1,B1),(CH1,CB1)), copy_term((H2,B2),(CH2,CB2)),
    bind_freevars_in_head(CH1,CB1), bind_freevars_in_head(CH2,CB2),
    anti_unify(CH1,CH2,H3), anti_unify_heads(L, H4, L2), anti_unify(H3,H4,H).

%Corrected the 12/2/98
%bind_freevars_in_head(H,IneqsB) :-
%    body_without_inequalities(IneqsB,_,B), free_variables(B,[],[],VarsB),
%    free_variables(H,VarsB,[],FreeVarsH), bind_vars(FreeVarsH).

bind_freevars_in_head(H,IneqsB) :-
    body_without_inequalities(IneqsB,_,B), free_variables(B,[],[],VarsB),
    free_variables(H,VarsB,[],FreeVarsH), bind_vars(H,FreeVarsH).

bind_vars(_,[]).
bind_vars(H,[X|L]):- (occurrences_of_var(X,H,1) -> X=fv ; true), bind_vars(H,L).

bind_vars([]).
bind_vars([fv|L]):- bind_vars(L).
my_concat_atom(L,N) :- name_list(L,Names), append(Names,Name), name(N,Name).

name_list([],[]).
name_list([N|L],[Name|L2]) :- name(N,Name), name_list(L,L2).

del_4(X,1,[X|L],L).
del_4(X,N,[Y|L],[Y|L1]) :- N > 1, M is N-1, del_4(X,M,L,L1).
del_4(1,[_|L],L).
del_4(N,[Y|L],[Y|L1]) :- N > 1, M is N-1, del_4(M,L,L1).


needs_modes :- (strat(after_first_step_unfold_leftmost_consumer);
               strat(after_first_step_unfold_consumer);
               strat(apply_case_split(_));
	       strat(generalize_output_in_bodies_before_new_def(_));
	       strat(generalize_head_output_in_continuation)).



packets_of_one_clause([],[]).
packets_of_one_clause([C|Cs],[[C]|Ps]) :-
    packets_of_one_clause(Cs,Ps).

split_into_packets_of_mut_excl_heads([],[]).
split_into_packets_of_mut_excl_heads([F|Fs],[LF|L2]) :-
   split_into_packets_of_mut_excl_heads(Fs,F,[F],LF,[],LRest),
   split_into_packets_of_mut_excl_heads(LRest,L2).

split_into_packets_of_mut_excl_heads([],_,Irev,I,Lrev,L) :-
   reverse(Irev,I), reverse(Lrev,L).
split_into_packets_of_mut_excl_heads([C2|Fs],C,LFI,LF,LRestI,LRest) :-
   head_of_clause(C,H), head_of_clause(C2,H2), functor(H,F,N),
   (exist_modes(F,N,_) -> \+ unify_wrt_input_mode(H,H2) ; 
                          \+ unify_with_occurs_check(H,H2)),
   split_into_packets_of_mut_excl_heads(Fs,C,LFI,LF,[C2|LRestI],LRest).
split_into_packets_of_mut_excl_heads([F2|Fs],C,LFI,LF,LRestI,LRest) :-
   split_into_packets_of_mut_excl_heads(Fs,C,[F2|LFI],LF,LRestI,LRest).


vars_to_be_generalized(B,VarsG) :-
	bagof(A, is_an_output_argument(B,A), Args), !,
	mult_occurs_in_output_args_bd(Args,[],VarsG).
vars_to_be_generalized(_,[]).

mult_occurs_in_output_args_bd([],_,[]).
mult_occurs_in_output_args_bd([A|Args],ArgsI,[A|VarsG]) :-
	var(A), occurrences_of_var(A,(Args,ArgsI),N), N >= 1, !,
	mult_occurs_in_output_args_bd(Args,[A|ArgsI],VarsG).
mult_occurs_in_output_args_bd([A|Args],ArgsI,VarsG) :-
	mult_occurs_in_output_args_bd(Args,[A|ArgsI],VarsG).
