:- dynamic init_prog_max_coeff/1.

initialize_max_coeff(P) :-
	retractall(init_prog_max_coeff(_)),
	prog_max_coeff(P,M),
	assert(init_prog_max_coeff(M)).
%nl, write('max coeff '), write(M), nl.
	
prog_max_coeff(P,M) :-
	findall(X,
	(member(C,P),
	 clause_bd_cns(C,Cns),
	 max_coeff(Cns,X)),
	Xs),
	pos_list_max(Xs,M).

embedded_clause((_,B1),(_,B2)) :- embedded_cns_goal(B1,B2).

less_cns(C1,C2) :-
	max_coeff(C1,M1),
	max_coeff(C2,M2),
	init_prog_max_coeff(Maxcoeff),
	{M1<M2,
	 Maxcoeff<M2}.

embedded_cns_goal(CGoal1,CGoal2) :-
	CGoal1 = bd(C1,Goal1),
	CGoal2 = bd(C2,Goal2),
	variant(Goal1,Goal2),
%	Goal1=Goal2,
	less_cns(C1,C2).
%	embedded_cns(C1,C2).
%	homeo_embedded_goal(Goal1,Goal2).

embedded_cns(C1,C2) :-
	max_coeff_list(C1,0,M1),
	max_coeff_list(C2,0,M2),
	{M1=<M2}.

max_coeff_list([],M,M).
max_coeff_list([C|Cs],M,N) :-
	max_coeff(C,MC),
	max_num(M,MC,M1),
	max_coeff_list(Cs,M1,N).

max_coeff(C,M) :- var(C), !, {M=1}.
max_coeff(rat(X,Y),R) :-
	!,
	X1 is abs(X),
	Y1 is abs(Y),
	R=rat(X1,Y1).
max_coeff(C,M) :-
	number(C), !,
	M1 is abs(C),
	M=rat(M1,1).
max_coeff(C,MC) :-
	C=..[_|Args],
	max_coeff_list(Args,0,MC).



max_num(X,Y,Z) :- {X>=Y} -> Z=X ; Z=Y.

pos_list_max(L,M) :- plm(L,0,M).

plm([],M,M).
plm([X|Xs],A,M) :-
	max_num(X,A,B),
	plm(Xs,B,M).


%%% homeomorphic embedding (from F. homeo_embed.pl)

homeo_embedded_goal([],_).
homeo_embedded_goal([A|As],G) :-
	member(B,G),
	homeo_embedded_atom(A,B),
	homeo_embedded_goal(As,G).	

homeo_embedded_atom(A,B) :-
	A=..[P|AArgs],
	B=..[P|BArgs],
	homeo_embedded_term_list(AArgs,BArgs).

homeo_embedded_term(X,Y) :- 
	var(Y),!,
	var(X).

/*
homeo_embedded_term(M,N) :- 
	integer(N), !,
	integer(M), 
	X is abs(M), 
	Y is abs(N),
	X=<Y.

homeo_embedded_term(T,rat(N2,D2)) :-
	!,
	nonvar(T),
	T=rat(N1,D1),
	X1 is abs(N1), 
	X2 is abs(N2),
	X1=<X2,
	D1=<D2.

homeo_embedded_term(S,T) :-
	T =.. [_|Args],
	member(Ti,Args),
	homeo_embedded_term(S,Ti), !.


homeo_embedded_term(S,T) :-
	S =.. [F|SArgs],
	T =.. [F|TArgs],	
	homeo_embedded_term_list(SArgs,TArgs).

homeo_embedded_term_list([],[]).
homeo_embedded_term_list([ST|S],[TT|T]) :-
	homeo_embedded_term(ST,TT),!,
	homeo_embedded_term_list(S,T).

*/







