:- use_module(library(terms)). %variant
:- dynamic v_setting/1, vs/1, less_clause/2, generalize/3, clause_widening/3, init_prog_max_coeff/1.
:- retractall(less_clause(_,_)), 
   retractall(generalize(_,_,_)), 
   retractall(clause_widening(_,_,_)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

initialize_max_coeff(P) :-
	retractall(init_prog_max_coeff(_)),
	prog_max_coeff(P,M),
	assert(init_prog_max_coeff(M)).
%nl, write('max coeff '), write(M), nl.
	
prog_max_coeff(_P,M) :-
	vs(prog_maxcoeff(off)),
	M=0.
prog_max_coeff(P,M) :-
	vs(prog_maxcoeff(on)),
	findall(X,
	(member(C,P),
	 clause_bd_cns(C,Cns),
	 max_coeff(Cns,X)),
	Xs),
	pos_list_max(Xs,M).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

vs_msg :- 
	retractall(vs(_)),
	assert(vs(prog_maxcoeff(off))),
	assert(vs(less(maxcoeff))),
	assert(vs(gen(widen))),
	assert(vs(widen(e_leq_maxcoeff))),
	nl, write('current verification settings:'), nl,
	svs,
	nl, write('cvs to change settings.'), nl,
	write('svs to show current settings.'), nl, nl.

cvs :-  
	write('choose among the following settings:'),
	findall(A,clause(v_setting(A),_),Settings),
	nl,write(Settings),nl,
	read(S),
	v_setting(S).

svs :- 	
	findall(A,vs(A),Settings),
	nl,write(Settings),nl.
	
v_setting(prog_maxcoeff(on)) :-
	retractall(vs(prog_maxcoeff(_))),
	assert(vs(prog_maxcoeff(on))).
v_setting(prog_maxcoeff(off)) :-
	retractall(vs(prog_maxcoeff(_))),
	assert(vs(prog_maxcoeff(off))).
v_setting(less(variant)) :-
	retractall(less_clause(_,_)),
	assert((less_clause(C,D) :- less_variant(C,D))),
	retractall(vs(less(_))),
	assert(vs(less(variant))).
v_setting(less(maxcoeff)) :-
	retractall(less_clause(_,_)),
	assert((less_clause(C,D) :- less_maxcoeff(C,D))),
	retractall(vs(less(_))),
	assert(vs(less(maxcoeff))).
v_setting(less(maxsum)) :-
	retractall(less_clause(_,_)),
	assert((less_clause(C,D) :- less_maxsum(C,D))),
	retractall(vs(less(_))),
	assert(vs(less(maxsum))).

v_setting(gen(widen)) :- 
	retractall(generalize(_,_,_)),
	assert((generalize(D1,CD,Def) :- clause_widening(D1,CD,Def))),
	retractall(vs(gen(_))),
	assert(vs(gen(widen))).
v_setting(gen(ch_widen)) :- 
	retractall(generalize(_,_,_)),
	assert((generalize(D1,CD,Def) :-
		convex_hull_cl(D1,CD,CH),
		clause_widening(D1,CH,Def))),
	retractall(vs(gen(_))),
	assert(vs(gen(ch_widen))).

v_setting(widen(plain)) :- 
	retractall(clause_widening(_,_,_)),
	assert((clause_widening(D1,D2,D3) :- plain_widening(D1,D2,D3))),
	retractall(vs(widen(_))),
	assert(vs(widen(plain))).
v_setting(widen(e_leq_maxcoeff)) :- 
	retractall(clause_widening(_,_,_)),
	assert((clause_widening(D1,D2,D3) :- e_leq_maxcoeff(D1,D2,D3))),
	retractall(vs(widen(_))),
	assert(vs(widen(e_leq_maxcoeff))).
v_setting(widen(e_leq_maxsum)) :- 
	retractall(clause_widening(_,_,_)),
	assert((clause_widening(D1,D2,D3) :- e_leq_maxsum(D1,D2,D3))),
	retractall(vs(widen(_))),
	assert(vs(widen(e_leq_maxsum))).
v_setting(widen(leq_coeff)) :- 
	retractall(clause_widening(_,_,_)),
	assert((clause_widening(D1,D2,D3) :- leq_maxcoeff(D1,D2,D3))),
	retractall(vs(widen(_))),
	assert(vs(widen(leq_coeff))).
v_setting(widen(leq_sum)) :- 
	retractall(clause_widening(_,_,_)),
	assert((clause_widening(D1,D2,D3) :- leq_maxsum(D1,D2,D3))),
	retractall(vs(widen(_))),
	assert(vs(widen(leq_sum))).
v_setting(widen(e_ch_leq)) :- 
	retractall(clause_widening(_,_,_)),
	assert((clause_widening(D1,D2,D3) :- e_ch_leq(D1,D2,D3))),
	retractall(vs(widen(_))),
	assert(vs(widen(e_ch_leq))).


%%%%%%%%%%%%%%%%%%%%%%%

has_smaller_anc(CD,As,D1) :-
	similar_anc(CD,As,Ancs),
	Ancs=(_As1,[D1|_As2]), 
	less_clause(D1,CD).

less_clause(C,D) :- less_maxcoeff(C,D).

generalize(D1,CD,Def) :- clause_widening(D1,CD,Def).

clause_widening(D1,D2,D3) :- e_leq_maxcoeff(D1,D2,D3).

% MODE: similar_anc(+Cl,+Cls, -As)
% TYPE: similar_anc(clause,list(clause),(list(clause),list(clause)))
% SEMANTICS: As is of the form (As1,[D|As2]) where the bodies of Cl and D
% are variants.

similar_anc(C,[D|Ds],As) :-
	similar_cl(C,D),!,
	As=([],[D|Ds]).
similar_anc(C,[D|Ds],As) :-
	As=([D|As1],[D1|As2]),
	similar_anc(C,Ds,(As1,[D1|As2])).


% MODE: similar_cl(+Cl1,+Cl2)
% TYPE: similar_cl(clause,clause)
% SEMANTICS: the bodies of clauses Cl1 and Cl2 are variants

similar_cl(Cl1,Cl2) :-
	Cl1=(_,bd(_,[A])),
	Cl2=(_,bd(_,[B])),
	variant(A,B).


%%%%%%%% less_variant  %%%%%%%%%%

less_variant((_,B1),(_,B2)) :- less_variant_cns_goal(B1,B2).

less_variant_cns_goal(CGoal1,CGoal2) :-
	CGoal1 = bd(_C1,Goal1),
	CGoal2 = bd(_C2,Goal2),
	variant(Goal1,Goal2).


%%%%%%%% less_maxcoeff %%%%%%%%%%

less_maxcoeff((_,B1),(_,B2)) :- less_maxcoeff_cns_goal(B1,B2).

less_maxcoeff_cns_goal(CGoal1,CGoal2) :-
	CGoal1 = bd(C1,Goal1),
	CGoal2 = bd(C2,Goal2),
	variant(Goal1,Goal2),
	less_maxcoeff_cns(C1,C2).

less_maxcoeff_cns(C1,C2) :-
	max_coeff(C1,M1),
	max_coeff(C2,M2), 
	init_prog_max_coeff(Maxcoeff),
	{M1<M2,
	 Maxcoeff<M2}.

%leq_maxcoeff_cns(C1,C2) :-
%	max_coeff_list(C1,0,M1),
%	max_coeff_list(C2,0,M2),
%	{M1=<M2}.


max_coeff(C,M) :- var(C), !, {M=1}.
max_coeff(rat(X,Y),R) :-
	!,
	X1 is abs(X),
	Y1 is abs(Y),
	R=rat(X1,Y1).
max_coeff(C,M) :-
	number(C), !,
	M1 is abs(C),
	M=rat(M1,1).
max_coeff(C,MC) :-
	C=..[_|Args],
	max_coeff_list(Args,0,MC).

max_coeff_list([],M,M).
max_coeff_list([C|Cs],M,N) :-
	max_coeff(C,MC),
	max_num(M,MC,M1),
	max_coeff_list(Cs,M1,N).

max_num(X,Y,Z) :- {X>=Y} -> Z=X ; Z=Y.

pos_list_max(L,M) :- plm(L,0,M).

plm([],M,M).
plm([X|Xs],A,M) :-
	max_num(X,A,B),
	plm(Xs,B,M).


%%%%%%%%%%%%%%%%% less_maxsum

less_maxsum((_,B1),(_,B2)) :- less_maxsum_cns_goal(B1,B2).

less_maxsum_cns_goal(CGoal1,CGoal2) :-
	CGoal1 = bd(C1,Goal1),
	CGoal2 = bd(C2,Goal2),
	variant(Goal1,Goal2),
	less_maxsum_cns(C1,C2).

less_maxsum_cns(C1,C2) :-
	max_coeff_sums(C1,0,M1),
	max_coeff_sums(C2,0,M2),
	init_prog_max_coeff(Maxcoeff),
	{M1<M2,
	 Maxcoeff<M2}.

max_coeff_sums([],M,M).
max_coeff_sums([C|Cs],M,N) :-
	coeff_sum(C,SC),
	max_num(M,SC,M1),
	max_coeff_sums(Cs,M1,N).

coeff_sum(C,S) :- var(C), !, {S=0}.
coeff_sum(rat(X,Y),S) :-
	!,
	X1 is abs(X),
	Y1 is abs(Y),
	S=rat(X1,Y1).
coeff_sum(C,S) :-
	number(C), !,
	S1 is abs(C),
	S=rat(S1,1).
coeff_sum(C,S) :-
	C=..[_|Args],
	coeff_sum_list(Args,0,S).

coeff_sum_list([],S,S).
coeff_sum_list([C|Cs],A,S) :-
	coeff_sum(C,SC),
	{A1=SC+A},
	coeff_sum_list(Cs,A1,S).


%%%%%%%%% plain_widening

plain_widening(D1,D2,D3) :-
	copy_term(D1,(_H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
%	term_variables(H1,V1),
%	term_variables(H2,V2),
%	V1=V2,
	B1=B2,
	primitives(C1,I1),
	plain_cns_widening(I1,C2,C3),
	D3=(H2,bd(C3,B2)).

plain_cns_widening(C1,C2,C3) :-
	keep_entailed(C1,C2,E3),
	term_variables(E3,V3),
	solve(E3,V3,C3).


%%%%%%%%% e_leq_maxcoeff

e_leq_maxcoeff(D1,D2,D3) :-
	copy_term(D1,(_H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
%	term_variables(H1,V1),
%	term_variables(H2,V2),
%	V1=V2,
	B1=B2,
	primitives(C1,I1),
	leq_maxcoeff_cns(I1,C2,C3),
	D3=(H2,bd(C3,B2)).

leq_maxcoeff_cns(C1,C2,C3) :-
	keep_entailed(C1,C2,E1),
	keep_leq_maxcoeff_cns(C2,C1,E2),
	append(E1,E2,E3),
	term_variables(E3,V3),
	solve(E3,V3,C3).

%%% keep_entailed(C1,C2,C3)
%%% C3 is the subset of the constraints in C1 entailed by C2

keep_entailed([],_C2,[]).
keep_entailed(C1,C2,E) :-
	C1=[PC1|PC1s],
	(entails(C2,[PC1]) ->
	    E=[PC1|E1],
	    keep_entailed(PC1s,C2,E1)
	;
	    keep_entailed(PC1s,C2,E)
	).

%%% keep_leq_maxcoeff_cns(C1,C2,C3)
%%% C3 is the subset of constraints in C1 which are leq than some constraint in C2

keep_leq_maxcoeff_cns([],_C2,[]).
keep_leq_maxcoeff_cns(C1,C2,E) :-
	C1=[PC1|PC1s],
	((member(PC2,C2), leq_maxcoeff_cns(PC1,PC2)) ->
	    (E=[PC1|E1],
	    keep_leq_maxcoeff_cns(PC1s,C2,E1))
	;
	    keep_leq_maxcoeff_cns(PC1s,C2,E)
	).

leq_maxcoeff_cns(C1,C2) :-
	max_coeff(C1,M1),
	max_coeff(C2,M2),
	init_prog_max_coeff(Maxcoeff),
	{M1=<max(M2,Maxcoeff)}.


%%%%%%%%% e_leq_maxsum

e_leq_maxsum(D1,D2,D3) :-
	copy_term(D1,(H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(H1,V1),
	term_variables(H2,V2),
	V1=V2,
	B1=B2,
	primitives(C1,I1),
	leq_maxsum_cns(I1,C2,C3),
	D3=(H2,bd(C3,B2)).

leq_maxsum_cns(C1,C2,C3) :-
	keep_entailed(C1,C2,E1),
	keep_leq_maxsum_cns(C2,C1,E2),
	append(E1,E2,E3),
	term_variables(E3,V3),
	solve(E3,V3,C3).

%%% keep_leq_maxsum_cns(C1,C2,C3)
%%% C3 is the subset of constraints in C1 which are leq than some constraint in C2

keep_leq_maxsum_cns([],_C2,[]).
keep_leq_maxsum_cns(C1,C2,E) :-
	C1=[PC1|PC1s],
	((member(PC2,C2), leq_maxsum_cns(PC1,PC2)) ->
	    (E=[PC1|E1],
	    keep_leq_maxsum_cns(PC1s,C2,E1))
	;
	    keep_leq_maxsum_cns(PC1s,C2,E)
	).

leq_maxsum_cns(C1,C2) :-
	coeff_sum(C1,M1),
	coeff_sum(C2,M2),
	init_prog_max_coeff(Maxcoeff),
	{M1=<max(M2,Maxcoeff)}.


%%%%%%%% leq_maxcoeff(D1,D2,D3)

leq_maxcoeff(D1,D2,D3) :-
	copy_term(D1,(H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(H1,V1),
	term_variables(H2,V2),
	V1=V2,
	B1=B2,
	keep_leq_maxcoeff_cns(C2,C1,C3),
	D3=(H2,bd(C3,B2)).


%%%%%%%% leq_maxsum(D1,D2,D3)

leq_maxsum(D1,D2,D3) :-
	copy_term(D1,(H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(H1,V1),
	term_variables(H2,V2),
	V1=V2,
	B1=B2,
	keep_leq_maxsum_cns(C2,C1,C3),
	D3=(H2,bd(C3,B2)).


%%%%%%%% e_ch_leq(D1,D2,D3)

e_ch_leq(D1,D2,D3) :-
	copy_term(D1,(H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(H1,V1),
	term_variables(H2,V2),
	convex_hull(V1,C1,V2,C2,V3,E3),
	V1=V2,
	V2=V3,
	B1=B2,
	primitives(C1,I1),
	keep_entailed(I1,C2,E1), 
	append(E1,E3,E4),
	solve(E4,V3,E5),
	int_coeff_transf(E5,E6),
	keep_leq_maxcoeff_cns(E6,C1,C3),
	D3=(H2,bd(C3,B2)).


%%%%%%%%% convex_hull_cl

:- use_module(hull_clpq).

convex_hull_cl(D1,D2,D3) :-
	copy_term(D1,(_H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(B1,V1),
	term_variables(B2,V2),
	convex_hull(V1,C1,V2,C2,V3,C3),
	int_coeff_transf(C3,C4),
	V3=V2,
	D3=(H2,bd(C4,B2)).

int_coeff_transf([],[]).
int_coeff_transf([C|Cs],[D|Ds]) :-
	transf_constraint(C,D),
	int_coeff_transf(Cs,Ds).

transf_constraint(C,D) :-
	den_list(C,Dens),
	lcm(Dens,M),
	multiply(M,C,D).

den_list(C,Dens) :- 
	C=..[_,Lhs,Rhs],
	d_list(Lhs,LDens),
	d_list(Rhs,RDens),
	append(LDens,RDens,Dens).

d_list(V,[D]) :- 
	var(V), !, 
	D=1.
d_list(rat(_,Y),[D]) :-
	!,
	D=Y.
d_list(rat(_,Y)*V,[D]) :-
	var(V), !, 
	D=Y.
d_list(P1+P2,Dens) :- !,
	d_list(P1,Dens1),
	d_list(P2,Dens2),
	append(Dens1,Dens2,Dens).
d_list(-V,[D]) :- 
	var(V), !, 
	D=1.
d_list(-rat(_,Y),[D]) :-
	!,
	D=Y.
d_list(-rat(_,Y)*V,[D]) :-
	var(V), !, 
	D=Y.
d_list(P1-P2,Dens) :-
	d_list(P1,Dens1),
	d_list(P2,Dens2),
	append(Dens1,Dens2,Dens).

lcm([N],L) :- L=N.
lcm([N1,N2|Ns],L) :-
	N is ((N1*N2)/gcd(N1,N2)),
	lcm([N|Ns],L).

multiply(M,C1,C2) :- 
	C1=..[R,Lhs1,Rhs1],
	multiply_poly(M,Lhs1,Lhs2),
	multiply_poly(M,Rhs1,Rhs2),
	C2=..[R,Lhs2,Rhs2].

multiply_poly(M,V,C) :- 
	var(V), !, 
	{Coeff=M*1},
	({Coeff=1} -> C=V ; C=Coeff*V).
multiply_poly(M,rat(X,Y),C) :-
	!,
	{N=:=M*rat(X,Y)},
	C=N.
multiply_poly(M,rat(X,Y)*V,C) :-
	var(V), !, 
	{Coeff=:=M*rat(X,Y)},
	({Coeff=1} -> C=V ; C=Coeff*V).
multiply_poly(M,P1+P2,Q1+Q2) :-
	multiply_poly(M,P1,Q1),
	multiply_poly(M,P2,Q2).
multiply_poly(M,-V,C) :- 
	var(V), !, 
	{Coeff=M*1},
	({Coeff=1} -> C=(-V) ; C=(-Coeff)*V).
multiply_poly(M,-rat(X,Y),C) :-
	!,
	{N=:=M*rat(X,Y)},
	C=(-N).
multiply_poly(M,-rat(X,Y)*V,C) :-
	var(V), !, 
	{Coeff=:=M*rat(X,Y)},
	({Coeff=1} -> C=(-V) ; C=(-Coeff)*V).
multiply_poly(M,P1-P2,Q1-Q2) :-
	multiply_poly(M,P1,Q1),
	multiply_poly(M,P2,Q2).


