% goal_replacement_rules.pl
%
% This file contains the predicates which specify the effects of applying the
% transformation rules associated with the predicates replace_goal, move_lit,
% add_goal, and delete_goal, defined in top_level.pl, after selection and
% checking of the input parameters.

:- use_module(library(lists), [ append/3, member/2 ]).
:- use_module(library(terms), [ subsumes/2 ]).

/*-----------
apply_replace_goal(Cl,Poss,Law): replace goals at position Poss in clause Cl using law Law
------------*/
apply_replace_goal(Clause_Number,Atom_Numbers,Law_Number) :- 
     current_prog(PN,CNs), member(Clause_Number,CNs),
     (label_arc(PN,(Clause_Number,(rep,Atom_Numbers, Law_Number),NewCN),PNSucc) ->
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc))); 
     (replacement(Clause_Number,Atom_Numbers,Atoms, Law_Number,Clause),
     add_clauses(NewCN,[Clause],aux),            
     update_programs((Clause_Number,(rep,Atom_Numbers, Law_Number),NewCN), _))),
     (derivation_trace_on ->
     new_clause(replace_goal,Clause_Number,Atoms,Law_Number,NewCN);true).

replacement(Clause_Number,Atom_Numbers,Atoms,Law_Number,(H,B)) :- 
      clause_number_to_clause(Clause_Number,(H,B1)),            
      (substitute(B1,Atom_Numbers,Atoms,Law_Number,B) -> true ; (msg_impossible_rep, fail)).

substitute(B,Atom_Numbers,Atoms,Law_Number,B1) :- 
      law(Law_Number,(Left,Right)), positions_to_elts(B,Atom_Numbers,Atoms),
      delete_2(Atoms,B,B2), subsumes(Left,Atoms),
      (Atom_Numbers = [Pos|_] -> conc_position(B2,Pos,Right,B1) ; append(B2,Right,B1)).

/*-----------
apply_move_lit(Cl,Pos1,Pos2):
move the atom of clause Cl at position Pos1 to position Pos2.
------------*/
apply_move_lit(Clause_Number,N,M) :-
    current_prog(PN,CNs), member(Clause_Number,CNs),
    (label_arc(PN,(Clause_Number,(moveatom,N,M),NewCN), PNSucc) ->
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)));           
    (apply_move_lit(Clause_Number,N,M,Clause), add_clauses(NewCN,[Clause],nil),
    update_programs((Clause_Number,(moveatom,N,M),NewCN),PNSucc))),
    (derivation_trace_on ->
        new_clause(moveatom, Clause_Number,N,M, NewCN)
        ;true).
apply_move_lit(Clause_Number,N,M,(H,B2)) :-
    clause_number_to_clause(Clause_Number,(H,B)),
    positions_to_elts(B,[N],[Atom]),
    del_1(Atom,N,B,B1), (N<M ->
    (M1 is M-1, split(B1,M1,Before,A,After),
    append(Before,[A|[Atom|After]],B2))
    ; (split(B1,M,Before,A,After),
    append(Before,[Atom|[A|After]],B2))).

/*-----------
apply_add_goal(Cl,Pos,L): add goal L in clause Cl at position Pos.
------------*/
apply_add_goal(Clause_Number,Pos,L) :-
    current_clauses(CNs), member(Clause_Number,CNs),
    apply_add_goal(Clause_Number,Pos,L,Clause), add_clauses(NewCN,[Clause],nil), 
    update_programs((Clause_Number,(addgoal,Pos,L),NewCN),_), 
    (derivation_trace_on ->
        new_clause(addgoal,Clause_Number,Pos,L,NewCN)
        ;true).
apply_add_goal(Clause_Number,Pos,Goal,(H,B1)) :-
    clause_number_to_clause(Clause_Number,(H,B)), Pos1 is Pos-1,
    length1(Prefix,Pos1), append(Prefix,Suffix,B), append(Prefix,Goal,Suffix,B1).
/*-----------
apply_delete_goal(Cl,Positions): delete atoms at position Positions in the body of Cl
------------*/
apply_delete_goal(Clause_Number,Atom_Numbers) :-
    current_prog(PN,CNs), member(Clause_Number,CNs),
    (label_arc(PN,(Clause_Number,(delgoal,Atom_Numbers),NewCN), PNSucc) ->
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)));           
    (apply_delete_goal(Clause_Number,Atom_Numbers,Clause), add_clauses(NewCN,[Clause],nil),
    update_programs((Clause_Number, (delgoal,Atom_Numbers),NewCN),PNSucc))), 
    (derivation_trace_on ->
        new_clause(delgoal,Clause_Number,Atom_Numbers,NewCN);true).
apply_delete_goal(Clause_Number,Atom_Numbers,NewC) :- 
    clause_number_to_clause(Clause_Number,C),
    apply_delete_goal_(C,Atom_Numbers,NewC).
apply_delete_goal_((H,B),Atom_Numbers,(H,B1)) :-
    positions_to_elts(B,Atom_Numbers,Atoms), delete_2(Atoms,B,B1).

apply_delete_goal2(Clause_Number,Atom_Numbers,NewCN) :-
    current_prog(PN,CNs), member(Clause_Number,CNs),
    (label_arc(PN,(Clause_Number,(delgoal,Atom_Numbers),NewCN), PNSucc) ->
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)));           
    (apply_delete_goal(Clause_Number,Atom_Numbers,Clause), add_clauses(NewCN,[Clause],nil),
    update_programs((Clause_Number, (delgoal,Atom_Numbers),NewCN),PNSucc))), 
    (derivation_trace_on ->
        new_clause(delgoal,Clause_Number,Atom_Numbers,NewCN)
        ;true).


