% MODE: bf_unfold(+Def,+Prog-UDefs)
% TYPE: bf_unfold(def,list(def),def)
% SEMANTICS: unfold the negative literal in the boy of the given clause
bf_unfold((P,[(H,bd(C,B))]), Prog, (P,UnfoldedC)) :-
        unfold_all_clauses((P,[(H,bd(C,[],B))]), [], Prog, UnfoldedC).

% MODE: unfold_all_clauses(+Def,[],+Prog,-Done)
% TYPE: unfold_all_clauses(def,list,list(def),list(def))
% unfold_all_clauses(Cs,Prog,U) :- U is the result of unfolding all clauses
% in the lists Cs w.r.t. the definitions in Prog. A clause in Cs is represented
% as (H,B1,B2), where B1@B2 is the body. The first element of B2 is the literal
% selected for unfolding (if B2=[] no unfolding is performed)
unfold_all_clauses((_,[]),Done,_,Done).
unfold_all_clauses((P,[(H,bd(C,B,[]))|Cs]),CC,Prog,Done) :-
	append(CC,[(H,bd(C,B))],CCs),
		%(do_stats -> add_to_num_cls_u(1) ; true),
	unfold_all_clauses((P,Cs),CCs,Prog,Done).
unfold_all_clauses((P,[(H,bd(C,B,[L|Ls]))|Cs]),UCs,Prog,Done) :-
	unfold_one_clause((H,bd(C,B,[L|Ls])),Prog,UC),
		%(do_stats -> add_to_num_steps_u(1) ; true),
	append(UC,Cs,NewCs),
	unfold_all_clauses((P,NewCs),UCs,Prog,Done).

unfold_one_clause((H,bd(C,B,[])), _, (H,bd(C,B,[]))).
unfold_one_clause((H,bd(C,B,[A|Ls])), Prog, UC) :-
        A\=(\+_),
        !,
        findall(C1,pos_unfold((H,bd(C,B,[A|Ls])),Prog,C1),UC).
unfold_one_clause((H,bd(C,B,[(\+ A)|Ls])), Prog, UC) :-
        neg_unfoldable(A,Prog),
        !,
        findall(C1,neg_unfold((H,bd(C,B,[(\+ A)|Ls])),Prog,C1),UC).
unfold_one_clause((_,bd(_,_,[(\+ A)|_])),_,_) :-
        error_msg(neg_lit,[(\+ A)]),
        % discard_tmp_program, -- CONSISTENCY ?
        fail.

neg_unfoldable(A,Prog) :-
        \+ exists_head_which_binds_or_body_with_existential_vars(A,Prog).

exists_head_which_binds_or_body_with_existential_vars(A,Prog) :-
        functor(A,P,Arity),
	def(P/Arity,Prog,Def),
	find_unifiable_clause(A,Def,(H,bd(_,B))), !,
        % functor(H,P,Arity),
        % tmp_program_clause(H,bd(_,B)),
        % unifiable(A,H),
        binds_or_body_with_existential_vars(A,H,B).


pos_unfold((H,bd(C,B,[A|Ls])),Prog,(H,bd(Constraint,NewBody,Ls))) :-
        functor(A,P,Arity),
	def(P/Arity,Prog,Def),
	find_unifiable_clause(A,Def,(Head,bd(Const,Body))),
	%copy_term((H,bd(C,B,)),(Head2,bd(Const2,Body2)))
	unify_with_occurs_check(A,Head),
	% tmp_program_clause(A,bd(C2,Body)),
        append(B,Body,NewBody),
        append(C,Const,Cand),
        % copy_term(Cand,CTest),
        % satisfiable(CTest),
	linking_variables((H,bd(NewBody,Ls)),Cand,Vars),
	solve(clpr,Cand,Vars,Constraint),
	(Constraint==[false] -> fail ; true).
	
	% linking_variables((H,B,[A|Ls]),C,Vars1),
	% linking_variables((Head,Body),Const,Vars2),
        % ( (\+ is_truth_value(C,Vars1), \+ is_truth_value(Const,Vars2))
        %	->
        %		Constraint = Cand
        %			;
        %		( is_truth_value(C,Vars1) ->
	%			Constraint = Const;
	%			Constraint = C
        %		)
        %).

neg_unfold((H,bd(C,B,[(\+ A)|Ls])),Prog,(H,bd(NewC,NewB,Ls))) :-
	unifying_def(A,Prog,Def),		% unifying definitions but not unified
	(Def==(_,[]) ->
		(NewB=B, NewC=C) ;	% if there is no clause unifying delete atom
		(negated((H,bd(C,B,[(\+ A)|Ls])),Def,bd(NewC,NewB)),
		 linking_variables((H,NewB),NewC,Vars),
		 solve(clpr,NewC,Vars,NewC2),		% eliminate existential variables
		 (NewC2==[false] -> fail; true))	% check not tautology
	).
		% ,
		% conjunction(C1,CU2,Cand),
		% copy_term(Cand,CTest),
		% satisfiable(CTest),
		% append(B,Body,B1),
		% linking_variables((H,B,[A|Ls]),C,Vars1),
		% linking_variables((H,Body),CU2,Vars2),
		% ( (\+ is_truth_value(C1,Vars1), \+ is_truth_value(C2,Vars2))
        	% 	->
        	% 		Constr = Cand
        	% 			;
		% 			( is_truth_value(C1,Vars1) -> Constr = CU2
        	% 				   ;
        	% 			      Constr = C1
		% 			 )
		% )

% MODE: unifying_def(+A,+Prog,-UDef)
% TYPE: unifying_def(atom,list(def),def)
% SEMANTICS: returns [] either if the predicate is undefined in Prog or if the
% given atom has no unifying clause
unifying_def(A,Prog,(P/Arity,UDef)) :-
	functor(A,P,Arity),
	(def(P/Arity,Prog,Def) ->
	  (findall((H,bd(C,B)),(find_unifiable_clause(A,Def,(H,bd(C,B)))),UDef)); % ,unify_with_occurs_check(A,H)
	   UDef = []).

% TYPE: negated(clause,def,body?).
% MODE: negated(+Clause,+Def,-Body).
% SEMANTICS: notice CL = (H,bd(C1,B,[(\+ A)|Ls]))
negated((_,bd(Cons,Body,_)),(_,[]),bd(Cons,Body)).
% Negate a constraint and take the rest of constraints
negated((Head,bd(Cons,Body,[(\+ A)|Ls])),(X,[(H,bd(C,_))|Defs]),bd(FinalConstraint,NewB)) :-
	unify_with_occurs_check(A,H),
        extract(C,Constr,Before,_),				% extract form a set an element
								% and return the others

	%%% linking_variables((H,bd(Before,After,B)),Constr,Vars1),
	%%% solve(clpr,[Constr],Vars1,NewConstr),		% eliminate existential variables
	NewConstr=[Constr],
	(NewConstr==[] -> fail; true),				% check not tautology
	
        negated((Head,bd(Cons,Body,[(\+ A)|Ls])),(X,Defs),bd(NewCs,NewB)),
	
	append(Before,NewCs,NewCs2),
	%%% linking_variables((Head,Body),(NewCs2,NewConstr),Vars2),
	% negate and convert into a (disjunction of) basic constraints
	(is_equality(NewConstr) ->
			(	% it is an equality
				complement_eq(NewConstr,NotC1,NotC2),	% negation of ONE equality
									% is TWO constraints
				(append(NotC1,NewCs2,NewConstraint);	% in OR so that "findall" catches both
				 append(NotC2,NewCs2,NewConstraint))
			);
			(	% it is NOT an equality
				complement_c(NewConstr,NotNewConstr),
				append(NotNewConstr,NewCs2,NewConstraint)
			)
	),	
	%%% solve(clpr,NewConstraint,Vars2,FinalConstraint),	% test if false and minimize
	FinalConstraint=NewConstraint,
	(FinalConstraint==[false] -> fail ; true).
% Take all the constraints and negate an atom
negated((Head,bd(Cons,Body,[(\+ A)|Ls])),(X,[(H,bd(C,B))|Defs]),bd(FinalConstraint,NewB2)) :-
	unify_with_occurs_check(A,H),
        extract(B,L,_,_),
	
        complement_lit(L,NotL),
        negated((Head,bd(Cons,Body,[(\+ A)|Ls])),(X,Defs),bd(NewCs,NewB)),
	
        append(C,NewCs,NewConstraint),
	append(NewB,[NotL],NewB2),
	%%% linking_variables((Head,[NotL|NewB]),NewConstraint,Vars2),
	%%% solve(clpr,NewConstraint,Vars2,FinalConstraint),	% test if false and minimize
	FinalConstraint=NewConstraint,
	(FinalConstraint==[false] -> (fail,!) ; true).

binds_or_body_with_existential_vars(A,H,_):-
        \+ subsumes_chk(H,A),
        !.
binds_or_body_with_existential_vars(_,H,B):-
        existential_variable((H,B),_),
        !.

%existential_variable(+(H,B),?X) not functional
existential_variable((H,B),X) :-
          term_variables(H,VH),
          term_variables(B,VB),
          member(X,VB),
          is_not_member(X,VH).

