% ==============================================
% DEFINE & FOLD SUB-STRATEGY
% Single clause, block-folding

% MODE: define_fold_arrays(+Def,-FoldedDef,-IntroducedDefs)
% TYPE: define_fold_arrays(def,def,list(def))

% SEMANTICS:	
define_fold_arrays((Pred,[]),(Pred,[]),[]).
define_fold_arrays((Pred,[(H,bd(C1,C2,B))|Cls]),(Pred,[FoldedC|FoldedCls]),IntroducedDefs) :-
	% fold one clause and possibly generate some new definitions (at most one)
        define_fold_one_clause_arrays((H,bd(C1,C2,B)),FoldedC,NewDC), !,
	% add the new definition (at most one) to the set of introduced definitions
	( NewDC=[Def] -> IntroducedDefs=[Def|IntrDef] ; IntroducedDefs=IntrDef ),
	% fold the rest of clauses
        define_fold_arrays((Pred,Cls),(Pred,FoldedCls),IntrDef).

% MODE: define_fold_one_clause_arrays(+Clause,-FoldedC,-NewDefs)
% TYPE: define_fold_one_clause_arrays(((predicate/arity),clause),clause,list(def))
% SEMANTICS:
%	     skip if
%		1. is a fact
%		2. has no existential variables
%	     otherwise
%		1. provide a candidate definition for folding
%		2. check whether it already exists in the store (definition_ar/2)
%		3. fold the clause
%		4. (possibly) return the new definition that has been introduced
%
define_fold_one_clause_arrays((H,bd(C,CEx,[])),(H,bd(NewC,[])),[]) :-
	append(C,CEx,NewC).
define_fold_one_clause_arrays((H,bd(C,CEx,B)),(H,bd(NewC,B)),[]) :-
	(existential_variable((H,B),_) -> fail; append(C,CEx,NewC)). % otherwise comment this
%	(CEx \== [] ->
%		(append(C,CEx,C2),
%		 term_variables((H,B),Vars),
%		 solve(clpr,C2,Vars,NewC)) ;		% MAYBE USELESS !!
%		NewC = C
%	).
define_fold_one_clause_arrays((H,bd(C,CEx,B)),(H,bd(C,[Head])),IntroducedDef) :-

%	setof(X,existential_variable((H,B),X),ExVars),
%	separe_exist_body(B,ExVars,BEx,BNoEx),				% select existential subgoal

	candidate_definition_arrays((H,bd(C,CEx,B)),CDef,VarsSubst),	% compute candidate definition
	test_if_necessary_arrays(H,CDef,Head,IDef),			% return the head of the selected Def
									% here also assert the new def if necessary
	copy_term(IDef,IDef2),
	bind_substitutions(VarsSubst),
	(IDef==[] ->
		IntroducedDef=[] ;
		( functor(Head,Pred,Arity),
		  IntroducedDef=[(Pred/Arity,IDef2)] )
	).


candidate_definition_arrays((H,bd(C,CEx,B)),(Newp,bd(GenNewCExF,BNoF)),VarsSubst) :-

	move_functions_away(B,(BNoF,F)),	% B --> BNoF

	term_variables((H,C,BNoF),VarsOkCExF),
	append(CEx,F,CExF),
	solve(clpr,CExF,VarsOkCExF,NewCExF),	% CEx --> NewCExF

	setof(X,existential_variable((H,(NewCExF,BNoF)),X),ExVars),

	remove_neg_zero(NewCExF,NewCExF2),
	generalize_on_sum(NewCExF2,ExVars,GenNewCExF,VarsSubst,VarsH),
	
	setof(X,existential_variable((H,(GenNewCExF,BNoF)),X),ExVars2),
	
	term_variables((GenNewCExF,BNoF),Vars),
	subtract2(ExVars2,VarsH,ExVars3),
	subtract2(Vars,ExVars3,NewVars),
        Newp=..[dummy_pred|NewVars].

% MODE: test_if_necessary(+CandidateDef,-Head,-IntroducedDef)
% TYPE: test_if_necessary(clause,atom,clause)
% SEMANTICS: 
test_if_necessary_arrays(H,(Newp,bd(CExF,BExNoF)),Head,IDef) :-
	(foldable_arrays(H,Newp,bd(CExF,BExNoF),Head) -> 
		IDef = [] ; 
		(
		 Newp=..[_|Vars],
		 gen_new_name_functor(Pred),
		 Head=..[Pred|Vars],
		 functor(Head,Pred,Arity),
		 assert(definition_ar(Pred/Arity,[(Head,bd(CExF,BExNoF))])),
		 type_of_def(Pred,Arity,Vars,BExNoF),
		 IDef = [(Head,bd(CExF,BExNoF))]
		)
	).

% ------------------------------------
% TEST IF CLAUSE IS FOLDABLE

% MODE: foldable_ex(+Body,+Definitions,-Head)
% TYPE: foldable_ex(body,list(def),atom)
% SEMANTICS: foldable((H,B),(H1,B1)) :-  H1:-B1 is a definition clause
%   such that B is a variant of B1 and, after unifying B and B1,
%   forall X in B, X existential in H:-B iff X existential in H1:-B
%   we also verify condition
%			forall Y. (C <=> C1) EQUIVALENCE !!! (as we want that C disappears!!)
%   where C1 is the set of constraints in the clause used for folding
%   and Y is the set fv(C) U fv(C1) of variables of the constraints
% ~~~~~~ foldable_ex(H,bd(C,CEx,B),Head) :-
foldable_arrays(H,Newp,bd(CEx,B),Head) :-	% H         : the original head
						% Newp      : the candidate head
						% bd(CEx,B) : the body of the clause
						% Head      : the folded definition
	definition_ar(_,[(Head,bd(D,Body))]),
	
	functor(Newp,_,Arity),functor(Head,_,Arity),
	
	introduce_slack_variables(B,G,Slack1),
	introduce_slack_variables(Body,Gprime,Slack2),

	v_sort(G,Gprime,SortedG),	% G is sorted into SortedG
	subsumes_chk(Gprime,SortedG),
	SortedG=Gprime,
	
	append(CEx,Slack1,CExSlack),
	append(D,Slack2,DSlack),
	
	linking_variables(H,(CExSlack,SortedG),Unprimed),
	linking_variables(Head,(DSlack,Gprime),Primed),
	
	term_variables(CExSlack,V1),
	term_variables((H,SortedG),Cx1),
	term_variables(DSlack,V2),
	term_variables((Head,Gprime),Cx2),
	subtract2(V1,Cx1,ExVc),
	subtract2(V2,Cx2,ExVd),
	% setof(Vc,existential_variable((H,(CExSlack,SortedG)),Vc),TmpExVc),
	% setof(Vd,existential_variable((Head,(DSlack,Gprime)),Vd),TmpExVd),
		
	term_variables(Gprime,SlackVariables),
	
	% subtract2(TmpExVc,SlackVariables,ExVc),
	% subtract2(TmpExVd,SlackVariables,ExVd),
	
/*	
	move_functions_away(B,(BNoF,F)),	% B --> BNoF
	append(CEx,F,CExF),			% CEx --> CExF

	term_variables((Newp,BNoF),VarsOkCExF),
	solve(clpr,CExF,VarsOkCExF,NewCExF),	% CExF --> NewCExF
	
	v_sort(BNoF,Body,SortedB),	% BNoF is sorted into SortedB
	subsumes_chk(Body,SortedB),	
	%%%%% copy_term((Head,bd(Cfold,Body)),(Head1,bd(Cfold1,Body1))),

	move_functions_away(SortedB,(SortedBNoF,SBF)),	% SortedB --> SortedBNoF
	move_functions_away(Body,(BodyNoF,BF)),		% SortedB --> SortedBNoF
	
	SortedBNoF=BodyNoF,
	append(NewCExF,SBF,NewCExF2),
	append(Cfold,BF,Cfold2),
	
	linking_variables(H,(NewCExF2,SortedBNoF),Unprimed),
	linking_variables(Head,(Cfold2,BodyNoF),Primed),
	
	setof(Vc,existential_variable((H,bd(CExF,B)),Vc),ExVc),
	setof(Vd,existential_variable((Head,bd(Cfold,Body)),Vd),ExVd),
	
	term_variables(Body,SlackVariables),
*/

	% ISSUES !!!!!!!!!!!
	% "NewCExF2" and "Cfold2" TO FORMULA
	constr_to_formula(CExSlack,CexFormula),
	constr_to_formula(DSlack,DFormula),

	nl, print('Clause          : '), write((H,bd(CEx,B))),
	nl, print('Definition      : '), write((Head,bd(D,Body))),
	nl, print('Computing theta : '), write(a(SlackVariables,(e(ExVc,CexFormula) <-> e(ExVd,DFormula)))),
	
	% find suitable substitution
	solve_f(a(SlackVariables,(e(ExVc,CexFormula) <-> e(ExVd,DFormula))),[Solved]), !,
	
	nl, print('Computed theta : '), write(Solved),
	nl, print('Foldable test  : '), write(a(Unprimed,e(Primed,Solved))), nl,nl,
	
	% verify if the folding entails
	solve_f(a(Unprimed,e(Primed,Solved)),[true]), !,
	
	choose_one_solution(Solved,C),
	
	bind_equalities(C).

bind_equalities([]).
bind_equalities([(A=B) |As]) :-
	A=B, bind_equalities(As).
bind_equalities([(A=:=B) |As]) :-
	A=B, bind_equalities(As).

constr_to_formula([],true).
constr_to_formula([A],A).
constr_to_formula([A|Rest], (A and FRest)) :-
	constr_to_formula(Rest, FRest).

% QUALE  IL CRITERIO CON CUI VIENE SCELTA UNA TRA LE SOLUZIONI ???
% PERCH dovrebbe essere per forza un insieme di EQUAZIONI?
choose_one_solution(A,C) :-
	constraint(A), flatten_and(A,C).
choose_one_solution(A or _, C) :-
	choose_one_solution(A, C).

	
	% ~~~~~~~~~ %
/*
	SortedB=Body,
	term_variables(Body,BVars),
	term_variables(NewCExF,Vars1),
	term_variables(Cfold,Vars2),
	subtract2(Vars1,BVars,NewVars1),
	subtract2(Vars2,BVars,NewVars2),
	unify_permute(NewVars1,NewVars2),	% unify only those variables that are NOT in the body
	% ~~~~~~ append(C,NewCExF,NewC),  %%%
	% ~~~~~~ entails(NewC,Cfold). %%%
	equivalent_constraints(NewCExF,Cfold). %%% entails(NewCExF,Cfold).
	%%% entails(CEx,Cfold).
	%%%%% term_variables(Cfold1,CfoldVars),
	%% solve(clpr,Cfold1,CfoldVars,SolvedCfold),
	
	%%%%% (variant_constraint(CEx,SolvedCfold) ; variant_constraint(CEx,Cfold1)).
	%% entails(C,Cfold1).
*/

% ===========================================
% MODE: move_functions_away(+Body,-BodyNfun)
% TYPE: move_functions_away(list(atom),(list(atom),list(constraint)))
% SEMANTICS:
move_functions_away([],([],[])).
move_functions_away([A|Cls],([(NewA)|NewCls],Fun)) :-
	A\=(\+_),
	replace_functions_by_vars(A,NewA,F),
	append(F,Fs,Fun),
	move_functions_away(Cls,(NewCls,Fs)).
move_functions_away([\+A|Cls],([(\+NewA)|NewCls],Fun)) :-
	replace_functions_by_vars(A,NewA,F),
	append(F,Fs,Fun),
	move_functions_away(Cls,(NewCls,Fs)).
% move_functions_away([p(A),\+q(X-1,Y),q1([X+1|Xs])],Out).
% > Out = [p(A),q(_A,Y),q1([_B|Xs])],[_A=X-1,_B=X+1]

% MODE: replace_functions_by_vars(+A,-NewA,-F)
% TYPE: replace_functions_by_vars(atom,atom,list(var=term)),
replace_functions_by_vars(H,NewH,F) :-
	H =.. [Pred|Args],
	replace_each_function_by_variable(Args,NewArgs,F), !,
	NewH =.. [Pred|NewArgs].
% replace_functions_by_vars(p(A,[X|Xs],B+1-C,[D+1|T]),NewP,F).
% F = [_A=B+1-C,_B=D+1],
% NewP = p(A,[X|Xs],_A,[_B|T])

% MODE: replace_each_function_by_variable(+Args,-NewArgs,-F)
% TYPE: replace_each_function_by_variable(list(term),list(term),list(var=term))
replace_each_function_by_variable([],[],[]).
replace_each_function_by_variable([A|Args],[A|NewArgs],F) :-
	(var(A);ground(A)),
	replace_each_function_by_variable(Args,NewArgs,F).
replace_each_function_by_variable([A|Args],[A|NewArgs],F) :-
	A=[H|_], var(H),
	replace_each_function_by_variable(Args,NewArgs,F).
replace_each_function_by_variable([A|Args],[B|NewArgs],[(X=H)|F]) :-
	nonvar(A), A=[H|T], nonvar(H), B=[X|T],
	replace_each_function_by_variable(Args,NewArgs,F).
replace_each_function_by_variable([A|Args],[X|NewArgs],[(X=A)|F]) :-
	nonvar(A),
	replace_each_function_by_variable(Args,NewArgs,F).

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% MODE: introduce_slack_variables(+Body,-BodyNfun)
% TYPE: introduce_slack_variables(list(atom),(list(atom),list(constraint)))
% SEMANTICS:
introduce_slack_variables([],[],[]).
introduce_slack_variables([A|Cls],[(NewA)|NewCls],Fun) :-
	A\=(\+_),
	replace_all_arguments_by_vars(A,NewA,F),
	append(F,Fs,Fun),
	introduce_slack_variables(Cls,NewCls,Fs).
introduce_slack_variables([\+A|Cls],[(\+NewA)|NewCls],Fun) :-
	replace_all_arguments_by_vars(A,NewA,F),
	append(F,Fs,Fun),
	introduce_slack_variables(Cls,NewCls,Fs).
% introduce_slack_variables([p(A),\+q(X-1,Y),q1([X+1|Xs])],Out,Slacks).
% > Out = [p(A),q(_A,Y),q1([_B|Xs])]
% > Slacks = [_A=X-1,_B=X+1]


% MODE: replace_all_arguments_by_vars(+A,-NewA,-F)
% TYPE: replace_all_arguments_by_vars(atom,atom,list(var=term)),
replace_all_arguments_by_vars(A,NewA,F) :-
	A =.. [Pred|Args],
	replace_each_argument_by_variable(Args,NewArgs,F), !,
	NewA =.. [Pred|NewArgs].

% MODE: replace_each_argument_by_variable(+Args,-NewArgs,-F)
% TYPE: replace_each_argument_by_variable(list(term),list(term),list(var=term))
replace_each_argument_by_variable([],[],[]).
% it is a list :
replace_each_argument_by_variable([A|Args],[[X|T]|NewArgs],[(X=H)|F]) :-
	subsumes([H|T],A),
	replace_each_argument_by_variable(Args,NewArgs,F).
% it is not a list :
replace_each_argument_by_variable([A|Args],[X|NewArgs],[(X=A)|F]) :-
	\+ subsumes([_|_],A),
	replace_each_argument_by_variable(Args,NewArgs,F).
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

remove_neg_zero([],[]).
remove_neg_zero([C|Cs],[NewC|NewCs]) :-
	C=..[Relation,Arg1,Arg2],
	Arg1==(-0.0),
	NewC=..[Relation,(0.0),Arg2],
	remove_neg_zero(Cs,NewCs).
remove_neg_zero([C|Cs],[NewC|NewCs]) :-
	C=..[Relation,Arg1,Arg2],
	Arg2==(-0.0),
	NewC=..[Relation,Arg1,(0.0)],
	remove_neg_zero(Cs,NewCs).
remove_neg_zero([C|Cs],[C|NewCs]) :-
	remove_neg_zero(Cs,NewCs).

generalize_on_sum([],_,[],[],[]).
generalize_on_sum([C|Cs],ExVars,[GenC|GenCs],NewVarsSubst,NewVarsH) :- 
	C=..[Relation,Arg1,Arg2],
	\+ ground(Arg1), \+ ground(Arg2), \+ var(Arg1), \+ var(Arg2),
	max_polynomial_with_no_exvars(Arg1,ExVars,PolyEx1,PolyNoEx1),
	max_polynomial_with_no_exvars(Arg2,ExVars,PolyEx2,PolyNoEx2),
	GenC=..[Relation,(X + PolyEx1),(Y + PolyEx2)],
	(PolyNoEx1==(0.0) ->
		(NewArg1=PolyEx1, SubstArg1=[], VarsArg1=[]) ;
		(NewArg1=(X+PolyEx1), SubstArg1=[X=PolyNoEx1], VarsArg1=[X])
	),
	(PolyNoEx2==(0.0) ->
		(NewArg2=PolyEx2, SubstArg2=[], VarsArg2=[]) ;
		(NewArg2=(Y+PolyEx2), SubstArg2=[Y=PolyNoEx2], VarsArg2=[Y])
	),
	GenC=..[Relation,NewArg1,NewArg2],
	
	append(SubstArg1,SubstArg2,NewSubst),
	append(NewSubst,VarsSubst,NewVarsSubst),
	
	append(VarsArg1,VarsArg2,NewVars),
	append(NewVars,VarsH,NewVarsH),
	
	generalize_on_sum(Cs,ExVars,GenCs,VarsSubst,VarsH).
generalize_on_sum([C|Cs],ExVars,[GenC|GenCs],NewVarsSubst,NewVarsH) :- 
	C=..[Relation,Arg1,Arg2],
	ground(Arg1), \+ ground(Arg2), \+ var(Arg2),
	max_polynomial_with_no_exvars(Arg2,ExVars,PolyEx2,PolyNoEx2),
	(PolyNoEx2==(0.0) ->
		(GenC=..[Relation,Arg1,PolyEx2],
		 NewVarsSubst=VarsSubst,
		 NewVarsH=VarsH) ;
		(GenC=..[Relation,Arg1,(X + PolyEx2)],
		 NewVarsSubst=[X=PolyNoEx2|VarsSubst],
		 NewVarsH=[X|VarsH])
	),
	generalize_on_sum(Cs,ExVars,GenCs,VarsSubst,VarsH).
generalize_on_sum([C|Cs],ExVars,[GenC|GenCs],NewVarsSubst,NewVarsH) :- 
	C=..[Relation,Arg1,Arg2],
	\+ ground(Arg1), ground(Arg2), \+ var(Arg1),
	max_polynomial_with_no_exvars(Arg1,ExVars,PolyEx1,PolyNoEx1),
	(PolyNoEx1==(0.0) ->
		(GenC=..[Relation,PolyEx1,Arg2],
		 NewVarsSubst=VarsSubst,
		 NewVarsH=VarsH) ;
		(GenC=..[Relation,(X + PolyEx1),Arg2],
		 NewVarsSubst=[X=PolyNoEx1|VarsSubst],
		 NewVarsH=[X|VarsH])
	),
	generalize_on_sum(Cs,ExVars,GenCs,VarsSubst,VarsH).
generalize_on_sum([C|Cs],ExVars,[C|GenCs],VarsSubst,VarsH) :- 
	generalize_on_sum(Cs,ExVars,GenCs,VarsSubst,VarsH).

% MODE: max_polynomial_with_no_exvars(+InitialPoly,+ExVars,-MaxPolyEx,-RestOfPoly)
% TYPE: max_polynomial_with_no_exvars(polynomial,list(var),polynomial,polynomial)
% SEMANTICS:
max_polynomial_with_no_exvars(InitialPoly,ExVars,PolyEx,PolyNoEx) :-
	subsumes(Arg1+Arg2,InitialPoly),		% InitialPoly=..[+,Arg1,Arg2],
	max_polynomial_with_no_exvars(Arg1,ExVars,PolyEx1,PolyNoEx1),
	max_polynomial_with_no_exvars(Arg2,ExVars,PolyEx2,PolyNoEx2),
	sum_poly(PolyEx1,PolyEx2,PolyEx),
	sum_poly(PolyNoEx1,PolyNoEx2,PolyNoEx).
max_polynomial_with_no_exvars(Addendum,ExVars,PolyEx,RestOfPoly) :-
	\+ subsumes(_+_,Addendum),		% Addendum\=(_+_),
	term_variables(Addendum,Vars),
	((member(V,Vars), is_member(V,ExVars)) ->	% if has an existential variable
		(PolyEx=Addendum, RestOfPoly=(0.0)) ;	% then put into the existential part of the polynomial
		(PolyEx=(0.0),    RestOfPoly=Addendum)	% else put into the non-existential part
	).
% max_polynomial_with_no_exvars(A+B+C,[

sum_poly(A,B,C) :-
	( (A\==(0.0), B\==(0.0)) ->
		C=(A + B) ;
		( A==(0.0) -> C=B ; C=A)
	).

%%%%% E' sound????????
unify_permute(_,[]).
unify_permute([],_).
unify_permute([V|Vars1],Vars2) :-
	del_elem(V2,Vars2,RemainingVars2),
	V=V2,
	unify_permute(Vars1,RemainingVars2).	


bind_substitutions([]).
bind_substitutions([S|Ss]) :-
	S, bind_substitutions(Ss).


type_of_def(Pred,Arity,Vars,Body) :-
	collect_vars_and_types(Body,VarsT),
	find_types(Vars,VarsT,Type),
	assert(type(Pred,Arity,Type)).

collect_vars_and_types([],[]).
collect_vars_and_types([list(X)|Ls],[X/list|VarsT]) :-
	collect_vars_and_types(Ls,VarsT).
collect_vars_and_types([(\+ A)|Ls],VarsT) :-
	functor(A,P,Arity),
	A=..[P|Vars],
	type(P,Arity,Type),
	associate_vars_w_types(Vars,Type,VarsT1),
	append(VarsT1,VarsT2,VarsT),
	collect_vars_and_types(Ls,VarsT2).
collect_vars_and_types([A|Ls],VarsT) :-
	A\=(\+_),
	functor(A,P,Arity),
	A=..[P|Vars],
	type(P,Arity,Type),
	associate_vars_w_types(Vars,Type,VarsT1),
	append(VarsT1,VarsT2,VarsT),
	collect_vars_and_types(Ls,VarsT2).

associate_vars_w_types([],[],[]).
associate_vars_w_types([V|Vars],[T|Type],[(V/T)|VarsT]) :-
	associate_vars_w_types(Vars,Type,VarsT).

find_types([],_,[]).
find_types([V|Vars],VarsT,[list|Type]) :-
	is_member((V/list),VarsT),
	find_types(Vars,VarsT,Type).
find_types([_|Vars],VarsT,[none|Type]) :-
	find_types(Vars,VarsT,Type).


