%------------------------------------
% Eliminate_Conj Strategy

% MODE: definite-conj-elim(+Clause,+Pin,-Pout)
% TYPE: definite-conj-elim(clause,list(Def),list(Def))
% SEMANTICS: we are given the definition Def of a given predicate
% such that each clause in Def has the conjunction of two or more atoms
% in the body, and we obtain a new definition DefOk where each clause has
% at most one atom in the body and, possibly, a new set of definitions NewDefs
definite_conj_elim((P,CLs),Pin,Pout) :-
	%%% assert(definition(P,CLs)),
	write(' -- Eliminate Conjunction Substrategy --'),nl,nl,
	eliminate_conjunction_strategy([(P,CLs)],Pin,Pout).

% MODE: eliminate_conjunction_strategy(+ListDef,+Prog,-NewProg)
% TYPE: eliminate_conjunction_strategy(list(def),list(def),list(def))
% SEMANTICS: eliminate negation from the given definition and
%   all the definitions introduced along the way
eliminate_conjunction_strategy([],NewProg,NewProg).
eliminate_conjunction_strategy([(_,[(H,bd(C,B))])],Prog,NewProg) :-

	( derivation_trace_on -> (write('   New definition :'),nl,show_def((_,[(H,bd(C,B))])),nl) ; true),

	functor(H,Pred,Arity),
	
	% breadth-first unfold
	bf_unfold((Pred/Arity,[(H,bd(C,B))]),Prog,UDef),
	append(UDef,Prog,NewProg).
	
	% ,	% breadth-first unfold 1-level

	% write('Unfold :'),nl,
	% show_def(UDef),nl,
	
	% simplify_arrays(UDef,SimplifiedUDef),	% printing embedded

	% define_fold(SimplifiedUDef,FoldedCLs,ToDoDefs),		% define-fold

	% write('Fold :'),nl,
	% show_def(FoldedCLs),nl,

	% append(ToDoDefs,CLs,NewCLs),
	% eliminate_conjunction_strategy(NewCLs,[FoldedCLs|Prog],NewProg).


%------------------------------------
% Eliminate_Negation Strategy

% MODE: neg-elim(+Clause,+Prog,-NewProg)
% TYPE: neg-elim(clause,list(def),list(def))
% SEMANTICS: we are given the definition Def of a given predicate
%   such that each clause in Def has the atom in the body negated,
%   and we obtain a new definition DefOk with no negated atoms and,
%   possibly, a new set of definitions NewDefs
neg_elim((P,CLs),Pin,Pout) :-
	assert(definition_neg(P,CLs)),
	write('B. Eliminate Negation Substrategy'),nl,nl,
	eliminate_negation_strategy([(P,CLs)],Pin,Pout).

% MODE: eliminate_negation(+ListDef,+Prog,-NewProg)
% TYPE: eliminate_negation(list(def),list(def),list(def))
% SEMANTICS: eliminate negation from the given definition and
%   all the definitions introduced along the way
eliminate_negation_strategy([],NewProg,NewProg).
eliminate_negation_strategy([(_,[(H,bd(C,B))])|CLs],Prog,NewProg) :-

	(derivation_trace_on -> (write(' - New definition :'),nl,show_def((_,[(H,bd(C,B))])),nl) ; true),

	functor(H,Pred,Arity),
	
	% breadth-first unfold
	bf_unfold((Pred/Arity,[(H,bd(C,B))]),Prog,UDef),	% breadth-first unfold 1-level

	(derivation_trace_on -> (write(' - Unfold :'),nl,show_def(UDef),nl) ;true),
	
	simplify_arrays(UDef,SimplifiedUDef),		% printing embedded
	
	constraint_replacement(SimplifiedUDef,NewUDef),
	
	(derivation_trace_on -> (write(' - Replace Constraint :'),nl,show_def2(NewUDef),nl) ; true),
	
	%%%%%%%
	%%% convert_clauses_syntax(),
	%%%%%%%
	
	define_fold_neg(SimplifiedUDef,FoldedCLs,ToDoDefs),		% define-fold

	(derivation_trace_on -> (write(' - Fold :'),nl,show_def(FoldedCLs),nl) ; true),

	% append(Prog,FoldedCLs,Prog2),				% we add the folded clauses to the program
	% append(IntDefs,Defs,Defs2),				% we finish the last transformation (depth-first)
	append(ToDoDefs,CLs,NewCLs),
	eliminate_negation_strategy(NewCLs,[FoldedCLs|Prog],NewProg).

% MODE: 
% TYPE: 
% SEMANTICS: convert from (H,bd(C,B)) to (H,bd(C,CEx,B)) syntax, where CEx=[]
convert_clauses_syntax([],[]).
convert_clauses_syntax([(H,bd(C,B))|CLs],[(H,bd(C,[],B))|NewCLs]) :-
	convert_clauses_syntax(CLs,NewCLs).

%------------------------------------
% Eliminate_Ex_Var Strategy

% MODE: exist-elim(+Clause,+Pin,-Pout)
% TYPE: exist-elim(clause,list(def),list(def))
% SEMANTICS: we are given the definition Def of a given predicate
% such that each clause in Def has the atom in the body containing
% an existential variable, and we obtain a new definition DefOk with
% no existential variables and, possibly, a new set of definitions NewDefs
exist_elim((P,CLs),Pin,Pout) :-
	assert(definition_ex(P,CLs)),
	write('C. Eliminate Existential Variable Substrategy'),nl,nl,
	eliminate_existential_variable_strategy([(P,CLs)],Pin,Pout).	

eliminate_existential_variable_strategy([],NewProg,NewProg).
eliminate_existential_variable_strategy([(_,[(H,bd(C,B))])|CLs],Prog,NewProg) :-

	(derivation_trace_on -> (write(' - New definition :'),nl,show_def((_,[(H,bd(C,B))])),nl) ; true),

	functor(H,Pred,Arity),
	
	% breadth-first unfold
	bf_unfold((Pred/Arity,[(H,bd(C,B))]),Prog,UDef),			% breadth-first unfold 1-level
	
	(derivation_trace_on -> (write(' - Unfold :'),nl,show_def(UDef),nl) ; true),
	
	simplify_arrays(UDef,SimplifiedUDef),

	constraint_replacement(SimplifiedUDef,NewUDef),

	(derivation_trace_on -> (write(' - Replace Constraint :'),nl,show_def2(NewUDef),nl) ; true),
	
	define_fold_ex(NewUDef,FoldedCLs,ToDoDefs),		% define-fold

	(derivation_trace_on -> (write(' - Fold :'),nl,show_def(FoldedCLs),nl) ; true),

	% append(Prog,DefOk,Prog2),				% we add the folded clauses to the program
	% append(IntDefs,Defs,Defs2),				% we finish the last transformation (depth-first)
	append(ToDoDefs,CLs,NewCLs),
	eliminate_existential_variable_strategy(NewCLs,[FoldedCLs|Prog],NewProg).

%------------------------------------
% Eliminate Existential Variables and Negation Strategy

% MODE: eliminate_exist_vars(+Clause,+Pin,-Pout)
% TYPE: eliminate_exist_vars(clause,list(def),list(def))
% SEMANTICS: we are given the definition Def of a given predicate
% such that each clause in Def has the atom in the body containing
% an existential variable, and we obtain a new definition DefOk with
% no existential variables and, possibly, a new set of definitions NewDefs
eliminate_exist_vars((P,CLs),Pin,Pout) :-
	assert(definition_ar(P,CLs)),
	write('Eliminate Existential Variable and Negation Substrategy'),nl,nl,
	eliminate_existential_variable_and_negation_strategy([(P,CLs)],Pin,Pout).	

eliminate_existential_variable_and_negation_strategy([],NewProg,NewProg).
eliminate_existential_variable_and_negation_strategy([(_,[(H,bd(C,B))])|CLs],Prog,NewProg) :-

	(derivation_trace_on -> (write(' - New definition :'),nl,show_def((_,[(H,bd(C,B))])),nl) ; true),

	functor(H,Pred,Arity),
	
	% breadth-first unfold
	bf_unfold((Pred/Arity,[(H,bd(C,B))]),Prog,UDef),			% breadth-first unfold 1-level
	
	(derivation_trace_on -> (write(' - Unfold :'),nl,show_def(UDef),nl) ; true),
	
	simplify_arrays(UDef,SimplifiedUDef),

	constraint_replacement(SimplifiedUDef,NewUDef),

	(derivation_trace_on -> (write(' - Replace Constraint :'),nl,show_def2(NewUDef),nl) ; true),
	
	define_fold_arrays(NewUDef,FoldedCLs,ToDoDefs),		% define-fold

	(derivation_trace_on -> (write(' - Fold :'),nl,show_def(FoldedCLs),nl) ; true),

	% append(Prog,DefOk,Prog2),				% we add the folded clauses to the program
	% append(IntDefs,Defs,Defs2),				% we finish the last transformation (depth-first)
	append(ToDoDefs,CLs,NewCLs),
	eliminate_existential_variable_and_negation_strategy(NewCLs,[FoldedCLs|Prog],NewProg).

