% MODE: types_of_prog(+Prog)
% TYPE: types_of_prog(list(def))
% SEMANTICS: we ASSUME that each clause in a given definition
%            has the same type
types_of_prog([]).
types_of_prog([(P/A,[Cl|_])|Defs]) :-
	types_of_clause_arrays(P/A,Cl),
	types_of_prog(Defs).

% MODE:
% TYPE:
% SEMANTICS: we ASSUME that each Def is made
types_of_clause_arrays(P/A,(H,bd(_,B))) :-
        functor(H,P,A),
	  H =.. [P|Vars],
        types_of_vars_arrays(Vars,B,List_of_types),
        assert(type(P,A,List_of_types)).

% MODE: types_of_vars_arrays(+HeadVars,+Body,-TypesList)
% TYPE:
% SEMANTICS:
types_of_vars_arrays([],_,[]).
types_of_vars_arrays([V|Vars],G,[T|Ts]) :-
       occurs_in_lit(V,G,L),
       functor(L,P,_),
	 ((P = \+) -> (L =.. [\+,Atom]); L = Atom),
	 functor(Atom,P2,A2),
	 type(P2,A2,Types),
	 Atom =.. [_|AtomVars],
	 associate_type(AtomVars,V,Types,T),
	 types_of_vars_arrays(Vars,G,Ts).
types_of_vars_arrays([_|Vars],G,[none|Ts]) :-	% if it fails then
	 types_of_vars_arrays(Vars,G,Ts).	% it is a constraint

associate_type([AV|_], V, [Type|_], Type) :-
	 AV == V.
associate_type([_|AtomVars], V, [_|Types], Type) :-
	 associate_type(AtomVars, V, Types, Type).


% MODE:
% TYPE:
% SEMANTICS: add type atoms  to all clauses of Cs
add_types_to_prog([],[]).
add_types_to_prog([Def|Defs],[NewDef|NewDefs]) :-
        add_types_to_def(Def,NewDef),
        add_types_to_prog(Defs,NewDefs).


% MODE:
% TYPE:
% SEMANTICS: add_types(Def,NewDef) :- D is derived from C by adding a type atom
%            for each variable occurring in an instantiated or negated atom

% notice: no need for typing over constraints
%         also, there is no predicate with both bit types and set types
%         or with both bit types and nat types
%         thus it is not going to happen that any bit tipe is introduced
%         in Body and not in Constraints
add_types_to_def((P,[]),(P,[])).
add_types_to_def((P,[Cl|Cls]),(P,[NewCl|NewCls])) :-
         add_types_to_cl(Cl,NewCl),
         add_types_to_def((P,Cls),(P,NewCls)).

add_types_to_cl((H,bd(C,B)),(H,bd(C,NewB))) :-
        occurs_in_neg_and_not_constr(B,VarsT),
        add_type_of_vars(VarsT,Btype),
	  append(Btype,B,NewB).


% occurs_in_neg_and_not_constr(B,VarsT). :-  VarsT is the set of variables
% which occur in an nonvariable argument of a literal or in a
% negated atom in B
occurs_in_neg_and_not_constr([],[]).
occurs_in_neg_and_not_constr([L|Ls],VarsT) :-
	L =.. [\+,Atom],
	functor(Atom,Pr,Ar),
	type(Pr,Ar,Type),
      Atom=..[_|Args],
	find_vars_not_constr(Args,Type,VarsT1),
	occurs_in_neg_and_not_constr(Ls,VarsT2),
	union(VarsT1,VarsT2,VarsT).
occurs_in_neg_and_not_constr([_|Ls],VarsT) :-
	occurs_in_neg_and_not_constr(Ls,VarsT).

find_vars_not_constr([],[],[]).
find_vars_not_constr([Arg|Args],[Type|Types],[Arg/Type|Vars]) :-
	Type \== none,
	find_vars_not_constr(Args,Types,Vars).
find_vars_not_constr([_|Args],[Type|Types],Vars) :-
	Type == none,
	find_vars_not_constr(Args,Types,Vars).


% add_type_of_vars(Vars,B,NewB) :- NewB is a goal derived from goal B
% by adding to B a type atom for each variable occurring in
% Vars

add_type_of_vars([],[]).
add_type_of_vars([V/T|Vars],[Type|Btype]) :-
	Type =.. [T,V],
        add_type_of_vars(Vars,Btype).

useful_type_atom(A,Ls) :-
        A=..[_,X],
        var(X),
        occurs_in_nonvar_or_neg_chk(X,Ls).

occurs_in_nonvar_or_neg_chk(V,B) :-
        member(L,B),
        L=..[_|Args],
        member(Arg,Args),
        nonvar(Arg),
        term_variables(Arg,Vs),
        is_member(V,Vs).

