% PROGRAM  : ufprover_arrays_new
% provides : prove/0
% requires : module(),...,library(),...,

:- use_module(library(terms)).
:- use_module(library(ordsets)).

:- op(960,xfy, [<->]).
:- op(965,xfy, [->]).
:- op(950,xfy, [or]).
:- op(900,xfy, [and]).

% ==================
%  System Libraries
%% :- use_module(library(terms), [subsumes/2,subsumes_chk/2,term_variables/2,variant/2]).
:- use_module(library(lists), [remove_duplicates/2]).
:- use_module(library(clpr)).

% ================
%  Support Files
% THIS ALREADY LOADED IN [make_map]
% :- use_module(utilities).%,[def/3,def_star/3,union/3,subtract/3,extract/3,is_member/2,is_not_member/2,member_chk/2,unifiable/2,existential_variable/2,linking_variables/3,test/3]).
:- use_module(solver_clpr).%,[solve/3,satisfiable/1,entails/2,solve_constraints/2,has_satisfiable_constraint/1]).
:- use_module(formulas).

% ================
%  Sub-Strategies
:- use_module(substrategies).
:- use_module(bf_unfold).
:- use_module(simplify).
:- use_module(constraint_replace).
:- use_module(fold).
%%% :- use_module(fold_neg).
%%% :- use_module(fold_ex).
:- use_module(typing).

%%% :- consult(basic_op).
%%% :- consult(input_output).

% DYNAMIC predicates
:- dynamic definition_ar/2.
:- dynamic definition_ex/2.
:- dynamic definition_neg/2.
:- dynamic type/3.
:- assert(derivation_trace_on).

% -------------------------------------------
% programs have the following syntax :
%
% clause: p1(H,...,m(vector)) :- c(X,n(vector)), p2(T,...,n(vector)).
%
%	either	H = []
%	    or	H = [X|Xs] and T = Xs
%
%	m(vector) is a vector of variables on natural numbers
%	n(vector) is a vector of variables and (possibly) constants on natural numbers
%
%	s.t. 	m(vector) = f(n(vector))	for some LINEAR function f
%		X = g(n(vector))		for some LINEAR function g
%
% program is provided as a 'list(def)' [see below in TYPES for details]
%
% -------------------------------------------
%
% we assume that:
% - every predicate has the types odered s.t. all the list arguments
%   are before all the integer arguments (i.e. p(list,...,list,nat,...,nat))
%
% we represent:
% - a clause as a term '(H,bd(C,B))'
%   where H is atom
%         C is list(constraint)
%         B is list(literal)
%   [see below]
%
% -------------------------------------------
%
% TYPES
%
% Base types        : term, predicate, nat, rat, any.
% Type constructors : list/1, sorted_list/1.
% Type definitions  :
%    atom       == predicate(term1,...,termN).
%    literal    == atom; \+ atom.
%    clause     == (atom, bd(list(constraint),list(literal))).
%    clause_sel == (atom,bd(list(constraint),list(literal),list(literal))).
%                  (the first literal of the second list of literals
%                   is the literal selected for unfolding)
%    body       == bd(list(constraint),list(literal)).
%    neg_clause == (\+atom, bd(list(constraint),list(literal))).
%    arity      == nat.
%    def        == (predicate/arity,list(clause)).
%    def_sel    == (predicate/arity, list(clause_sel)).
%    answer     == true; false; unknown.
%    constraint == rat=<rat; rat>=rat; rat<rat; rat>rat; rat=:=rat; rat=\=rat.
%    bit        == 0; 1.%
%
%    formula...
%
% -------------------------------------------

initialize :-
	retractall(is_type(_)),
	retractall(type(_,_,_)),
	retractall(definition_ar(_,_)),
	retractall(definition_neg(_,_)),
	retractall(definition_ex(_,_)),
	retractall(current_new_def_functor(_)),
	retractall(do_subs),
	retractall(do_bcdet),
	retractall(do_useless_elim),
	assert(is_type(list(_))).
	% ,retractall(derivation_trace_on).
	% assert(current_new_def_functor(0)).

% ===================================================
% VERIFY PREDICATE

prove :-
	check_exists_transf_tree,
	initialize,
	initial_program_and_types(Pin),
	read_property(Phi,Pin,Property), % NOTICE: Phi is a predicate name
	check_closed_fo_formula(Property),
	
	statistics(walltime,[T1,_]),
	
	lloyd_topor_transf_arrays(Phi,Property,Defs,NDefs),	% (predicate,formula,definitions,number_of_definitions)
	(derivation_trace_on ->
		(write('--- Lloyd-Topor Transformation ---'),nl,nl,
		 show_all_defs_num(Defs,NDefs), nl);
		 true
	),
	
	write('--- Existential Variables Elimination Strategy ---'), nl, nl,
	t(Defs,Pin,Pout,NDefs), nl,
	% (derivation_trace_on ->
	% 	(write('--- Final Program ---'), nl, nl,
	%	 show_all_defs(Pout), nl, nl);
	%	true
	% ),
	
	write('--- Final program ---'), nl,
	inspect(Pout,Phi,Answer), nl,nl,
	write('Answer      : '), write(Answer), nl,
	statistics(walltime,[T2,_]),

	T is T2-T1,
	write('Elapsed time: '), write(T), write(ms), nl.

t([], P, P,0).
t([Def|Defs], P1, Pfin, N) :- 
	write(N),M is N-1,write('. '),
	strategy(Def, P1, P2),
	t(Defs, P2, Pfin, M). 

%=========================================================
% Main Transformation Loop

strategy(Def,Pin,Pout) :-
	no_variables_in_bodies(Def),
	append([Def],Pin,Pout),
	write('No transformation applied'),nl,nl.

% strategy(Def,Pin,Pout) :-
%	needs_neg_elim(Def),
%	neg_elim(Def,Pin,Pout).

strategy(Def,Pin,Pout) :-
%%% 	needs_exist_elim(Def),
%%% 	% exist_elim(Def,Pin,Pout).
 	eliminate_exist_vars(Def,Pin,Pout).

%%% strategy((P,C),Pin,Pout) :-
%%% 	write('A. One-Step Breadth-first Unfold Substrategy'),nl,nl,
%%% 	
%%% 	(derivation_trace_on -> (write('New definition :'),nl,show_def((_,C)),nl) ; true),
%%% 
%%% 	% breadth-first unfold
%%% 	bf_unfold((P,C),Pin,UDef),
%%% 
%%% 	(derivation_trace_on -> (write('Unfold :'),nl,show_def(UDef),nl) ; true),
%%% 	
%%% 	simplify_arrays(UDef,SimplifiedUDef),		% printing embedded
%%% 	
%%% 	append([SimplifiedUDef],Pin,Pout).

%%% needs_neg_elim((_,[(_,bd(_,B))])) :-
%%% 	member((\+ _),B).

%%% needs_exist_elim((_,[(H,bd(C,B))])) :-
%%% 	existential_variable((H,bd(C,B)),_).

/*
% MODE: strategy(+Clause,+Prog,-FinalProg)
% TYPE: strategy(clause,list(def),list(def))
% clause ok
strategy(C, P1, P2) :- 
	definite_lr_clause(C),  
	P2 = {C} U P1.
% comes from LT-transformation
% it is a definition, is is a definite clause
% and needs Eliminate-Conjunction
strategy(C, P1, P2) :- 
	definite_conj_def(C),
	definite_conj_elim(C,P1,P2).
% comes from LT-transformation
% it is a definition and needs Eliminate-Negation
strategy(C, P1, P2) :- 
	neg_def(C),
	neg_elim(C,P1,P2).
% comes from LT-transformation
% it is a definition and needs Eliminate-Exist-Var
strategy(C, P1, P2) :- 
	exist_def(C),
	exist_elim(C,P1,P2).
% comes from the initial program
% it is NOT a definition and has a negative
% literal in the body
strategy(C, P1, P2) :- 
	C = (p([X|L],...) <- c, \+ A	),
	D = (not-A <- list(...), \+ A),
	E = (p([X|L],...) <- c, not-A),
	strategy(D, P1, P11),
	P2 = {E} U P11.
% Extra
% it is NOT a definition and has a negative
% literal in the body
strategy(C, P1, P2) :- 
	C = (H <- c, G1, \+ A, G2),
	D = (not-A <- list(...), \+ A),
	E = (H <- c, G1, not-A, G2),
	strategy(D, P1, P11),
	strategy(E, P11, P2).  	
% Extra
% it is NOT a definition and has an existential
% variable in the body
strategy(C, P1, P2) :- 
	C = (H <- c, G), 
	Y = vars(G) superset vars(H),
	D = (new(Y) <- G),
	E = (H <- c, new(Y),
	strategy(D, P1, P11),
	exist-elim(E, P11, P2).
*/
% ------------------------

% MODE: initial_program_and_property(-Prog,-Phi)
% TYPE: initial_program_and_property(list(def),predicate)
% SEMANTICS: Prog is the list of clauses of the current program
% Phi is a nullary predicate symbol defined by a single clause in the program

initial_program_and_types(Prog) :-
        findall((H1,bd(Cns1,Lits1)),
        	(current_prog(_,Ns),member(N,Ns),clause3(N,(H1,B1),_),
		 split_bd(B1,bd(Cns1,Lits1))),Cls),
	findall(type(Pr,Ar,Ts),type_definition(Pr,Ar,Ts),Types),
	assert_all(Types),
	cls_to_defs(Cls,Prog).

% MODE: assert_all(+Cls)
% TYPE: assert_all(list(clause))
assert_all([]).
assert_all([T|Ts]) :- assert(T), assert_all(Ts).

% MODE:
% TYPE:
% SEMANTICS:
cls_to_defs([],[]).
cls_to_defs([Cl|Cls],[Def|Defs]) :-
	find_complete_def(Cl,Cls,NewCls,Def),
	cls_to_defs(NewCls,Defs).

% MODE:
% TYPE:
% SEMANTICS:
find_complete_def((H,bd(C,B)),Cls,RemainingCls,(P/A,[(H,bd(C,B))|Clauses])) :-
	functor(H,P,A),
	find_all_clauses(P/A,Cls,Clauses,RemainingCls).

% MODE:
% TYPE:
% SEMANTICS:
find_all_clauses(_,[],[],[]).
find_all_clauses(P/A,[(H,bd(C,B))|Cls],[(H,bd(C,B))|Clauses],RemainingCls) :-
	functor(H,P,A),
	find_all_clauses(P/A,Cls,Clauses,RemainingCls).
find_all_clauses(P/A,[(H,bd(C,B))|Cls],Clauses,[(H,bd(C,B))|RemainingCls]) :-
	\+ functor(H,P,A),
	find_all_clauses(P/A,Cls,Clauses,RemainingCls).

% =====================================================

% MODE: read_property(-Phi, +Prog)
% TYPE: read_property(predicate,list(def))
% SEMANTICS: Phi is read from the user.
% Phi is a nullary predicate defined by a single clause in the current program Prog.

read_property(Phi,Prog,Property) :-
        write('property to be proved:'), nl,
        write('enter a nullary predicate defined by a single clause in the current program'), nl,
        read(Phi),
        (check_property(Phi,Prog,Property) ->
		(assert(type(Phi,0,[])), nl);
		(nl, write('error: not a nullary predicate defined by a single clause in the current program'), nl, fail)).


% MODE:
% TYPE: t_sort_arrays(list(def),list(def))
% SEMANTICS:
%
% t_sort(+Cs,?S) :-  S is a list of clauses derived by reordering
% the clauses in Cs s.t. if C1 precedes C2 in S then
% no literal in the body of C1 depends on the head of C2
t_sort_arrays([],[]).
t_sort_arrays([C|Cs],[M|S]) :-
        del_minimal_arrays([C|Cs],M,Cs1),
        t_sort_arrays(Cs1,S).

del_minimal_arrays([M],M,[]).
del_minimal_arrays([C1,C2|Cs],M,Cs1) :-
        dep_arrays(C1,[C2|Cs]), !,
        Cs1=[C1|Cs2],
        del_minimal_arrays([C2|Cs],M,Cs2).
del_minimal_arrays([C1,C2|Cs],M,Cs1) :-
        M=C1,
        Cs1=[C2|Cs].

% dep(+C,+Cs) :- a predicate in the body of C depends on
% the head of a clause in Cs.
dep_arrays((_,bd(_,B)),Cs) :-
        member(L,B),
        pred(L,P/A),
        member((H1,bd(_,_)),Cs),
        pred(H1,P/A).

% ltt(+Stat,+Aux_Sts1,?Defs,?Aux_Sts2)
% ltt(Stat,[],Defs,_) :- Defs is the list of clauses derived by
% applying the LT transformation starting from the statement Stat.
% Distinct clauses for different occurrences of the same subformula are
% not introduced.
% The second and fourth arguments are used to keep track
% of all statements introduced during the LT transformation.

ltt_arrays((_,(\+ true)), Aux_Sts, [], Aux_Sts).
ltt_arrays((H,L), Aux_Sts1, Sts, Aux_Sts2) :-
        conj_of_literals(L),
        !,
        Sts=[(H,L)],
        Aux_Sts1=Aux_Sts2.
ltt_arrays((H,(\+ (\+ F))), Aux_Sts1, Sts, Aux_Sts2) :-
        ltt_arrays((H,F), Aux_Sts1, Sts, Aux_Sts2).
% optimize subformulas
% ltt_arrays((H,(\+ (F1 and F2))), Aux_Sts1, Sts, Aux_Sts2):-
%	member((A,Phi), Aux_Sts1), variant(Phi, (F1 and F2)),
%	!,
%	functor(A,Pred,_),
%	fv((F1 and F2),Vars),
%	Newp=..[Pred|Vars],
%	copy_term((H,(\+ Newp)),(H1,(\+ Newp1))),
%	Sts=[(H1,(\+ Newp1))],
%	Aux_Sts2=Aux_Sts1.
ltt_arrays((H,(\+ (F1 and F2))), Aux_Sts1, Sts, Aux_Sts2):-
        new_statement(Newp, (F1 and F2)),
        ltt_arrays((Newp, (F1 and F2)), [(Newp, (F1 and F2))|Aux_Sts1], Sts12, Aux_Sts2),
	%%% copy_term((H,(\+ Newp)),(H1,(\+ Newp1))),
        Sts=[(H,(\+ Newp)) | Sts12].
% optimize subformulas
% ltt_arrays((H, (\+ e(X,F))), Aux_Sts1, Sts, Aux_Sts2) :-
%	member((A,Phi), Aux_Sts1), variant(Phi, e(X,F)),
%	!,
%	functor(A,Pred,_),
%	fv(e(X,F),Vars),
%	Newp=..[Pred|Vars],
%	copy_term((H,(\+ Newp)),(H1,(\+ Newp1))),
%	Sts=[(H1,(\+ Newp1))],
%	Aux_Sts2=Aux_Sts1.
ltt_arrays((H, (\+ e(X,F))), Aux_Sts1, Sts, Aux_Sts2) :-
        new_statement(Newp, e(X,F)),
        ltt_arrays((Newp,F), [(Newp,e(X,F))|Aux_Sts1], Sts12,Aux_Sts2),
	%%% copy_term((H,(\+ Newp)),(H1,(\+ Newp1))),
        Sts=[(H,(\+ Newp)) | Sts12].
ltt_arrays((H, (F1 and F2)), Aux_Sts1, [(H, (G1 and G2)) | NewSts], Aux_Sts3) :-
        ltt_arrays((H,F1), Aux_Sts1, [(H,G1) | NewSts1], Aux_Sts2),
        ltt_arrays((H,F2), Aux_Sts2, [(H,G2) | NewSts2], Aux_Sts3),
	%%% copy_term((H,G1,G2),(NewH,NewG1,NewG2)),
        append(NewSts1,NewSts2,NewSts).
ltt_arrays((H, e(_,F)), Aux_Sts1, Sts, Aux_Sts2) :-
	% drop_quantifier(e(X,F), NewF),
	% new_statement(Newp, NewF),
        % functor(Newp,Pred,_),
	% fv((H and F),V),
	% Newp1=..[Pred|V],
	ltt_arrays((H,F), Aux_Sts1, Sts, Aux_Sts2).


conj_of_literals(L) :-
	(lit(L) ; is_a_constraint(L) ; L==true).
conj_of_literals(L and M) :-
	conj_of_literals(L),
	conj_of_literals(M).

% ltt((H, e(X,F)), Aux_Sts1, Sts, Aux_Sts2) :-
%         drop_quantifier(e(X,F), NewF),
%         ltt((H, NewF), Aux_Sts1, Sts, Aux_Sts2).

% =====================================================

no_variables_in_bodies((_,[])).
no_variables_in_bodies((P,[(H,bd(_,B))|Cls])) :-
	term_variables(B,[]),
	term_variables(H,[]),
	no_variables_in_bodies((P,Cls)).

% =====================================================

% prop :- \+ p
% p :- t
% p :- s
% p :- q
% q :- n
% n :- n
% n :- t
% s :- r
% r.
% t :- m
% m :- n
%
% [(prop/0,[(prop,bd([],[\+ p]))]),(p/0,[(p,bd([],[t])),(p,bd([],[s])),(p,bd([],[q]))]),(q/0,[(q,bd([],[n]))]),(n/0,[(n,bd([],[n])),(n,bd([],[t]))]),(s/0,[(s,bd([],[r]))]),(r/0,[(r,bd([],[]))]),(t/0,[(t,bd([],[m]))]),(m/0,[(m,bd([],[n]))])]
%
% {p,q,n,s,r,t,m}
% {p,q,n,r,t,m}
% {p,q,n,t,m}
% {q,n,t,m} ok!
%
% p :- t
% p :- s
% p :- q
% s :- r
% r.
%
% [(prop/0,[(prop,bd([],[\+ p]))]),(p/0,[(p,bd([],[t])),(p,bd([],[s])),(p,bd([],[q]))]),(s/0,[(s,bd([],[r]))]),(r/0,[(r,bd([],[]))])]
%
% after some unfolding...
%
% p.
%

% MODE:
% TYPE: 
% SEMANTICS: we ASSUME (by correctness of the strategy) that
%            if the strategy terminates, then Def-star(Phi)
%            is propositionaL
inspect(Pout,Phi,Answer) :-
	def_star(Phi/0,Pout,DefST),
	(derivation_trace_on -> show_all_defs(DefST) ; true),
	decide_truth_value(Phi/0,DefST,Answer).

decide_truth_value(P/A,DefST,Answer) :-
	useless_cls_elim(P/A,DefST,NewDefST),
	def(P/A,DefST,Def),
	append([Def],NewDefST,NewDefST2),
	% write('After useless elimination:'), nl,
	% show_all_defs(NewDefST2),
	reverse(NewDefST2,RevNewDefST2),
	unfold_star_prop(RevNewDefST2,FinalDefST),
	(has_no_def(P/A,FinalDefST) -> Answer=false ; Answer=true).
	
useless_cls_elim(P/A,DefST,NewDefST) :-
	delete_def(P/A,DefST,DefST2),
	detect_useless(DefST2,UselessD),	
	delete_useless_defs(UselessD,DefST2,NewDefST).

detect_useless(Def,UselessDef) :-
	remove_those_not_useless(Def,Def,NewDef),
	(NewDef==Def ->
		(UselessDef=Def) ;
		detect_useless(NewDef,UselessDef)).

remove_those_not_useless([],_,[]).
remove_those_not_useless([D|Ds],Prog,NewDs) :-
	has_clause_not_useless(D,Prog),
	remove_those_not_useless(Ds,Prog,NewDs).
remove_those_not_useless([D|Ds],Prog,[D|NewDs]) :-
	remove_those_not_useless(Ds,Prog,NewDs).

%  has_clause_not_useless((_,[]),_).
has_clause_not_useless((P,[Cl|Cls]),Prog) :-
	((Cl = (_,bd([],[]))) ;
	  (Cl = (_,bd([],B)),
	   member(Atom,B),
	   functor(Atom,Predicate,Arity),
	   (has_no_def(Predicate/Arity,Prog);
	    def(Predicate/Arity,Prog,(_,[])))
	   % (functor(L,\+,_) -> (L=..[\+,Atom]); (L=Atom)),
	   % functor(Atom,P,A),
	   % (def(P/A,Prog,[]);def(P/A,Prog,(_,[]))) 
	  )) ->
	  	true ;
		has_clause_not_useless((P,Cls),Prog).
% has_clause_not_useless((P,[_|Cls]),Prog) :-
% 	Cl \== (_,bd([],[])),
%	has_clause_not_useless((P,Cls),Prog).

has_no_def(Predicate/Arity,Prog) :-
	\+ def(Predicate/Arity,Prog,_).

delete_useless_defs([],Prog,Prog).
delete_useless_defs([(P,_)|UDs],Prog,NewProg) :-
	delete_def(P,Prog,Prog2),
	delete_useless_defs(UDs,Prog2,NewProg).

% ------------------------------------------

unfold_star_prop(Prog,ProgFin) :-
	unfold_all_prop(Prog,Prog,NewProg),
	(NewProg==Prog ->
		(ProgFin=Prog) ;
		unfold_star_prop(NewProg,ProgFin)).

unfold_all_prop([],_,[]).
unfold_all_prop([Def|Defs],Prog,[Def|NewDefs]) :-
	all_facts(Def),
	unfold_all_prop(Defs,Prog,NewDefs).
unfold_all_prop([Def|Defs],Prog,NewDefs) :-
	(unfold_def_prop(Def,Prog,NewDef) ->
		append(NewDef,Defs,NewDefs) ;
		unfold_all_prop(Defs,Prog,NewDefs)
	).

% MODE:
% TYPE:
% SEMANTICS: we ASSUME only ONE predicate in the body
%            predicate-clauses
% unfold_def_prop((_,[]),_,[]).
unfold_def_prop((P,[(H,bd([],[L]))|_]),Prog,[(P,[(H,bd([],[]))])]) :-
	functor(L,\+,_),
	L=..[\+,Atom],
	(has_no_def(Atom/0,Prog);def(Atom/0,Prog,(Atom/0,[]))).
unfold_def_prop((P,[_|Cls]),Prog,Def) :-
	unfold_def_prop((P,Cls),Prog,Def).
unfold_def_prop((P,[(H,bd([],[L]))|_]),Prog,[(P,[(H,bd([],[]))])]) :-
	\+ functor(L,\+,_),
	def(L/0,Prog,Def),
	has_facts(Def).
% unfold_def_prop((P,[_|Cls]),Prog,Def) :-
%	\+ functor(L,\+,_),
%	has_no_def(L/0,Prog),
%	has_facts(Def).

%	unfold_def_prop((P,Cls),Prog,Def).

all_facts((_,[])).
all_facts((_,[Cl|Cls])) :-
	fact(Cl),
	all_facts((_,Cls)).

has_fact((_,[Cl|_])) :- fact(Cl).
has_fact((_,[_|Cls])) :- has_fact((_,Cls)).

fact((_,bd([],[]))).

% MODE: mydelete(+,-,-)
mydelete([X|Y],Z,Y) :- X=Z.
mydelete([X|Y],Z,[X|R]) :- mydelete(Y,Z,R).

