% input_output for Windows

:- use_module(library(system), [ file_exists/1,
                                 shell/1 ]).
:- use_module(library(lists), [ nth/3 ]).

/*-------
input: Shows the list of available programs in subdirectory "Progs/" and prompts the user
to select one which becomes the initial program of a new transformation process.
This predicate is used at top level by start.
---------*/
input  :- 
  nltab, write('Choose program:'), nl, my_directory_files('Progs',LL),
  !, show_available_programs(1,LL,LL,Num_P), !,
  ask_between('Initial program', 1, Num_P, 'no selection', Answer),
  (Answer = 'no selection' -> fail ; (nth(Answer,LL,Prog),
  my_concat_atom(['Progs/',Prog],Path), input(Path), assert(file_input(Path)))).
input :- msg_no_available_progs.

input(File) :- see(File), read_list(1), seen, show_initial_program.
input_determ(File) :-
      (strat(generalize_head_output_in_continuation) ->
	  (assert(mode('@',1,[-])), input_cont(File)) ; input(File)).
input_cont(File) :-
      see(File), read_list_cont(1), seen, show_initial_program.

show_available_programs(N,[File1],_L_progs,N) :- !,
	nl, write(N), write(' ) '), write(File1), nl, repeat, nl.
%	(yesno('Do you want to look at a program before selection',n) -> 
%	(ask_between('Which one ?', 1, N, 'none', Answer), nl,
%	(Answer = 'none' -> 
%	true ; 
%	(nth(Answer,L_progs,Prog), 
%	 name(Prog,ProgAscii),
%	 append("more Progs/",ProgAscii, S_Unix_command),
%	 name(Unix_Command,S_Unix_command), 
%	 shell(Unix_Command), fail ))) ; true).

show_available_programs(N,[File1|L],L_progs,NF) :-
	nl, write(N), write(' ) '), write(File1), 
  	N2 is N+1,show_available_programs(N2,L,L_progs,NF).

/*---------------
select_theory(N,[T],L_theories) is used at top level by select_law_in_theory.
----------------*/

select_theory(N,[T],L_theories) :- !,
     nl, write(N), write(' ) '), write(T), nl, nl,
     ask_between('Theory number', 1, N, 'no selection', Answer),
     (Answer = 'no selection' -> fail ;
	 (   nth(Answer,L_theories,Theory),
	     my_concat_atom(['Theories/',Theory], File_Theory),
	     read_laws(File_Theory,Laws),
	     apply_add_laws(Laws))).					    
select_theory(N,[T1|L],L_theories) :-
      nl, write(N), write(' ) '), write(T1), N2 is N +1, select_theory(N2,L,L_theories).


apply_show_initial_program:-
 nl, write('Initial program:'), nl, nl, clause3(N,C,prog), show_numb_clause((N,C)), nl, fail.
apply_show_initial_program.


/*-----------
apply_show_history: show history
------------*/
apply_show_history(N) :-
    ancestors(N,NPs), append(NPs,[N],NPps), nl, nl, show_information(NPps).
show_information([P,P1|Ps]) :-
     (P=1 -> write_inf(1),nl; true), label_arc(P1,LA), write_inf(P,LA,P1),
     nl, show_information([P1|Ps]).
show_information([P]) :- P=1 -> write_inf(1), nl, nl;true.

/*-----------
apply_show_definition: show available definitions (in the ancestor nodes of current program)
------------*/
apply_show_definition :-
   current_program(P), 
   setof(D, Pr^(definition_clauses(Pr,D), check_ancestor(Pr,P)), L),
   append(L,L2), show_clauses_ok(L2).

/*-----------
apply_show_law: show available laws
------------*/
apply_show_law:- nl, law(N,L),	show_law_number((N,L)),	fail.
apply_show_law.

/*-----------
apply_save_history(N): save history of program N in subdirectory Histories/
in a file named hist_P_N where P is the name of the initial program for
the current transformation process.
------------*/

apply_save_history(N) :- 
   file_input(File), name(File,L), append("Progs/",Name_L,L), append("hist_",Name_L,Hist_L),
   append(Hist_L,"_",L_), name(N,L2), append(L_,L2,ProgN), append("Histories/",ProgN,History_F),
   msg_ask_save_history(History_F,File2), name(Name_File2,File2),	
   msg_save_history(L,Name_File2), ancestors(N,Ns), append(Ns,[N],Ms),
   tell(Name_File2), show_short_commented_programs(Ms), told, nl.

/*--------
show_available_final_programs used at top level by dir_f.
--------*/
show_available_final_programs(N,[File1],L_progs,N) :- !,
     nl, write(N), write(' ) '), write(File1), nl, repeat, nl, 
     ask_between('Which program do you want to look at?', 1, N, 'none', Answer), nl, 
          (Answer = 'none' -> true ;
     (nth(Answer,L_progs,Prog), name(Prog,Ascii),
     append("more Final_Progs/",Ascii, S_Unix),
     name(Unix,S_Unix), shell(Unix), fail )).
show_available_final_programs(N,[File1|L],L_progs,NF) :-
     nl, write(N), write(' ) '), write(File1), N2 is N +1,
     show_available_final_programs(N2,L,L_progs,NF).

/*--------
show_available_sessions used at top level by dir_s.
--------*/
show_available_sessions(N,[File1],N) :- !, nl, write(N), write(' ) '), write(File1).
show_available_sessions(N,[File1|L],N3) :-
   nl, write(N), write(' ) '), write(File1), N2 is N +1, show_available_sessions(N2,L,N3).

/*--------
show_available_theories used at top level by dir_t.
--------*/
show_available_theories(N,[File1],L_progs,N) :- !,
   nl, write(N), write(' ) '), write(File1), nl, repeat, nl, 
   (yesno('Do you want to look at a theory ',n) -> 
   (ask_between('Which one ?', 1, N, 'none', Answer), nl,
   (Answer = 'none' -> true ;
   (nth(Answer,L_progs,Prog), show_theory2_ok(Prog), fail))) ; true).
show_available_theories(N,[File1|L],L_progs,NF) :-
   nl, write(N), write(' ) '), write(File1), N2 is N +1,
   show_available_theories(N2,L,L_progs,NF).

show_theory2_ok(Prog) :-
        name(Prog,ProgAscii), append("more Theories/",ProgAscii, S_Unix_command),
	name(Unix_Command,S_Unix_command), shell(Unix_Command).

/*
select_theory(N,[T],L_theories) :- !,
  nl, write(N), write(' ) '), write(T), nl, nl,
  ask_between('Theory number', 1, N, 'no selection', Answer),
  (Answer = 'no selection' -> fail ; (nth(Answer,L_theories,Theory),
  my_concat_atom(['Theories/',Theory], File_Theory), see(File_Theory), read_laws,  seen)).
select_theory(N,[T1|L],L_theories) :-
   nl, write(N), write(' ) '), write(T1), N2 is N +1, select_theory(N2,L,L_theories).
*/

/*-------
input_session: Shows the list of available sessions in subdirectory "Sessions/" and prompts
the user to select the which must become the new transformation tree.
This predicate is used at top level by load_session.
---------*/
input_session  :-
   nltab, write('Available sessions:'),nl, my_directory_files('Sessions',LL),
   non_empty_directory(LL,L), !, show_available_sessions(1,L,Num_P),!,
   ask_between('Session number', 1, Num_P, 'no selection',Ans),(Ans = 'no selection' -> fail
  ; (nth(Ans,L,Prog), my_concat_atom(['Sessions/',Prog],Path),input_session(Path))).
input_session  :- msg_no_available_sessions.

input_session(File) :- see(File), read_session, seen.
read_session :- 	
	read(C), C\==end_of_file, !, (functor(C, library_directory, _) -> true ;
	(functor(C, file_search_path, _) -> true ; (functor(C, file_search_path, _) -> true ;
        assert(C)))), read_session.
read_session.

/*----------
apply_add_law_in_theory(T): add a law to file T. The law is typed by the user.
He may add more than one law. This predicate is used at top level by add_law_in_theory(T).
------------*/
apply_add_law_in_theory(File_Theory) :-
   (yesno('Do you want to see the already existing laws ',n) -> 
   show_theory_ok(File_Theory) ; true), nl, write('Replacement law to add: '),
   repeat, read(Law), nl,
   (check_law(Law) -> concat_law(File_Theory,Law) ; msg_syntax_law),
   (yesno('Another law to add',n) -> nl, nl, write(': '), fail ; true),
   nltab, write('New theory saved in file: '),	write(File_Theory), nl .

/*----------
apply_delete_law_from_theory(T): delete a law from file T. The number of the law is chosen
by the user via a menu. He may delete more than one law.
This predicate is used at top level by delete_law_from_theory(T).
------------*/

apply_delete_law_from_theory(Theory,File_Theory) :-
   get_file(File_Theory,L), extract_file_laws(L,Laws,_), show_laws(0,Laws,Max), repeat, nl,
   ask_between('enter law to delete', 1, Max, 'no selection',Answer2), (Answer2 = 'no selection' ->
  true ; (nth(Answer2,Laws,Law), delete_law_from_theory_file(Law,Theory))),
  (yesno('Another law to delete ',n) -> fail ; true).

show_laws(0,[],_) :- !, msg_no_law_in_theory.
show_laws(N,[],N) :- nl.
show_laws(N,[Law|Laws],Max) :- N2 is N+1, show_law_num(N2,Law), show_laws(N2,Laws,Max).

extract_file_laws([],[],[]).
extract_file_laws([Line|Lines], [Law|Laws],[Law2|Laws2]):-
       law_line(Law,Law2,Line),!, extract_file_laws(Lines, Laws,Laws2).
extract_file_laws([_|Lines], Laws,Laws2):- extract_file_laws(Lines, Laws,Laws2).
delete_law_from_theory_file(Law,File_Theory) :-
	my_concat_atom(['old_', File_Theory], Old_File),
	my_concat_atom(['mv Theories/', File_Theory], B_S_Unix_command),
	my_concat_atom([B_S_Unix_command, ' Theories/'], M_S_Unix_command),
	my_concat_atom([M_S_Unix_command, Old_File], S_Unix_command),
	shell(S_Unix_command), name(Quote, [39]),
	my_concat_atom(['sed s/', Quote, Law, Quote],X), my_concat_atom([X,'//g Theories/'],Y),
	my_concat_atom([Y, Old_File], Z), my_concat_atom([Z, ' > Theories/'], T),
	my_concat_atom([T, File_Theory], U), shell(U), msg_save_file_laws(File_Theory).


/*----------
select_law_in_theory(N,T,T2) is used at top level by seslect_theory.
------------*/
select_law_in_theory(_,[],_) :- !, nl,
   write('Empty file of theories (stored in sub-directory Theories/) '), nl. 
select_law_in_theory(N,[T],L_theories) :- !,
   nl, write(N), write(' ) '), write(T), nl,
   ask_between('Theory number', 1, N, 'no selection', Answer),
   (Answer = 'no selection' -> fail ; (nth(Answer,L_theories,Theory),
   my_concat_atom(['Theories/',Theory], File_Theory), 
   read_laws(File_Theory,LLaws), show_laws(0,LLaws,Max), nl, repeat, nl,
   ask_between('Law number', 1, Max, 'no selection', Answer2),
   (Answer2 = 'no selection' -> true ; (nth(Answer2,LLaws,Law), apply_add_law(Law))),
   (yesno('Another law to select ',n) -> fail ; true))).
select_law_in_theory(N,[T1|L],L_theories) :-
  nl, write(N), write(' ) '), write(T1), N2 is N +1, select_law_in_theory(N2,L,L_theories).

show_theory_ok(Prog) :-
   name(Prog,ProgAscii), append("more ",ProgAscii, S_Unix_command),
   name(Unix_Command,S_Unix_command), shell(Unix_Command).

read_laws(T,Laws) :- see(T), read_laws2([],Laws), seen.
read_laws2(L1,L) :- read(C), C\==end_of_file, !, (C=(_ -> _) ->
        (append(L1,[C],LL1), read_laws2(LL1,L)) ; read_laws2(L1,L)).
read_laws2(L,L).

apply_show_program(Clause_Numbers) :-
    numb_fetch_clauses(Clause_Numbers,Numbered_Clauses), show_numb_clauses(Numbered_Clauses).

apply_show_short_program(Clause_Numbers) :-
    label_node(1,InitialClauses), subtract(Clause_Numbers, InitialClauses, Clauses),
    numb_fetch_clauses(Clauses,Numbered_Clauses), show_numb_clauses(Numbered_Clauses).

/*
apply_show_initial_program:-
     nl, write('Initial program:'), nl, nl, clause3(N,C,prog),
     show_numb_clause((N,C)), nl, fail.
apply_show_initial_program.
*/

apply_show_current_program(NCs) :- numb_fetch_clauses(NCs,Cls), show_numb_clauses(Cls).

apply_show_clauses :- clause_number_to_clause(N,C), show_numb_clause((N,C)), nl, fail.
apply_show_clauses.

show_numb_clauses([NC|NCs]) :- show_numb_clause(NC), nl, show_numb_clauses(NCs).
show_numb_clauses([]).
show_numb_clause(NC) :-   
    NC=(Num,C), number_clause(Num), !,  write(Num), write('. '),
    print_length(write(Num),LNum), InitPos is LNum + 2,
    show_clause(InitPos, C).
show_numb_clause(C) :-	show_clause(0,C).



show_clause(InitPos, (H,B)):-
    copy_term((H,B),(H1,B1)),
    numbervars((H1,B1),0,_),
    writeq(H1), (B1=[] -> write('.');
		   (print_length(write(H1),LH1),
		       LH2 is LH1 + 4 + InitPos,
		       write(' :- '), show_body(B1,LH2))).

/*--------------
apply_show_only_clauses is used at top level by save current program.
--------------*/

apply_show_only_clauses(NCs) :- numb_fetch_clauses(NCs,Cls), show_not_numb_clauses(Cls).

show_not_numb_clause(NC) :-  NC=(Num,C), number_clause(Num), !, show_clause(0,C).
show_not_numb_clause(C) :- show_clause(0,C).
show_not_numb_clauses([]).
show_not_numb_clauses([C|Clauses]) :-
	show_not_numb_clause(C), nl, show_not_numb_clauses(Clauses).

%show_body(Body,PosBody) :- show_body_lines(Body,PosBody,PosBody,70).

show_body(L,LH) :-	show_body_lines(L,10,LH,80).
show_body_lines([H], PosInitBody,PosLine, Max) :- !, 
	print_length(write(H),Length), NewPos is PosLine + Length + 1,
	(NewPos > Max -> nl, tab(PosInitBody); true),writeq(H), write('.').
show_body_lines([H|T],PosInitBody,Pos,Max) :- 
	print_length(write(H),Length), NewPos is Pos + Length + 2,
	(NewPos > Max ->
	    (	nl, tab(PosInitBody), Pos2 is PosInitBody +  Length + 2)
	;   Pos2=NewPos),   writeq(H), write(', '),
	show_body_lines(T, PosInitBody,Pos2,Max).

show_goal(L) :-	show_goal_lines(L,0,0,80).
show_goal_lines([H], PosInitBody,PosLine, Max) :- !, 
	print_length(write(H),Length), NewPos is PosLine + Length + 1,
	(NewPos > Max -> nl, tab(PosInitBody); true), writeq(H).
show_goal_lines([H|T],PosInitBody,Pos,Max) :- 
	print_length(write(H),Length), NewPos is Pos + Length + 2,
	(NewPos > Max -> nl, tab(PosInitBody), Pos2=PosInitBody; Pos2=NewPos),
        writeq(H), write(', '),	show_goal_lines(T, PosInitBody,Pos2,Max).

show_law_num(N,L) :- write(N), write('. '), numbervars(L,0,_), write(L), nl.
show_law_number((N,(L,R))) :-
   numbervars((N,(L,R)),0,_), write(N), write('. '), show_lhs(L),
   write(' -> '), show_rhs(R), nl.
show_lhs([]).
show_lhs([E]) :- !, write(E).
show_lhs([E|L]) :- write(E), write(', '), show_lhs(L).
show_rhs([]) :- write('.').
show_rhs([E]) :- !, write(E), write('.').
show_rhs([E|L]) :- write(E), write(', '), show_rhs(L).

apply_save_current_theory :-
	repeat, nl, write('Name of the new theory :  '),
	get_line(S_File), (S_File =[] -> msg_no_selection ;
	(my_last(S_File,Head_S_File,46), append("Theories/",Head_S_File,File2),
	name(Name_File2,File2),	(file_exists(Name_File2) ->
	msg_choose_another_filename(Name_File2), fail ; true),
        msg_save_current_theory(Name_File2), tell(Name_File2), show_notnumbered_laws, told)).
	
show_notnumbered_laws :- nl, law(_,L), show_law(L), fail.
show_notnumbered_laws.
show_law(((L,R))) :- numbervars((L,R),0,_), show_lhs(L), write(' -> '), show_rhs(R), nl.
show_short_commented_programs([]).
show_short_commented_programs([N]) :-
      (N=1 -> write_inf(1), nl; true), (label_node(N,_) -> (show_short_program(N), nl,
      check_assert_comment_history(N); write(' does not exist.'))).
show_short_commented_programs([N,N1|Ns]) :- 
        (N=1 -> write_inf(1), nl ; true), (label_node(N,_) -> (show_short_program(N), nl,
      check_assert_comment_history(N), label_arc(N1,LA), write_inf(N,LA,N1), nl);
	write(' does not exist.')), nl, show_short_commented_programs([N1|Ns]).
check_assert_comment_history(P) :-
    comment_history(P,Comment), !, write_comment_history(Comment).
check_assert_comment_history(_).

write_comment_history([]) :- nl.
write_comment_history([X|Xs]) :- nl, name(T,X), write(T), write_comment_history(Xs).

apply_insert_comment_history(P,L) :-
   get_line(X), (X =[] -> assert_comment_history(P,L) ; apply_insert_comment_history(P,[X|L])).

show_list_h([]) :- write('.'), nl.
show_list_h([E|L]) :- tab(2), write(E), show_list_h(L). 
show_list_v([]) .
show_list_v([E|L]) :- nltab, write(E), show_list_v(L). 

nltab :- nl, tab(5).
show_packets([P]) :- !, new_clauses(P,NewClauses), show_numb_clauses(NewClauses).

show_packets([P|Ps]) :- 
  new_clauses(P,NewClauses), show_numb_clauses(NewClauses), nl, show_packets(Ps).

give_modes([],_).
give_modes([X|L],1) :- ask_oneof('First argument ', ['+','-'], '+', X), give_modes(L,2).
give_modes([X|L],2) :- ask_oneof('Second argument ', ['+','-'], '+', X), give_modes(L,3).
give_modes([X|L],3) :- ask_oneof('Third argument ', ['+','-'], '+', X), give_modes(L,4).
give_modes([X|L],N) :- N > 3, ask_oneof('Next argument ', ['+','-'], '+', X), give_modes(L,N).

listing_transf_tree :-
    listing(node), listing(number_clauses), listing(clause3),
    listing(current_prog), listing(file_input), listing(max_clause),
    listing(max_law), listing(maxnode), listing(number_clauses),
    listing(law), listing(new_file_input).


% Added by Maurizio 3 Oct 2002
% dtrace(on) shows the derivation trace during determ, prove, synth

:- dynamic derivation_trace_on/0.

%:- assert(derivation_trace_on).
dtrace(on) :- assert(derivation_trace_on).
dtrace(off) :- retractall(derivation_trace_on).


