% lap.pl 
%	
% This file contains the predicates which are used by the "lap"
% Loop Absorbtion strategy. It is described in
%      Proietti, M. and Pettorossi, A.:
%      Construction of Efficient Logic Programs by Loop Absorption and
%      Generalization. Proceedings Meta'90 (M. Bruynooghe, ed.), Leuven,
%      Belgium (1990), p. 57-81.
%      and in
%      Proietti, M. and Pettorossi, A.:
%      The Loop Absorption and the Generalization Strategies for the
%      the Development of Logic Programs and Partial Deduction.
%      Journal of Logic Programming, 16(1,2):123-161, 1993.

:- use_module(library(lists), [ append/3,
                                reverse/2 ]).
:- use_module(library(terms), [ subsumes_chk/2  ]).

lap :- check_exists_transf_tree,
       statistics(runtime,[To|_]), load(Prog1,Cl,Upred,NCl),
       mark_predicates(Upred,Pre),
       write('Enter horizontal bound: '),read(H_bound),
       write('Enter  vertical  bound: '),read(V_bound),nl,nl,
       assert(bound(H_bound,V_bound)), /* bound(Horizontal,Vertical) */
       Cl=[Head|Bd], Head=..[Nam|L], find_filename(Nam,Filename),
       tell(Filename), write_marked_proc(Prog1,Pre,Prog),
       faq([L|Bd],Prog,[],[],Defs,Pr,Pre),
       retractall(bound(_,_)), make_proc(Nam,Pr,Pr1),
       exclude_subsume(Pr1,Pr2), write_proc(Pr2), write('/*'),nl,
       scrivi_def(Defs), close(Filename),
       set_current_output(Filename,NCl), 
       format('Horiz_bound was ~d,
                Vertical_bound was ~d ~n',[H_bound,V_bound]), nl,nl,
       statistics(runtime,[Tn|_]), Tmm is Tn - To,
       format('it all took ~3d sec. ~n', [Tmm]),
       message_final_prog_lap(Filename), p.

find_filename(Nam,Filename) :-
 name(Nam,L1), name('Final_Progs/lap_',L2), append(L2,L1,L), name(Filename,L).

write_clause([H]):- write(H),write('.'),nl.
write_clause([H|C]):- write(H),write(' :-'),nl,write_goals(C).
write_goals([]):-nl.
write_goals([X]):-write('    '),write(X),write('.'),nl.
write_goals([X|L]):-write('    '),write(X),write(','),nl,write_goals(L).
write_proc([]) :- nl.
write_proc([Cl1|Rcl]) :-
  copy_term(Cl1,Cl), numbervars(Cl,0,_), write_clause(Cl), write_proc(Rcl).
set_current_output(Filename,NCl) :- 
   see(Filename), max_clause(MC), maxnode(N1), SN is N1+1, Pn is MC+1,
   read_list1(_,Pn,L,SN), seen, current_prog(N,P),
   expand(N,(NCl,loopabs),L,SN), update_current_prog((N,P),(SN,L)).
read_list1([C|Cs],N,L,SN) :- 
   read(C), C\==end_of_file, !, L=[N|K], clau(C,N,SN),  M is N+1,
   read_list1(Cs,M,K,SN).
read_list1([_|_],N,K,_) :- 
   M is N-1, max_clause(MC), retract(max_clause(MC)), 
   assert(max_clause(M)), K=[], assert(number_clauses(M)).
clau(C,N,NN):- C=..[:-, Head, Body], conj_to_list((Body),Body1), 
   assert(clause3(N,(Head,Body1),NN)).
clau(C,N,NN):- assert(clause3(N,(C,[]),NN)).

/* write_marked_proc
   1st arg: the list of all the procedures				input
   2nd arg: the list of the relevant predicates				input
   3rd arg: the list of all the procedures of relevant predicates	output
*/
write_marked_proc([],_,[]).
write_marked_proc([Proc1|Rest],Pre,[Proc1|Rest1]) :-
   Proc1 =[[Head|_]|_], functor(Head,Na,Ar),
   membert([Na,Ar],Pre), write_marked_proc(Rest,Pre,Rest1).
write_marked_proc([Proc|Rest],Pre,Rest1) :-
   write_proc(Proc), write_marked_proc(Rest,Pre,Rest1).
membert(H,[K|_]):-H==K.
membert(H,[K|T]):- H\==K, membert(H,T).
scrivi_def([]) :- write('*/').
scrivi_def([D1|Restdef]) :- numbervars(D1,0,_),write_clause(D1), 
                            scrivi_def(Restdef).

/*load
  1st arg: the list of the procedures, i.e. the program			output
  2nd arg: the clause							output
  3rd arg: the list of the defined predicates				output
*/
load(Prog,Cl, Pre,Numb) :-  
    nl, nl, write('Current program'), nl, p,
    nl, ask_number('Enter the clause number ', 1, Numb),
    clause3(Numb,T,_), get_aioni(T,Cl), current_prog(_,L),
    load_clauses(L,L_cls,Numb), divide_in_procedures(L_cls,Prog,Pre),!.

load_clauses([Numb|P],C,Numb) :- load_clauses(P,C,Numb).
load_clauses([K|P],[C|S],Numb) :-   clause3(K,C1,_),
  				    get_aioni(C1,C), load_clauses(P,S,Numb).
%load_clauses([_],[],_).
load_clauses([],[],_).

get_aioni(Cl,C) :- Cl=..[',',Hd,Bd], (Bd ==[] -> C=[Hd] ; C=[Hd|Bd]).
divide_in_procedures([],[[]|[]],[_|[]]).
divide_in_procedures([Cl|R_cl],[[Cl|R_proc]|R_procs],[Pr1|R_pre]) :-
    Cl=[Head|_],functor(Head,N,Ar), [N,Ar]==Pr1,
    divide_in_procedures(R_cl,[R_proc|R_procs],[Pr1|R_pre]).
divide_in_procedures([Cl|R_cl],[[]|R_procs],[Pr1|R_pre]) :-
    Cl=[Head|_], functor(Head,N,Ar), nonvar(Pr1),[N,Ar] \== Pr1,
    divide_in_procedures([Cl|R_cl],R_procs,R_pre).
divide_in_procedures([Cl|R_cl],[[Cl|R_proc]|R_procs],[[N,Ar]|R_pre]) :-
    Cl=[Head|_],functor(Head,N,Ar),
    divide_in_procedures(R_cl,[R_proc|R_procs],[[N,Ar]|R_pre]).
mark_predicates(Unmarked,Marked):-
	clear_screen(5),
	write('LOOP ABSORPTION PROCEDURE'),nl,
	write('-------------------------'),nl,
        nl,nl, write('predicates defined :'),
	nl,write_blocks(Unmarked),nl,
	(yesno('Do you want to mark the predicates (y/n) ',n) -> 
	marking(Unmarked,Marked); Marked=Unmarked).
marking(Unmarked,Marked):- nl,nl,
	write('Mark predicates with: '),nl,
	write('return or r  if you want to mark that predicate,'),nl,
	write('		 i  if you dont want to'),nl,nl,
	mark_each_predicate(Unmarked,Marked),
	prompt(_ZZ,'|:  '),
	nl,nl,write('Remaining predicates'),nl,
	write_blocks(Marked),nl,nl,nl.   
mark_each_predicate([],[]).
mark_each_predicate([X|Un],Ma):-write('Mark :'), write_pr(X),
	ask_chars('	(r/i)',1,1,Mark), tra(MARK,Mark), 		
	cont_marking(MARK,X,Un,Ma).
tra(MARK,[0'r]):- MARK='r'.
tra(MARK,[0'i]):- MARK='i'.
tra(_,[_]):- write('insert (r/i) follow by RETURN').
cont_marking(i,_,Un,Ma):-mark_each_predicate(Un,Ma).
cont_marking(r,X,Un,[X|Ma]):-mark_each_predicate(Un,Ma).
cont_marking(_,X,Un,Ma):-write('MARK     '), write_pr(X), 
			ask_chars('	(r/i)',1,1,Mark), tra(MARK,Mark), 
			cont_marking(MARK,X,Un,Ma).
read_mark(r) :- ask_chars('(r/i)',1,1,R), read1(R).
read1('r').
read1('i').
read1(_) :- write('insert (r/i) follow by RETURN'), read_mark(_).
message_final_prog_lap(Filename) :-
	nl, nl, write('Final program save in '),
	write(Filename), nl, nl.
clear_screen(0).
clear_screen(X):-nl,X1 is X-1,clear_screen(X1).
write_blocks([]).
write_blocks([B|BB]):- write('	...'), write_pr(B), nl,write_blocks(BB).
write_pr(B):- B=[Pred,Ar], write(Pred), write('/'), write(Ar) .

/* faq
   1st arg: a pseudoclause, tclause, with as head a list of terms	input
   2nd arg: the program which consists in a list of procedures		input
   3rd arg: the list of the ancestors blocks				input
   4th arg: the list of the definitions already found			input
   5th arg: the list of the new definitions found			output
   6th arg: the list of the tclauses,which are leaves of 1st		output
   7th arg: the list of the relevant predicates				input
*/
faq([Head|Bd],Prog,Ant,Ddef,Add_def,Tclss,Pred) :- 
    gather_builtins(Bd,Pred,Ir,Re),
    (Re =[] -> (Tclss = [[Head|Bd]], Add_def = [],!) ;
        (db(Re,[],Bloxx),
         distri1(Bd,Bloxx,Ir,Irr1,Sclss),
         links_care([Head|Bd],Sclss,Tsclss,Links),
         unf_T_subclauses(Tsclss,Prog,Ant,Ddef,Add_def,Tclss1,Pred),
         bagof(Tcl,make_tclss([Head|Irr1],Links,Tclss1,Tcl),Tclss),!)).
/* unf_T-subclauses
   1st arg: the list of T_subclauses  to unfold				input
   2nd arg: the program, remember a list of procedures			input
   3rd arg: the ancestor tloxs						input
   4th arg: the list of the definitions we already have			input
   5th arg: the list of the definitions we add to the already		output
            existing ones
   6th arg: the list of lists of the tclauses, leaves of 1st		output
   7th arg: the list of the relevant predicates				input
*/
unf_T_subclauses([],_,_,_,[],[],_).
unf_T_subclauses([Tscl1|Tsclss],Prog,Ant,Ddef,Add_def,Tclss,Pred) :-
  get_tlock(Tscl1,Pred,Tl1,[Head|Bdd],Link),
  unf_one_tlo(Tl1,Prog,Ant,Ddef,Add_def1,Tcll,Pred),
  append(Ddef,Add_def1,Ddef1),!, garbage_collect,
  unf_T_subclauses(Tsclss,Prog,Ant,Ddef1,Add_def2,Tclss2,Pred),
  ck_new_def(Tl1,Add_def1,Tcll,[Head|Bdd],Link,Tcl1),
  Tclss=[Tcl1|Tclss2], append(Add_def1,Add_def2,Add_def).
/* unf_one_tlo
   1st arg: a tlock, remember a tlock is a list of variables 		input
            as first element, followed by the atoms
   2nd arg: the program							input
   3rd arg: the ancestors tlocks					input
   4th arg: the old definitions						input
   5th arg: the new definitions						output
   6th arg: the list of the tclauses, leaveas of 1st			output
   7th arg: the list of the relevant predicates				input
*/
unf_one_tlo(Tl,_,_,Ddef,[],[Rat],_) :-	cfr_ls_def(Tl,Ddef,Rat),!.
unf_one_tlo(Tl,_,Ant,_,[Rdef],[Rat],_) :- cfr_ls_anc(Tl,Ant,Rdef,Rat),!.
unf_one_tlo(Tl,Prog,Ant,Ddef,Add_def,Tclss,Pred) :- 
    controllo(Tl), gen_step(Tl,Gbl), copy_term(Tl,Tl1),!,
    faq(Gbl,Prog,[Tl1|Ant],Ddef,Add_def,Tclss,Pred),!.
unf_one_tlo(Tl,Prog,Ant,Ddef,Add_def,Tclss,Pred) :-
    selez(Tl,Before,Atom,After), find_proc(Atom,Prog,Proc),!,
    go_dn_brs(Proc,Before,Atom,After,Ant,Ddef,Add_def,Tclss,Pred,Prog),
    !, Tclss\==[].
/* go_dn_brs     (go down all the branches)
   1st arg: the procedure defining the predicate of the atom to		input
             unfold
   2nd arg: the part of the clause before the atom			input
   3rd arg: the selected  atom						input
   4th arg: the part of the clause after the atom 			input
   5th arg: the list of the ancestors 					input
   6th arg: the list of the definitions					input
   7th arg: the list of the new definitions 				output
   8th arg: the list of the tclauses leaves of the tlox			output
   9th arg: the list of the relevant predicates				input
  10th arg: the program							input
*/
go_dn_brs([],_,_,_,_,_,[],[],_,_).
go_dn_brs([C1|Prc],Bef,At,Aft,Ant,Ddef,Add_def,Tclss,Pred,Prog) :-
  go_dn_one(C1,Bef,At,Aft,Ant,Ddef,Add_def1,Tclss1,Pred,Prog),
  !, garbage_collect, append(Ddef,Add_def1,Ddef1),
  go_dn_brs(Prc,Bef,At,Aft,Ant,Ddef1,Add_def2,Tclss2,Pred,Prog),
  append(Add_def1,Add_def2,Add_def), append(Tclss1,Tclss2,Tclss).
/* go_dn_one  (go down one branch)
   1st arg: the program clause with which unfold the block		input
   2nd arg: the part of the block before the selected atom		input
   3rd arg: the selected atom						input
   4th arg: the part after the atom					input
   5th arg: the list of the ancestors					input
   6th arg: the list of the definitions					input
   7th arg: the list of the new definitions				output
   8th arg: the list of the tclauses, leaves				output
   8th arg: the list of the relevant predicates				input
   9th arg: the program							input
*/
go_dn_one(C,Bef,At,Aft,Ant,Ddef,Add_def,Tclss,Pred,Prog) :-
   copy_term(C,Cl), copy_term([Bef,At,Aft],[Bef1,At1,Aft1]),
   append(Bef,[At|Aft],Blo), copy_term(Blo,Bl),
   unify_head(Bef1,At1,Aft1,Cl,Unf),
   faq(Unf,Prog,[Bl|Ant],Ddef,Add_def,Tclss,Pred).
go_dn_one(_,_,_,_,_,_,[],[],_,_).   
/* selez
   1st arg: a block							input
   2nd arg: the part before the selected atom				output
   3rd arg: the selected atom						output
   4th arg: the part after the selected atom				output
*/
selez([Head|Bl],[Head|Before],Atom,After) :- special_list_depth(Bl,Max),
	                                      split1(Bl,Max,Before,Atom,After).
/* special_list_depth
   1st arg: a list of atoms						input
   2nd arg: the hight of the hightest atom				output
*/
special_list_depth([],0).
special_list_depth([Atom|L],Max) :- depth(Atom,D1),
			            special_list_depth(L,D2),
			            max(D1,D2,Max).
/* split
   1st arg: a list of atoms, such as the body of a block		input
   2nd arg: the heigth of the heightest atom				input
   3rd arg: the list of atoms before the right one			output
   4th arg: the atom selected						output
   5th arg: the list of atoms after the right 				output
*/
split1([Atom|Last],Max,[],Atom,Last) :- depth(Atom,M), Max=M.
split1([X|L],Max,[X|M],Atom,After) :- split1(L,Max,M,Atom,After).
depth(T,0) :- simple(T),!.
depth(T,M) :- functor(T,_,N),maxdepth(T,N,M).
maxdepth(T,1,M) :- arg(1,T,TX),depth(TX,M1),M is M1+1.
maxdepth(T,N,M) :- arg(N,T,TX), depth(TX,DX), DX1 is DX+1, N1 is N-1, 
                   maxdepth(T,N1,M1), max(DX1,M1,M).
max(B,S,B):-B>=S,!.
max(_,B,B).

 /* unify_head
   1st arg: the part before the selected atom				input
   2nd arg: the selected atom						input
   3rd arg: the part after the selected atom				input
   4th arg: the clause with which the unfold will be done		input
   5th arg: the new clause						output
*/
unify_head(Bef,At,Aft,[Head|Bd],Nb) :- 
   Head=At, !, append(Bef,Bd,S), append(S,Aft,Nb).
exclude_subsume([],[]).
exclude_subsume([[C]|L],[[C]|M]) :- exclude_subsume(L,M).
exclude_subsume([A|L],[X|N]) :- exist_subsume(A,X,L,M),
                                exclude_subsume(M,N).
exist_subsume(A,A,[],[]).
exist_subsume(A,X,[B|L],M) :- permute_furb(B,A,Bp),
                              subsumes_chk(A,Bp), exist_subsume(A,X,L,M).
exist_subsume(A,X,[B|L],M) :- permute_furb(A,B,Ap),
                              subsumes_chk(B,Ap), exist_subsume(B,X,L,M).
exist_subsume(A,X,[B|L],[B|M]):-exist_subsume(A,X,L,M).
/* find_proc
   1st arg: an atom							input
   2nd arg: the list of procedures that constitute the program		input
   3rd arg: the procedure that defines the predicate of the atom	ourput
*/
find_proc(Atom,[Proc|_], Proc) :- functor(Atom,F,Ar),
                                  Proc=[Cl|_], Cl=[At|_], functor(At,F,Ar).
find_proc(Atom,[_|Prog],Proc) :- find_proc(Atom,Prog,Proc).
/* gather_builtins
   1st arg: a list of atoms						input
   2nd arg: the list of the relevant predicates				input
   3rd arg: the sublist of 1st of all the atoms with non relevant 	output
            predicates
*/
gather_builtins([],_,[],[]).
gather_builtins([H|B],Pred,I1,R1):-functor(H,Fu,Ar),
  (not_member([Fu,Ar],Pred)->
  (gather_builtins(B,Pred,I2,R1), I1=[H|I2]);
  (gather_builtins(B,Pred,I1,R2), R1=[H|R2])).

not_member(_,[]).
not_member(X,[A|L]):- X\==A, not_member(X,L).

/* make_tclss
   1st arg: a tclause							input
   2nd arg: the list of links of the head of 1st with its former body	input
   3rd arg: the list of lists of the leaves of the tlox 		input
   4th arg: a tclause which is made up from appending one leaf 	of	output
            each tlox
*/
make_tclss(Tcl,[],[],Tcl).
make_tclss([Head|Bd],Links,[Troc1|Trocc],Tcl) :- 
   one_tcl(Troc1,[Arg|Bd1]), copy_term([Arg|Bd1],[Arg2|Bd2]),
   copy_term([Links,[Head|Bd]],[[Lk|Lkk],Bef]), Lk=Arg2,
   append(Bef,Bd2,Td), make_tclss(Td,Lkk,Trocc,Tcl).
/* one_tcl
   1st arg: a list 							input
   2nd arg: an element of the list					output
*/
one_tcl([A|_],A).
one_tcl([_|B],C) :- one_tcl(B,C).
/* cfr_ls_def
   1st arg: a tlox							input
   2nd arg: a list of definitions					input
   3rd arg: 1st folded, if the goal succeeds				output
*/
cfr_ls_def(Bl,[D1|_],[Lv,Head]) :-  copy_term(D1,[Head|Bdd]),
   Head=..[_|Arg], faithful_instance(Bl,[Arg|Bdd],[Lv|Bd]), Bd=Bdd,!.
cfr_ls_def(Bl,[_|Lb],B1) :- cfr_ls_def(Bl,Lb,B1).
/* ck_new_def
   1st arg: a tlox							input
   2nd arg: the list of the defininitions				input
   3rd arg: the list of the tclauses, leaves of 1st			input
   4th arg: a list, its head is the head of the tclause, its tail	input
            instead is the tail of the tclause to build
   5th arg: the links							input
   6th arg: the list of tclauses, if 1st was a new defor an istance 	output
            of a new def we fold, else 3rd
*/
ck_new_def(Tl1,Def,Tclss,[Head|Bdd],Link,[[Heanw,Nat|Bdnw]]) :-
   cfr_ls_def(Tl1,Def,Rtl), Rtl=[_,At], [Rtl]\==Tclss, At =..[Name|Args],
   all_different_variables(Args),!, make_proc(Name,Tclss,Pro1),
   exclude_subsume(Pro1,Proc), write_proc(Proc), copy_term(Rtl,[Tes,Nat]),
   copy_term([Head,Bdd,Link],[Heanw,Bdnw,Lknw]), Lknw = Tes. 
ck_new_def(_,_,Tcl,[Head|Bdd],Link,TcllO) :-
   put_together(Tcl,Head,Bdd,Link,TcllO).
/* all_different_variables
   1st arg: a list of terms						input
   The goal succeeds if 1st is a list of variables with no repetitions
*/
all_different_variables([]).
all_different_variables([T|L]) :- var(T), not_member(T,L),
                                  all_different_variables(L).
/* cfr_ls_anc
   1st arg: a tlock							input
   2nd arg: a list of tloxs, 1st's ancestors				input
   3rd arg: the new definition						output
   4th arg: 1st folded							output
*/
cfr_ls_anc(Tl,[B1|_],[Head|Bd],[L,He]) :- 
   copy_term(B1,[Lv|Bd]), faithful_instance(Tl,[Lv|Bd],[L|Bdd]),
   gensym('new',Name), Head=..[Name|Lv], copy_term([Head|Bd],[He|B3]), B3=Bdd.
cfr_ls_anc(Tl,[_|Lb],D1,Rat) :- cfr_ls_anc(Tl,Lb,D1,Rat).
/* make_proc 
   1st arg: the name of the new predicate				input
   2nd arg: the list of the tclauses defining the new predicate		input
   3rd arg: the procedure defining the new predicate			output
*/
make_proc(Name,Tclss,Proc) :- 
   change_head(Name,Tclss,Proc1),rem_equ(Proc1,Proc2), 
   facts_first(Proc2,Proc),!.
/* change_head
   1st arg: the name of the new predicate				input
   2nd arg: the list of the tclauses defining 1st			input
   3rd the list of the clauses defining 1st				output
*/
change_head(_,[],[]).
change_head(Name,[[Ter|Bd]|Tcll],[[Head|Bd]|Cll]) :- 
   Head=..[Name|Ter], change_head(Name,Tcll,Cll).
/* reorders the clauses of a procedure */
facts_first(L,M):-divide_facts_others(L,Facts,Others),
		  append(Facts,Others,M).
divide_facts_others([],[],[]).
divide_facts_others([[C]|L],[[C]|M],N):-divide_facts_others(L,M,N).
divide_facts_others([C|L],M,[C|N]):-divide_facts_others(L,M,N).
/* rem_equ
   1st arg: a list of clauses 						input
   2nd arg: the same list of equalities without the equalities, 	output
            which have already been used
   if a clause has an inequality e.g. a= b then this clause is removed
   from the procedure
*/
rem_equ([],[]).
rem_equ([Cl1|Prc1],Prc3) :- 
   do_it_to_one(Cl1,Cl) -> (rem_equ(Prc1,Prc), Prc3 = [Cl|Prc]);
   rem_equ(Prc1,Prc3).
/* do_it_to_one
   1st arg: a clause							input
   2nd arg: the same clause but the equalites have been used		input
*/
do_it_to_one([],[]).
do_it_to_one([At|Rest],Rest1) :- 
   At =(_=_),!, call(At), do_it_to_one(Rest,Rest1).
do_it_to_one([At|Rest],[At|Rest1]) :- do_it_to_one(Rest,Rest1). 
/* db
   1st arg: the list of atoms to divide in bloxs			input
   2nd arg: intermediate value, set to [] on first call			input
   3rd arg: the list of blox wanted					output
*/
db([],Blox,Blox).
db([At|R],Bl1,Blox) :- add(At,Bl1,Bl2), db(R,Bl2,Blox).
add(At,Bl1,Bl2) :- 
   setof(Bl,exist_var(At,Bl1,Bl),Setbl),!, riblock(Setbl,At,Bl1,Bl2).
add(At,Bl1,Bl2) :- append(Bl1,[[At]],Bl2).
exist_var(At,[Bl|_],Bl) :- var_is_in(At,Bl).
exist_var(At,[_|Bl1],Bl) :- exist_var(At,Bl1,Bl).
var_is_in(At,[C|_]) :- shared_var(At,C).
var_is_in(At,[_|Bl]) :- var_is_in(At,Bl).
shared_var(At,C) :- variabili([At],V), variabili([C],W), shared(V,W).
shared([H|_],T) :- membert(H,T).
shared([_|L],T) :- shared(L,T).
/* riblock
   1st arg: the list of blocks to be merged in a new block		input
   2nd arg: the last atom						input
   3rd arg: the list of the blocks, in the right order of course	input
   4th arg: the new list of blocks, special care as been taken to       output
            make sure that a block comes before another one if its
            first atom comes before  the first atom of the other 
            block in the body of the clause  
*/
riblock(L,At,[H|R],[New_block|Rest_blocks]) :- 
   member3(L,H,L1),!, make_block(L,At,New_block), filter(L1,R,Rest_blocks).
riblock(L,At,[H|R],[H|Z]) :- riblock(L,At,R,Z).
/* make_block
   1st arg: a list of blocks to merge, the order is important		input
   2nd arg: the last atom of the new block 				input
   3rd arg: the new block was obtained appending in the right order	output
            the blocks in 1st and placing 2nd at the end
*/			
make_block([],At,[At]).
make_block([H|R],At,Bl) :- make_block(R,At,Bl1), append(H,Bl1,Bl).
/* filter
   1st arg: a list 							input
   2nd arg:  a second list						input
   3rd arg: the list of the elements of 2nd that aren't 	        output
            elements of 1st, the order is otherwise kept 
*/
filter(_,[],[]).
filter(L,[H|R],Z) :- membert(H,L),!, filter(L,R,Z).
filter(L,[H|R],[H|Z]) :- filter(L,R,Z).
/* distri1 
   1st arg: the list of all the atoms in the body of the clause		input
   2nd arg: the list of the blocks of the clause			input
   3rd arg: the list of the base and irrelevant atoms in the order	input
            in which they occur in the body of the clause	
   4th arg: the list of the atoms not relevant before the first		output
            relevant atom
   5th arg: the list of the subclauses in which I've divided the	output
            clause
   For all lists above mentioned the order is very important, in the 
   definition of the procedure I've taken advantage of it in several places
*/
distri1(_,Blxx,[],[],Blxx).
distri1([At1|Rbd],Blxx,[At1|Rirr],[At1|Rh],Mess) :-
                         distri1(Rbd,Blxx,Rirr,Rh,Mess).
distri1([At1|Rbd],Blxx,[At2|Rirr],[],Mess) :- At1 \== At2,
                         distri2(Rbd,Blxx,[At2|Rirr],Mess).
/* distri2
   1st arg: the list of all the atoms of the clause			input
   2nd arg: the list of the blocks of the clause			input
   3rd arg: the list of the base and irrelevant atoms 			input
   4th arg: the list of the subclauses in which I've divided the	output
            clause
*/
distri2(_,Blxx,[],Blxx).
distri2([At1|Rbd],[Bl1,Bl2|Rbxx],[At1|Rirr],Mess) :- 
                         distri2(Rbd,[[At1|Bl1],Bl2|Rbxx],Rirr,Mess).

/* in the above clause the order of the non relevant atoms is inverted, I'll
 correct this in get_tlock, instead in the last clause I need to invert on
purpose the order so that when I invert again in get_tlock I have at last
the rigth order
*/
distri2([At1|Rbd],[Bl1,Bl2|Rbxx],[At2|Rirr],Mess) :-
   At1 \== At2, membert(At1,Bl1),
   distri2(Rbd,[Bl1,Bl2|Rbxx],[At2|Rirr],Mess).
distri2([At2|Rbd],[Bl1,Bl2|Rbxx],[At1|Rirr],[Bl1|Mess]) :-
   At1 \== At2, not_member(At2,Bl1), membert(At2,Bl2),
   distri2(Rbd,[Bl2|Rbxx],[At1|Rirr],Mess).
distri2([At2|Rbd],[Bl1,Bl2|Rbxx],[At1|Rirr],Mess) :-
   At1 \== At2, not_member(At2,Bl1), not_member(At2,Bl2),
/* At2 is an atom of a previus block */
distri2(Rbd,[Bl1,Bl2|Rbxx],[At1|Rirr],Mess).
distri2(_, [Bl1],[A|Rirr],[Bl2]) :- reverse([A|Rirr],Rirr1),
                                       append(Rirr1,Bl1,Bl2).

/* distri1 distibutes the non relevant atoms in the body of the clause
according to the following guidelines:
1)if the non relevant atom At comes before any relevant atom in the body
of the tclause then place At in the list of atoms before any block, these
atoms will stay with the head of the tclause.
2) if At comes after the first atom of the i-th block and before the first
atom of the (i+1)-th block then  At is added to the i-th block.

distri2 is called when there has already been a relevant atom, so there isn't
any more the business of finding the atoms of the 1st kind.

a) If the atom At1  under examination is a non relevant one it is added to
the present block then we can go on to examine the nest atom of the clause.

b) If the At1 is not equal to the first non relevant atom of the ordered list
of non relevant atoms then At1 is relevant:

b1) if At1 is part of the present block the present block doesn't change, we
examine the next atom in the clause;

b2) if At1 is not part of the present block Bl1 but instead it's part of the
following block we wont need to add any more non relevant atoms to Bl1,Bl2
becomes the present block, we examine the next atom in the clause;

b3) if At1 isn't part of Bl1 nor of the next block Bl2 I belive At1 must be
part of a past block, just go on;

b4) if there is only one block left then it will get all the non relevant
atoms left.
*/
/* get_tlock
   1st arg: a T-subclause						input
   2nd arg: the list of the relevant predicates				input
   3rd arg: the tlock from the T-subclause				output
   4th arg: the head of the T-subclause followed by the non relevant    output
            predicates
   5th arg: the link between 4th and 3rd				output
*/
get_tlock([Head|Tscl],Pred,Tlo,[Head|Bds],Links) :- 
  find_non_rel(Tscl,Pred,Bds1,Relev), reverse(Bds1,Bds),
  variabili([Head|Bds],Lv1), variabili(Relev,Lv2),
  links(Lv1,Lv2,Links), copy_term([Links|Relev],Tlo).
/* find_non_rel
   1st arg: the body of the T-subclause					input
   2nd arg: the list of the relevant atoms				input
   3rd arg: the list of the non relevant atoms				output
   4th arg: the list of the relevant atoms				output
*/
find_non_rel([At1|Rest],Pred,[At1|R_irr],Rel) :- 
      functor(At1,Nam,Ar),
      not_member([Nam,Ar],Pred),
      find_non_rel(Rest,Pred,R_irr,Rel).
find_non_rel([At1|Rest],Pred,[],[At1|Rest]) :- functor(At1,Nam,Ar),
                                               membert([Nam,Ar],Pred).

/* put_together
   1st arg: the list of the tclauses descendants of the tlock		input
   2nd arg: the head of the T-subclause					input
   3rd arg: the non relevant atoms in the T-subclause, they'll go 	input
            at the end
   4th arg: the link between the tlock and the T-subclause		input
   5th arg: the list of the tclauses obtained				output
*/
put_together([],_,_,_,[]).
put_together([Tcl1|Rtcl],Head,Bdd,Link,[TclO|RtclO]) :- 
   copy_term(Tcl1,[Testa|Corpo]),
   copy_term([Head,Bdd,Link],[Heanw,Bddnw,Lknw]), Lknw = Testa,
   append([Heanw|Corpo],Bddnw,TclO), put_together(Rtcl,Head,Bdd,Link,RtclO). 
/* links_care
   1st arg: the entire tclause relevant and irrelevant predicates	input
   2nd arg: the list of the bloxx in which 1st has been divided		input
   3rd arg: the list of the tloxx of 1st				output
   4th arg: the list of the links between each tlox and 1st		output
*/ 
links_care(_,[],[],[]).  
links_care(Tcl,[Bx1|Bxx],[[Lv1|Bx1]|Rtxx],[Lv1|Rlnx]) :- 
  linking_vars(Bx1,Tcl,Lv1), links_care(Tcl,Bxx,Rtxx,Rlnx).
/* linking_vars
   1st arg: a list of atoms						input
   2nd arg: a second list of atoms including 1st			input
   3rd arg: the list of the linking variables of 1st in 2nd		output
*/
linking_vars(Bl,[H|R],L):-variabili(Bl,L1),
   remove_predicates(R,Bl,R2), variabili([H|R2],L2), links(L1,L2,L),!.
/* links
   1st arg: a list 							input
   2nd arg: a second list						input
   3rd arg: the list of the common elements of the two list		output
*/
links([],_,[]).
links([H|R],L,[H|R1]):-membert(H,L), links(R,L,R1).
links([H|R],L,R1):- not_member(H,L), links(R,L,R1).
/* remove_predicates
   1st arg: a list of atoms						input
   2nd arg: a second list of atoms					input
   3rd arg: the list of the elements of the first list which 		output
            aren't elements of the second list
*/
remove_predicates([],_,[]).
remove_predicates([H|R],Bl,R2):-membert(H,Bl),
                                remove_predicates(R,Bl,R2).
remove_predicates([H|R],Bl,[H|R2]):-remove_predicates(R,Bl,R2).
/* permute_furb
   1st arg: a list of atoms						input
   2nd arg: a second list of atoms, the length  is =< of 1st 		input
   3rd arg: a permutation of 1st, or of a sublist of 1st, if there	output
            is any, such that each element of 3rd is an instance of 
            the corresponding element  of 2nd
*/
permute_furb(_,[],[]).
permute_furb(S,[G|K],[M|J]) :- member_match(S,G,S1,M), permute_furb(S1,K,J).
/* member_match
   1st arg: a list of atoms						input
   2nd arg: an atom							input
   3rd arg: if in 1st there is an atom, which is an istance of 2nd	output
            the list is returned without this atom
   4th arg: the istance of 2nd						output
*/
member_match([S|L],G,L,S) :- subsumes_chk(G,S).
member_match([S|L],G,[S|K],M) :- member_match(L,G,K,M).
/* faithful_instance
   1st arg: a tlock							input
   2nd arg: a second tlock						input
   3rd arg: the permutation of 1st if this is a faithful instance	output
            of 2nd
    The goal succeeds if a permutation of 1st is a faithfull instance
    of 2nd; that is an istance and also the fact that to every existencial
    variable corresponds an existencial variable and no double occurrence
    of variables in the range of the bindings
*/
faithful_instance([Vs|S],[Vg|G],[Vs|Per_S]) :-
  length(S,Ns), length(G,Ng), Ns=Ng, permute_furb(S,G,Per_S), 
  find_exist_vars(Per_S,Vs,Ex_var), fi([Vg|G],[],[],Ex_var,[Vs|Per_S]).
/* fi
   1st arg: a block							input
   2nd arg: the list of links between existencial variables		input
            in the two blocks; set to [] on first call
   3rd arg: the list of links between a linking var and a term 		input
            without variables occurring more than once; [] first call
   4th arg: the list of the existencial variables of 5th		input
   5th arg: a second block						input
   The goal succeeds if 5th is, no permutations allowed, a faithful
   instance of 1st
*/
fi([_],_,_,_,[_]).
fi([La|[A1|A]],E,Lv,Evar,[Lb|[B1|B]]) :- 
   ground(A1), ground(B1), A1==B1, fi([La|A],E,Lv,Evar,[Lb|B]).

fi([La|[A1|A]],E,Lv,Evar,[Lb|[B1|B]]) :- 
   var(A1), not_member(A1,La), var(B1), membert(B1,Evar), !,
   /* two existencial variables */
   look_in_exis_ls(A1,B1,E,E1), fi([La|A],E1,Lv,Evar,[Lb|B]).

fi([La|[A1|A]],E,Lv,Evar,[Lb|[B1|B]]) :- var(A1), membert(A1,La),
   !, /* the first is a linking variable */
   (ck_term([B1],Lb,Nlb) -> ck_bindings(A1,B1,Lv,Lv1);
   (ck_bindings(A1,B1,Lv,Lv), Lv=Lv1,Lb=Nlb)),  fi([La|A],E,Lv1,Evar,[Nlb|B]).

/* the first ck_binding checks if something wasn't already given to A1, instead
 the second ck_bindings succedes only if a binding for A1, B1 has already been 
established */
fi([La|[A1|A]],E,Lv,Evar,[Lb|[B1|B]]) :- compound(A1), compound(B1),
   functor(A1,Nm,Ar), functor(B1,Nm,Ar), A1=..[_|Ma], B1=..[_|Mb],
   append(Ma,A,A2), append(Mb,B,B2), fi([La|A2],E,Lv,Evar,[Lb|B2]).
/* look_in_exis_ls
   1st arg: an existencial variable					input
   2nd arg: a second existencial variable				input
   3rd arg: a list of links between existencial variables		input
   4th arg: if there already exist a link between the two variables	output
            the 3rd list is returned, if none of the two variables
            occurs in any existing link, the link between them is added
*/
look_in_exis_ls(A1,B1,[],[[A1,B1]]).
look_in_exis_ls(A1,B1,[[A,B]|Rest],[[A,B]|Rest1]) :- A1\==A, B1\==B,
   look_in_exis_ls(A1,B1,Rest,Rest1).
look_in_exis_ls(A1,B1,[[A,B]|Rest],[[A,B]|Rest]) :- A1==A, B1==B .
/* ck_term
   1st arg: a list of terms						input
   2nd arg: the list of the linking variables of the block to which 	input
            the term belongs, please no repetitions
   3rd arg: the new list of the linking variables. The variables 
            which have been found in 1st were removed from 2nd.
   The goal succeeds if all the variables are linking, and they belong to 2nd
*/
ck_term([],S,S).
ck_term([B|R],Lb,Nbl) :- ground(B), ck_term(R,Lb,Nbl).
ck_term([B|R],Lb,Nbl) :- var(B), member3(Lb,B,Lb1), ck_term(R,Lb1,Nbl). 
ck_term([B|R],Lb,Nbl) :-
   compound(B), B=..[_|L], append(L,R,R1), ck_term(R1,Lb,Nbl).
/* ck_bindings
   1st arg: a linking variable						input
   2nd arg: a term							input
   3rd arg: a list of bindings between variables and terms		input
   4th arg: if a binding between the 1st and 2nd already exists 3rd     output
            is returned, if a binding between 1st an some term
            different from 2nd exist, the goal fails, otherwise
            the new binding is added to the list
*/
ck_bindings(A,B,[],[[A,B]]).
ck_bindings(A1,B1,[[A,B]|L],[[A,B]|L]) :- A==A1, B==B1.
ck_bindings(A1,B1,[[A,B]|L],[[A,B]|L1]) :- A\==A1, ck_bindings(A1,B1,L,L1).
/* member3
   1st arg: a list 							input
   2nd arg: a term							input
   3rd arg: if the term belongs to the list, the goal succeds and	output
            the rest of the list is returned in the same order as they
            appear in 1st
*/
member3([A|L],At,L) :- A==At.                   
member3([A|L],At,[A|L1]) :- A\==At, member3(L,At,L1). 
%   predicate just exists in aioni's transformer.
%
/* find_exist_vars
   1st arg: a list of atoms, i.e. a block				input
   2nd arg: the list of the linking variables of 1st			input
   3rd arg: the list of the existencial variables of 1st		output
*/
find_exist_vars(Bl,Lv,Ev) :- variabili(Bl,Variabili),setminus(Variabili,Lv,Ev).
/* setminus
   1st arg: the list of all the variables of a block			input
   2nd arg: the list of the linking variables of the block		input
   3rd arg: the list of the existencial variables of the block		output
*/
setminus([], _, []). 
setminus([A|Vrs], Lv, E) :- membert(A,Lv) -> setminus(Vrs,Lv,E);
                        (setminus(Vrs,Lv,E1),  E = [A|E1]).
/* gen_step
   1st arg: the block in need of generalization				input
   2nd arg: the block after the generalization				input
*/
gen_step([Head|Bl],[Head|Bg]) :- 
   bound(H,V), normV(Bl,V,Nbl,L,[]),normH(Nbl,H,Eq,New_bl),
   append(Eq,L,Total_eq), append(New_bl,Total_eq,Bg).
/* normV
   1st arg: list of atoms, 						input
   2nd arg: natural number, 						input
   3rd arg: the generalisation of 1st such that the heigth of each	output
            atom in this new list is =< than 2nd, the order is
            preserved.
   4th arg: the list of equalities to add to 5th 			output
   5th arg: the list of equalities to start with, [] on first call	input
*/
normV([], _, [], [],_).
normV([T|L], 0, [V|La], L_out,L_in) :- 
   compound(T), (search(T,L_in,V) -> normV(L,0,La,L_out,L_in); 
   (copy_term(_,V), normV(L,0,La,L_out1,[T=V|L_in]), [T=V|L_out1] = L_out)). 
normV([T|L], N, [T|La], L_out,L_in) :- (var(T); atomic(T)),
    normV(L, N, La, L_out,L_in).
normV([T|L], N, [Ta|La], L_out,L_in) :- N>0, compound(T), T=..[P|Arg],
  M is N-1, normV(Arg,M,AArg,L1,L_in), Ta=..[P|AArg],
  append(L1,L_in,L2), normV(L,N,La,L3,L2),  append(L1,L3,L_out).
/* normH
   1st arg: a block, that is a list of atoms, 				input
   2nd arg: the horizontal bound					input
   3rd arg: the list of the equalities  added to split  1st		output
   4th arg: the horizontal generalization of 1st			output
*/
normH(Bl,H,[],Bl) :- length(Bl,Lh), Lh=<H.
normH(Bl,H,E,Hgl) :- length(Bl,Lh),Lh>H,  bagof(M,spezza(Bl,H,M),Lm),
                     find_best(Lm,[Vc,X]), sostitut(Bl,X,Vc,[],Hgl,E).
/* spezza
   1st arg: a list of atoms, as for example a block,			input
   2nd arg: natural namber						input
   3rd arg: a couple of lists: the second is a subsequence of 1st       output
            first is  the list of the linking variables betwen the
            subsequence  and 1st      				
*/
spezza(Bl,H,[Vc,X]) :- subseq(X,Bl), length(X,Nx),
   Nx<H, ((X=[At], At=..[_|Args], not_all_diff_vars(Args))  ;
   X =[_,_|_]), filter(X,Bl,Y),
    ((Y=[At3], At3=..[_|Args1], not_all_diff_vars(Args1)) ; Y= [_,_|_]),
     variabili(X,Vx), variabili(Y,Vy), links(Vx,Vy,Vc).
not_all_diff_vars([A|_]) :- nonvar(A).
not_all_diff_vars([A|L]) :- 
   var(A), (membert(A,L) ->true; not_all_diff_vars(L)).
/* variabili
   1st arg: list of terms						input
   2nd arg: list of the variables of 1st, no repetitions		output
*/
variabili([],[]).
variabili([H|R],S) :- var(H), variabili(R,R1), app_sine_repet([H],R1,S).
variabili([H|R],R1) :- ground(H), variabili(R,R1).
variabili([H|R],R1) :- compound(H), H=..[_|X],
    variabili(X,L), variabili(R,L1), app_sine_repet(L,L1,R1).
/* subseq
   1st arg: a list							output
   2nd arg: a subsequence of 1st, i.e. the elements of it are	        input
            elements of the list and they occur in the same order
            of the list						
*/
subseq([A],[A|_]).
subseq([A|X],[A|B]) :- subseq(X,B).
subseq([A|X], [_|B]) :- subseq([A|X],B).
/* sostitut
   1st arg: list of terms\atoms						input
   2nd arg: sublist of 1st						input
   3rd arg: the list of the linking variables betwen 1st and 2nd	input
   4th arg: list of bindings betwen one of  3rd's variables and		input
            a fresh variable; set at [] on  first call		
   5th arg: the 1st changed so that the links betwen 2nd and the	output
            rest of 1st have been severed i.e. there are no more
            linking variables betwen them	
   6th arg: the list of equalities that had to be added to 4th          output
            to bring home the bacon, that is 5th			
*/
sostitut([A|Bl], X, Vc, L_in, [C|B_out], L_out) :- 
  membert(A,X), sost_var([A], Vc, L_in, L, [C]), append(L, L_in, L2),
  sostitut(Bl,X,Vc,L2,B_out,L3), append(L,L3,L_out).

sostitut([A|Bl],X,Vc,L_in,[A|B_out],L_out):-
   not_member(A,X),  sostitut(Bl,X,Vc,L_in,B_out,L_out).
sostitut([],_,_,_,[],[]).
/* sost_var
   1st arg: a list of atoms/terms					input
   2nd arg: a list of variables						input
   3rd arg: list of bindings betwen one of the 3rd variables and	input
            a fresh variable; set at [] on first call
   4th arg: the list of equalities that had to be added to 4th          output
            to get 5th
   5th arg: the 1st list has been changed so that there are no more	output
            occurences of the 2nd variables, they have been 
            replaced according with 1st and 4th
*/
sost_var([],_,_,[],[]).
sost_var([A|R], Vc,L_in,[A=D|L_out],[D|S]) :-
   var(A), membert(A,Vc), search_inv(A,L_in), copy_term(_,D), 
   sost_var(R,Vc,[A=D|L_in],L_out,S).
sost_var([A|R], Vc,L_in,L_out,[A|S]) :- 
   var(A), not_member(A,Vc), sost_var(R,Vc,L_in,L_out,S).
sost_var([A|R], Vc,L_in,L_out,[D|S]) :- 
   var(A), membert(A,Vc), search(A,L_in,D), sost_var(R,Vc,L_in,L_out,S).
sost_var([A|R], Vc,L_in,L_out,[A|S]) :- ground(A), sost_var(R,Vc,L_in,L_out,S).
sost_var([A|R], Vc,L_in,L_out,[D|S]) :- 
   \+(ground(A)), compound(A), A =..[P|Arg],
   sost_var(Arg,Vc,L_in,L1,AArg), D =..[P|AArg],  append(L1,L_in,L2),
   sost_var(R,Vc,L2,L3,S), append(L1,L3,L_out).
/* search_ inv
   1st arg: a variable or a term					input
   2nd arg: a list of bindings like variable/term = variable		input
   the goal succeds if there is no binding involving 1st
*/
search_inv(_,[]).
search_inv(A,[D|L]) :- D= (S=_), S\==A, search_inv(A,L).
/* search
   1st arg: a variable/term						input
   2nd arg: a list of bindings like "variable/term = variable" 		input
   3rd arg: if there exists a binding involving 1st, it reports		input
            the other membert
*/
search(A,[S|_],D) :- S=(F=D), F==A.
search(A,[S|L],D) :- S=(F=_), F\==A, search(A,L,D).
/* app_sine_repet
   1st arg: list without repetitions					input
   2nd arg: list without repetitions					input
   3rd arg: the concatenation of the two lists without repetions	output
*/
app_sine_repet([],M,M).
app_sine_repet([A|L],M,[A|Lr]) :- not_member(A,M), app_sine_repet(L,M,Lr).
app_sine_repet([A|L],M,Lr) :- membert(A,M), app_sine_repet(L,M,Lr).
/* controllo
   1st arg: the block, although the atoms aren't in the right		input
           order, of whom we are going to check the lenght and the
           height of its atoms
   The goal succeeds if at least one of the bounds are superati
*/
controllo([_|Bl]) :- bound(_,V), /* bound is a dynamic predicate*/
                     ck_depth(Bl,V).
controllo([_|Bl]) :- length(Bl,M), bound(H,_), M>H.
/* ck_depth
   1st arg: the list of atoms of whom we want to check the depth	input
   2nd arg: the vertical bound						input
*/
ck_depth([A|_],V) :- depth(A,N), N>V.
ck_depth([_|R],V) :- ck_depth(R,V).
/* find_best
   1st arg: a list of elements like the following [Vc,X], where 	input
            the second is a subsequence, the first is the list of the
            linking variables
   2nd arg: is  the element with the shortest list of linking 		output
            variables
*/
find_best([A],A).
find_best([A|L],Best) :- find_best(L,B1), better(B1,A,Best).
/* better
   1st arg: a list like the following one: [A,B] where A is a list 	input
            of variables and B is a list of atoms
   2nd arg: same as 1st							input
   3rd arg: the one of the two who has the shorter list of variables	output
*/
better([A,B],[C,D],[A,B]) :- 
   length(A,Ca), length(B,Cb), length(C,Cc), length(D,Cd),
   (Ca<Cc ; (Ca=Cc,Cb>Cd)).
better(_,[C,D],[C,D]).
gensym(Prefix, V) :-
   var(V), atomic(Prefix), (retract(gensym_counter(Prefix, M)) ;   M = 0 ),
   N is M+1, asserta(gensym_counter(Prefix, N)), concat(Prefix, N, V), !.
concat(Pref,N,V) :- name(Pref,L), name(N,L1), append(L,L1,S),  name(V,S).

