% man.pl
%
% This file contains the clauses for the man command which gives help
% about transformation rules, strategies and other commnads.
%

man(ts_define) :- !, man(ts_d).
man(ts_d) :- !,
   nl, write('ts_d:'),
   nl, write('Adds a new definition (using one clause only),'),
   nl, write('Example:'), nltab,
   write('| ?- ts_d.'), nltab,
   write('enter definition clause:'), nltab,
   write('|: r :- q.'), nltab,
   write('3. r :- q.'), nltab,
   write('| ?-'), 
   nl, write('The new clause is given the number 3 and is added to the current program.'), nl.

man(gen_define) :- !, man(d).
man(d) :- !, 
   nl, write('d:'),
   nl, write('Adds a new definition (using one or more clauses).'), 
   nl, write('Example:'), nltab,
   write('| ?- d.'), nltab,
   write('Do you want to check the existing definitions [n]?'), nltab,
   write('enter clauses (use the same new predicate for the heads):'), nltab,
   write('first clause: s :- p(a).'), nltab,
   write('Another definition [n]? y'), nltab,
   write('clause: s :- r, p(b), p(c).'), nltab,
   write('Another definition [n]? y'), nltab,
   write('clause: s :- r, p(b), p(a).'), nltab,
   write('Another definition [n]? '), nltab,
   write('4. s :- p(a).'), nltab,
   write('5. s :- r, p(b), p(c).'), nltab,
   write('6. s :- r, p(b), p(a).'), nltab,
   write('| ?-'),
   nl, write('The new clauses are given the numbers 4, 5, and 6 respectively,'),
   nl, write('and are added to the current program.'), nl.

man(unfold) :- !, man_(u).
man(u) :- !, man_(u), man(type_unf_1), nl, nl, 
    write('See also commands ud, ud_p, ud_i, ud_facts, ud_eqs.'), man_u_ex.

man(unfold_d_i) :- !, man(ud_i).
man(ud) :- !, man_(ud),  man(type_unf_2), nl, nl, 
    write('See also commands u, ud_p, ud_i, ud_facts, ud_eqs.'), man_u_ex.

man(ud_p) :- !, man_(ud_p), man(type_unf_2), nl, nl, 
    write('See also commands u, ud, ud_i, ud_facts, ud_eqs.'), man_u_ex.

man(ud_i) :- !, man_(ud_i), man(type_unf_2), nl, nl, 
    write('See also commands u, ud, ud_p, ud_facts, ud_eqs.'), man_u_ex.

man(type_unf_1) :-
   nl, nl, 
   write('(*) The unfolding wrt an atom which does not unify with the head'),
   nl, write('    of any clause fails (clause deletion is not performed).').
man(type_unf_2) :-
   nl, nl,
   write('(**) The unfolding of C wrt an atom whose predicate never occurs'),
   nl, write('    in the head of any clause fails. Otherwise, if the unfolding'),
   nl, write('    of C wrt an atom whose predicate occurs in the head of a'),
   nl, write('    clause and that atom does not unify with the head of any'),
   nl, write('    clause, then clause C is deleted.'), nl.

man(unfold_d_facts) :- !, man(ud_facts).
man(ud_facts) :- !, 
    nl, write('ud_facts(C):'),
    nltab, write('Unfolding (**) of clause C wrt all atoms which'),
    nltab, write('are instances of facts.'),
    nl, write('ud_facts:'),
    nltab, write('Unfolding (**) of all clauses of the current program'),
    nltab, write(' wrt all atoms which are instances of facts.'),
    man(type_unf_2), nl, nl, 
    write('See also commands u, ud, ud_p, ud_i, ud_eqs.'), man_u_ex.


man(unfold_d_eqs) :- !, man(ud_eqs).
man(ud_eqs) :- !, 
    nl, write('ud_eqs(C):'),
    nltab, write('Unfolding of clause C wrt all equalities.'),
    nl, write('ud_eqs:'), 
    nltab, write('Unfolding of all clauses of the current program'),
    nltab, write('wrt all equalities.'), nl.

man(ts_fold) :- !, man(ts_f).
man(ts_f) :- !,
    nl, write('ts_f(C,As,D):'),
    nltab,
    write('Tamaki and Sato folding of clause C wrt atoms As using definition D.'),
    nltab, write('C is a clause number in the current program.'),
    nltab,
    write('As is a list of atom numbers in the body of clause C.'),
    nltab, write('D is the number of a clause introduced by definition.'), nl,
    nl, write('Example:'), nltab,
    write('Assume that clause 5 has been introduced by definition:'), nltab,
    write('5. p(A,B,C,D,E,F) :- s_p(A,B,E,G), s_p(C,D,E,H), plus(G,H,F).'), nltab,
    write('and clause 13 is a clause in the current program:'), nltab,
    write('13. p([A|B],[C|D],[E|F],[G|H],s(I),J) :- s_p(B,D,I,K), mult(A,C,L),'), nltab,
   write('   plus(L,M,J), plus(K,N,O), plus(O,P,M), s_p(F,H,I,N), mult(E,G,P).'),
nl, nltab,
   write('| ?- ts_f(13, [1,6,4], 5).'), nltab,
   write('by folding clause 13 w.r.t. atoms [1,6,4] using 5 we get:'), nltab,
   write('14. p([A|B],[C|D],[E|F],[G|H],s(I),J) :- p(B,D,F,H,I,K), mult(A,C,L),'),
   nltab, write('   plus(L,M,J), plus(K,N,M), mult(E,G,N).'),  nltab,
   write('| ?-'), nl.
 
man(fold) :- !, man(f).
man(f) :- !,
   nl, write('f(Cls11,Cls2):'),
   nltab, write('Folding of clauses Cls1 using clauses Cls2.'),
   nltab, write('Cls1 is a list of clause numbers in the current program.'),
   nltab, write('Cls2 is a list of clause numbers in the current program, or'),
   nltab, write('in an ancestor of the current program.'),
   nltab, write('Cls1 and Cls2 must be ordered so that the i-th clause in Cls1'),
   nltab, write('is folded using the i-th clause in Cls2.'),
   nltab, write('No rearrangement of atoms in the bodies of the clauses is performed.'),
   nltab, write('See also command r_f.'), nl,
   nl, write('Example:'), nltab,
   write('Assume that the current program is made out of the following clauses'),
   nltab,
   write('1. p([a|A]) :- t(A), u(A), v(A).'), nltab,
   write('2. p([b|A]) :- t(A), u(A), s(A).'), nltab,
   write('3. p2(a,A) :- v(A).'), nltab,
   write('4. p2(b,A) :- s(A).'), nl, nltab,
   write('| ?- f([1,2], [3,4]).'), nltab,
   write('by folding clauses [1,2] using clauses [3,4] we get:'), nltab,
   write('5. p([A|B]) :- t(B), u(B), p2(A,B).'), nltab,
   write('| ?-'), nl. 

man(rearrange_and_fold) :- !, man(r_f).
man(r_f) :- !,
   nl, write('r_f(Cls11,Cls2):'),
   nltab, write('Folding of clauses Cls1 using clauses Cls2.'),
   nltab, write('Cls1 is a list of clause numbers in the current program.'),
   nltab, write('Cls2 is a list of clause numbers in the current program,'),
   nltab, write('or in an ancestor of the current program.'), nl,
   nl, write('Example:'), nltab,
   write('Assume that the current program is made out of the following clauses'),
   nltab,
   write('1. p([a|A]) :- t(A), u(A), v(A).'), nltab,
   write('2. p([b|A]) :- u(A), s(A), t(A).'), nltab,
   write('3. p2(a,A) :- v(A).'), nltab,
   write('4. p2(b,A) :- s(A).'), nl, nltab,
   write('| ?- f([1,2], [3,4]).'), nltab,
   write('by folding clauses [1,2] using clauses [3,4] we get:'), nltab,
   write('p([A|B]) :- p2(A,B), t(B), u(B).'), nltab,
   write('| ?-'), nl.

man(replace_goal) :- !, man(rep).
man(rep) :- !,
   nl, write('rep(C,As,L):'),
   nltab, write('Replaces the goal made out of atoms As in the body of clause'),
   nltab, write('C by another goal according to replacement law L. '),
   nltab, write('C is a clause number in the current program.'), nltab,
   write('As is a list of atom numbers in the body of clause C.'),
   nltab, write('L is a law number in the current theory.'), nltab,
   nl, write('Example:'), nltab,
   write('Assume that the current program contains clause 10:'),
   nltab,
   write('10. p([A|B],[C|D],[E|F],[G|H],s(I),J) :- s_p(B,D,I,K), mult(A,C,L),'),
   nltab,
   write('   plus(L,K,M), s_p(F,H,I,N), mult(E,G,O), plus(O,N,P), plus(M,P,J).'),
   nltab, write('and the current theory contains law 1:'), nltab,
   write('1. plus(A,B,C), plus(C,D,E) -> plus(A,F,E), plus(B,D,F).'), nl,
   nltab, write('| ?- rep(10,[3,7],1).'), nltab,
   write('by replacing goal [plus(A,B,C),plus(C,D,E)] by [plus(A,F,E),plus(B,D,F)]'), nltab,
   write('in clause 10 using law 1 we get:'), nltab,
   write('11. p([A|B],[C|D],[E|F],[G|H],s(I),J) :- s_p(B,D,I,K), mult(A,C,L),'),
   nltab, write('   plus(L,M,J), plus(K,N,M), s_p(F,H,I,O), mult(E,G,P), plus(P,O,N).'), nltab,
   write('| ?-'), nl.


man(move_lit) :- !, man(mv).
man(mv) :- !,
   nl, write('mv(C,Pos1,Pos2):'),
   nltab, write('Moves atom from position Pos1 to position Pos2'),
   nltab, write('in the body of clause C.'), nl,
   nl, write('Example:'), nltab,
   write('Assume that the current program contains clause 6:'),
   nltab,
   write('6. p(A) :- \+A=a, q(A,B), r(B).'),
   nl, nltab,
   write('| ?- mv(6,1,3).'), nltab,
   write('by moving atom 1 to position 3 in clause 6 we get:'), nltab,
   write('8. p(A) :- q(A,B), r(B), \+A=a.'), nltab,
   write('| ?-'), nl.
 
man(add_goal) :- !, man(add_g).
man(add_g) :- !,
   nl, write('add_g(C):'),
   nltab, write('Adds a goal in the body of clause C.'),
   nltab, write('The goal to be added is entered by the user interactively.'),
   nl, nl, write('Example:'), nltab,
   write('Assume that the current program contains clause 11:'), nltab,
   write('11. p(A) :- q(A,B), r(B).'), 
   nl, nltab,
   write('| ?- add_g(11).'), nltab,
   write('Goal: s(a,b).'), nltab,
   write('position of new goal [3]: 2'), nltab,
   write('by adding goal s(a,b). in clause 11 at position 2 we get:'), nltab,
   write('12. p(A) :- q(A,B), s(a,b), r(B).'), nltab,
   write('| ?-'), nl.
 

man(delete_goal) :- !, man(del_g).
man(del_g) :- !,
   nl, write('del_g(C,As):'),
   nltab, write('Deletes the goal made out of atoms As in the body of clause C.'),
   nl, nl, write('Example:'), nltab,
   write('Assume that the current program contains clause 10:'), nltab,
   write('10. p(A) :- \+A=b, q(A,B), r(B), \+A=a.'), nltab,
   nltab,
   write('| ?- del_g(10,[1,4]).'), nltab,
   write('by deleting goal [1,4] in clause 10 we get:'), nltab,
   write('11. p(A) :- q(A,B), r(B).'), nltab,
   write('| ?-'), nl.


man(restriction) :-!, man(restr).
man(restr) :-!,
   nl, write('restr(P/A):'),
   nl, write('Delete all clauses from current program that are not used'),
   nl, write('in the definition of predicate P of arity A.'),
   nl, write('Example:'), nltab,
   write('Assume that current program is made out of the following clauses:'),
   nltab, 
   write('1. average(A,B) :- sumlist(A,C), len(A,D), div(C,D,B).'), nltab,
   write('2. sumlist([],0).'), nltab,
   write('3. sumlist([A|B],C) :- sumlist(B,D), plus(A,D,C).'), nltab,
   write('4. len([],0).'), nltab,
   write('5. len([A|B],C) :- len(B,D), plus(D,1,C).'), nltab,
   write('6. div(A,B,C) :- C is A/B.'), nltab,
   write('7. plus(A,B,C) :- C is A+B.'), nltab,
   write('| ?- restr(sumlist/2).'), nltab,
   write(' elimination of transient clause(s) [3,4,5,7] '), nltab,
   write( '| ?- '), nl.

man(delete_clauses) :- !, man(del_cl).
man(del_cl) :- !,
   nl, write('del_cl(Cls):'),
   nltab, write('Deletes clause(s) Cls.'),
   nltab, write('Cs might be a clause number or a list of clause numbers.'),
   nl.

man(delete_failing_body_clauses) :- !, man(del_f_cl).
man(del_f_cl) :- !,
   nl, write('del_f_cl(Cls):'), nltab,
   write('Deletes among clause(s) Cls, those clauses which have a failing body,'),
   nltab, write('that is, which contain an atom whose predicate is defined in '),
   nltab, write('the current program but which does not unify with the head of any clause.'),
   nltab, write('Cls may be a clause number or a list of clause numbers.'),  
   nl, write('del_f_cl:'), nltab,
   write('short for del_f_cl(Cls)  where Cls is the list of all clauses'),
   nltab, write('in the current program.'), 
   nl, nl, write('Example:'), nltab,
   write('Assume that the current program is made out of the following clauses'),
   nltab,
   write('2. q(a,b).'), nltab,
   write('3. r(b).'), nltab,
   write('9. p(b) :- q(b,A), r(A), \+b=a.'), nl, nltab,
   write('| ?- del_f_cl.'), nltab,
   write('deletion of clause 9 because of failure in body'), nltab,
   write('| ?-'), nl.

man(add_clause) :- !, man(add_cl).
man(add_cl) :- !,
   nl, write('add_cl:'), nltab,
   nltab, write('Adds a clause to the current program.'), 
   nltab, write('The clause to be added is entered by the user interactively.'),
   nl, nl, write('Example:'), nltab,
   write('Assume that the current program is made out of the following clauses'),
   nltab,
   write('1. p([a|A]) :- t(A), u(A), v(A).'), nltab,
   write('2. p([b|A]) :- t(A), u(A), s(A).'), nltab,
   write('3. p2(a,A) :- v(A).'), nltab,
   write('4. p2(b,A) :- s(A).'), nl, nltab,
   write('| ?- add_cl.'), nltab,
   write('clause to be added:'), nltab,
   write('|: p([a]) :- q(a,A).'), nltab,
   write('Before which clause ? [last]: 3'), nltab,
   write('Addition of clause 5:'), nltab,
   write('1. p([a|A]) :- t(A), u(A), v(A).'), nltab,
   write('2. p([b|A]) :- t(A), u(A), s(A).'), nltab,
   write('5. p([a]) :- q(a,A).'), nltab,
   write('3. p2(a,A) :- v(A).'), nltab,
   write('4. p2(b,A) :- s(A).'), nltab,
   write('| ?-'), nl.



man(replace_clause) :- !, man(rep_cl).
man(rep_cl) :- !,
   nl, write('rep_cl(C):'), nltab,
   nltab, write('Replaces clause C by another clause in the current program,'),
   nltab, write('The new clause is entered by the user interactively and'),
   nltab, write('it takes the place of clause C in the current program'), nl.

man(move_clause) :- !, man(mv_cl).
man(mv_cl) :- !,
   write('mv_cl(C):'), nltab,
   nltab, write('Moves clause C in the current program to a position.'),
   nltab, write('which is entered by the user interactively.'), nl.

man(subsumption) :- !, man(subs).
man(subs) :- !,
   nl, write('subs(Cs):'), nltab,
   write('Applies subsumption to clauses Cs, that is, deletes among Cs'), nltab,
   write('all clauses that are subsumed by another clause in Cs.'), nltab,
   nl, write('subs:'), nltab,
   write('is a short for subs(Cs), that is, applies subsumption to current program.'), nltab,
   nl, write('Example:'), nltab,
   write('Assume that the current program is made out of the following clauses'),
   nltab,
   write('5. p(a) :- q(a,A), r(A).'), nltab,
   write('6. p(A) :- \+A=a, q(A,B), r(B).'),  nltab,
   write('7. p(a) :- q(a,A).'), nltab,
   write('| ?- subs.'), nltab,
   write('clause 5 subsumed by 7'), nltab,
   write('| ?- p.'), nltab,
   write('6. p(A) :- \+A=a, q(A,B), r(B).'), nltab,
   write('7. p(a) :- q(a,A).'), nl.


man(split_clause) :- !, man(split).
man(split) :- !,
   nl, write('split(C,A):'), nltab,
   write('Splits clause C w.r.t. atom A'),
   nl, nl, write('Example:'), nltab,
   write('Assume that the current program contains clause 4:'), nltab,
   write('4. p(A) :- q(A,B), r(B).'), 
   nl, nltab,
   write('| ?- split(4, p(a)).'), nltab,
   write('by case splitting clause 4 we get:'), nltab,
   write('5. p(a) :- q(a,A), r(A).'), nltab,
   write('6. p(A) :- \+A=a, q(A,B), r(B).'), nltab,
   write('| ?-'), nl.

man(instantiation_clause) :- !, man(i).
man(i) :- !, 
   nl, write('i(C,Vs,Ts):'), nltab,
   write('Instantiates clause C w.r.t. variable number(s) Vs of type(s) Ts.'),
   nltab,
   write('Variables are ordered according to the position of their first'),
   nltab, write('occurrence, starting from 1 for the leftmost variable.'),
   nltab, write('For instance, in the clause p(X,X,Y) :- q(X,Y,Z), r(Z).,'),
   nltab, write('variable numbers 1, 2, and 3 denote the variables'),
   write(' X, Y, and Z'),
   nltab, write('respectively.'),
   nltab, write('A type must be the name of a predicate of arity 1.'),
   nl, nl, write('Example:'), nltab,
   write('Assume that the current program contains the following clauses:'),
   nltab,  write(' 1. list([]).'),
   nltab,  write('2. list([A|B]) :- list(B).'),
   nltab,  write('3. nat(0).'),
   nltab,  write('4. nat(s(A)) :- nat(A).'),
   nltab,  write('5. p(A,B) :- length(A,B).'),
   nl, nltab,
   write('| ?- i(5,[1,2],[nat,list]).'), nltab,
   write('by instantiating clause 5 w.r.t. the variables [1,2]'),
   nltab, write(' with types [nat,list] we get:'), nltab,
   nltab,  write('6. p(0,[]) :- length(0,[]).'),
   nltab,  write('7. p(0,[A|B]) :- length(0,[A|B]).'),
   nltab,  write('8. p(s(A),[]) :- length(s(A),[]).'),
   nltab,  write('9. p(s(A),[B|C]) :- length(s(A),[B|C]).'),
   write('| ?-'), nl.


man(start) :- !, man(st).
man(st) :- !, 
       nl, write('st:'),
       nltab, write('Initialization of a transformation process by selecting'),
       nltab, write('an initial program available in the directory Progs/'),
       nl, write('st(P):'),
       nltab, write('Initialization of a transformation process with P'),
       nltab, write('as initial program.'),
       nltab, write('P must be an existing file in the directory Progs/'), nl.

man(load_session) :- !, man(load_s).
man(load_s) :- !, 
       nl, write('load_s:'), nltab, 
       write('Restores a transformation tree to be selected via a menu'), nltab,
       write('which lists all files of directory Sessions/. These files'), nltab,
       write('are the ones automatically produced by the s_sess command.'), nl.


man(set_current_program) :- !, man(set).
man(set) :- !,
    nl, write('set(P):'), nltab, 
    write('Moves to program P, that is,'), 
    write('P becomes the new current program.'), nl.

%man(load_th) :- !,
%    nl, write('load_th(T):'), nltab, 
%    write('adds all laws of theory T to current theory.'), nltab,
%    write('T  must be the name of a file in directory Theories/.'), 
%    nl, write('load_th:'), nltab, 
%    write('same as load_th(T) except that T is selected via a menu'), nltab,
%    write('which lists all files of directory Theories/.'), nl.

%man(load_l) :- !,
%    nl, write('load_l(T):'), nltab, 
%    write('Add one or more laws of theory T to current theory.'), nltab,
%    write('Laws are selected via a menu which lists all laws of theory T.'),
%    nltab,
%    write('T  must be the name of a file in directory Theories/.'),
%    nl, write('load_l:'), nltab, 
%    write('same as load_l(T) except that T is selected via a menu'), nltab,
%    write('which lists all files of directory Theories/.'), nl.
man(add_law_in_theory) :- !, man(add_l).
man(add_l) :- !,
   nl, write('add_l:'), nltab, 
   write('Adds a law to the current theory. The law is given interactively'),
   nltab, write('by the user.'), nl.

man(delete_law_from_theory) :- !, man(del_l).
man(del_l) :- !,
   nl, write('del_l:'), nltab, 
   write('Deletes a law from the current theory. The law is selected'), nltab,
   write('via a menu which lists all laws of the current theory.'), nl.

man(select_law_in_theory) :- !, man(sel_l).
man(sel_l) :- !,
    nl, write('sel_l:'), nltab, 
    write('Adds one or more laws from a theory T to the current theory.'), nltab,
    write('T is selected via a menu which lists all files'), nltab,
    write('of directory Theories/.'), nl.

man(select_theory) :- !, man(sel_t).
man(sel_t) :- !,
    nl, write('sel_t:'), nltab, 
    write('Adds theory T to the current theory.'), nltab,
    write('T is selected via a menu which lists all files'), nltab,
    write('of directory Theories/.'), nl.


man(show_program) :- !, man(p), man(p1).
man(show_programs) :- !, man(p).
man(p1) :- !,
    nl, write('p1:'), nltab, 
    nltab, write('Shows the initial program.'), nl.

man(p) :- !,
    nl, write('p:'), nltab, 
    nltab, write('Shows the current program.'), 
    nltab, write('See also command pp.'),
    nl, write('p(Ps):'), nltab, 
    nltab, write('Shows program(s) Ps. Ps may be a program number,'), nltab,
    nltab, write('or a list of program numbers.'),
    nltab, write('See also command pp(P).'), nl.


man(show_short_programs) :- !, man(pp).
man(pp) :- !,
    nl, write('pp:'), nltab, 
    nltab, write('Shows the current program.'), 
    nltab, write('Only clauses that are not in the initial program are shown.'),
    nl, write('pp(Ps):'), nltab, 
    nltab, write('Shows program(s) Ps. Ps may be a program number,'), nltab,
    nltab, write('or a list of program numbers.'),
    nltab, write('Only clauses that are not in the initial program are shown.'),
    nl.

man(show_history) :- !, man(h).
man(h) :- !,
    nl, write('h:'), nltab, 
    nltab, write('Shows history of the current program.'),
    nl, write('h(P):'), nltab, 
    nltab, write('Shows history of program P.'), nl.

man(show_clauses) :- !, man(cls).
man(cls) :- !,
    nl, write('cls(Cls):'), nltab,
    write('Shows clause(s) Cls. Cls may be a clause number or'), nltab,
    write('or a list of clause numbers.'), nl.

man(show_definitions) :- !, man(defs).
man(defs) :- !,
    nl, write('defs:'), nltab,
    write('Shows the definitions introduced so far.'), nl.

man(show_laws) :- !, man(l).
man(l) :- !,
    nl, write('l:'),
    nltab, write('Show laws in the current theory.'), nl.

man(directory_progs) :- !, man(dir_p).
man(dir_p) :- !,
    nl, write('dir_p:'),
    nltab, write('Shows the files contained in the directory Progs/.'), nl.

man(directory_final_progs) :- !, man(dir_f).
man(dir_f) :- !,
    nl, write('dir_f:'),
    nltab, write('Shows the files contained in the directory Final_Progs/.'), nl.

man(directory_sessions) :- !, man(dir_s).
man(dir_s) :- !,
    nl, write('dir_s:'),
    nltab, write('Shows the files contained in the directory Sessions/.'), nl.

man(directory_theories) :- !, man(dir_t).
man(dir_t) :- !,
    nl, write('dir_t:'),
    nltab, write('Shows the files contained in the directory Theories/.'), nl.


man(save_prog) :- !, man(s_p).
man(s_p) :- !,
    nl, write('s_p(P):'),
    nltab, write('Saves program P in the directory Final_Progs/.'),
    nl, write('s_p:'),
    nltab, write('Saves the current program in the directory Final_Progs/.'), nl.

man(save_history) :- !, man(s_h).
man(s_h) :- !,
    nl, write('s_h(P):'),
    nltab, write('Saves the history of program P in the directory Histories/.'),
    nl, write('s_h:'),
    nltab, write('Saves the history of the current program in the directory Histories/.'),
    nl.

man(save_current_theory) :- !, man(s_th).
man(s_th) :- !,
    nl, write('s_th:'),
    nltab, write('Saves the current theory in the directory Theories/.'),
    nl.

man(save_session) :- !, man(s_sess).
man(s_sess) :- !,
    nl, write('s_sess:'),
    nltab, write('Saves the current session in the directory Sessions/.'),
    nl.

man(insert_comment_history) :- !, man(com_h).
man(com_h) :- !,
   nl, write('com_h:'), nltab,
   write('Adds comments to the current history. Running this command'), nltab,
   write('gives a prompt to the user to add some comments.'), nltab,
   write('These comments will be integrated into the current history, that is,'),
   nltab, write(' they will appear in the text file produced by command s_h.'),
   nl.

man(helps) :- !, 
   nl, write('helps:'), nltab,
   write('Shows a general help menu, that is the following lines:'), nl, nltab,
   write('  st.            < start a new transformation process >'), nltab,
   write('  r.             < show available rules >'), nltab,
   write('  s.             < show available strategies >'), nltab,
   write('  c.             < show available other commands >'), nltab,
   write('  helps.         < show this five lines menu >'), nl.


man(rules) :- !, man(r).
man(r) :- !, 
   nl, write('r:'), nltab,
   write('Shows a menu of available transformation rules.'),
   nltab, write('See also command c and command s.'), nl.

man(commands) :- !, man(c).
man(c) :- !, 
   nl, write('c:'), nltab,
   write('Shows a menu of available commands.'),
   nltab, write('See also command r and command s.'), nl.

man(strategies) :- !, man(s).
man(s) :- !, 
   nl, write('s:'), nltab,
   write('Shows a menu of available strategies.'),
   nltab, write('See also command r and command c.'), nl.


man(man) :- !, 
   nl, write('man(X):'), nltab,
   write('Shows help about the use of rule, command, or strategy X.'),
   nl, write('Example:'), nltab,
   write('| ?- man(st)'), nltab,
   write('st:'), nltab,
   write('   Initialization of a transformation process by selecting'), nltab,
   write('   an initial program available in the directory Progs/'), nltab,
   write('st(P):'), nltab,
   write('   Initialization of a transformation process with P'), nltab,
   write('   as initial program.'), nltab,
   write('   P must be an existing file in the directory Progs/.'), nl.


man(lap) :- !,
    write('lap:'), nltab,
    nltab, write('This command is an implementation of the'),
    nltab, write('Loop Absorbtion strategy.'), nl,
    nltab, write('This strategy is defined in:'),
    nltab, write('Proietti, M. and Pettorossi, A.:'),
    nltab, write('The Loop Absorption and the Generalization Strategies for'),
    nltab, write('the Development of Logic Programs and Partial Deduction.'),
    nltab, write('Journal of Logic Programming, 16(1,2):123-161, 1993.'), nl.

man(uve) :- !,
    write('uve:'), nltab,
    nltab, write('This command is an implementation of the'),
    nltab, write('Unnecessary Variables Elimination strategy.'), nltab,
    write('This strategy is defined in: Proietti, M. and Pettorossi, A.:'),
    nltab, write('Unfolding-Definition-Folding, in this order,'),
    nltab, write('for Avoiding Unnecessary Variables in Logic Programs.'),
    nltab, write('Theoretical Computer Science 142(1):89-124, 1995.'), nl.


man(determ) :- !,
    write('determ:'), nltab,
    nltab, write('Implementation of a strategy for reducing nondeterminism.'),
    nltab, write('This strategy is defined in:'),
    nltab, write('Pettorossi, A., Proietti, M., Renault, S.:'),
    nltab, write('Reducing Nondeterminism while Specializing Logic Programs'),
    nltab, write('24th ACM Symposium on Principles of Programming Languages,'),
    nltab, write('Paris, France, January 15-17, 1997, p:414-427.'),
    nl.

man(X) :- 
    nltab, write('Sorry, no help for '), write(X), write('.'), nl.


man_(u) :- !, 
    nl, write('u(C,A):'),
    nltab, write('Unfolding (*) of clause C wrt atom A'),
    write(' using the initial program.'),
    nltab, write('C is a clause number in the current program.'),
    nltab, write('A is an atom number in the body of clause C.'),
    nl, write('u(C):'), 
    nltab, write('Short for u(C,1).').

man_(ud) :- !, 
    nl, write('ud(C,A):'),
    nltab, write('Unfolding (**) of clause C wrt atom A'),
    write(' using the initial program.'),
    nltab, write('C is a clause number in the current program.'),
    nltab, write('A is an atom number in the body of clause C.'),
    nl, write('ud(C):'),
    nltab, write('Short for ud(C,1)').

man_(ud_p) :- !, 
    nl, write('ud_p(C,A):'),
    nltab, write('Unfolding (**) of clause C wrt atom A'),
    write(' using the current program.'),
    nltab, write('C is a clause number in the current program.'),
    nltab, write('A is an atom number in the body of clause C.'),
    nl, write('ud_p(C):'),
    nltab, write('Short for ud_p(C,1)').

man_(ud_i) :- !, 
    nl, write('ud_i(Pi,C,A):'),
    nltab, write('Unfolding (**) of clause C wrt atom A'),
    write(' using program Pi.'),
    nltab, write('Pi is a program number, ancestor of the current program.'),
    nltab, write('C is a clause number in the current program.'),
    nltab, write('A is an atom number in the body of clause C.'),
    nl, write('ud_i(Pi,C):'),
    nltab, write('Short for ud_i(Pi,C,1)'),
    write(' using program Pi.'),nl.

man_u_ex :- nl,
   nl, write('Example:'), nltab,
   write('Assume that the current program is the following:'), nltab,
   write('1. p(A) :- q(A,B), r(B).'), nltab,
   write('2. q(a,b).'), nltab,
   write('3. q(b,a).'), nltab,
   write('4. r(a) :- q(A,B), q(B,A).'), nltab,
   write('5. r(b) :- q(A,A).'), nltab,
   write('6. r(c).'), nltab,
   write('| ?- u(1,2).'), nltab,
   write('by unfolding clause 1 w.r.t. atom 2 we get:'), nltab,
   write('7. p(A) :- q(A,a).'), nltab,
   write('8. p(A) :- q(A,b), q(B,B).'), nltab,
   write('9. p(A) :- q(A,c).'), nltab,
   write('| ?- u(8,2).'), nltab,
   write('no.'), nltab,
   write('| ?- ud(8,2).'), nltab,
   write('deletion of clause 8 because of failure in body'), nltab,
   write('| ?- u(4).'), nltab,
   write('by unfolding clause 4 w.r.t. atom 1 we get:'), nltab,
   write('10. r(a) :- q(b,a).'), nltab,
   write('11. r(a) :- q(a,b).'), nltab,
   write('| ?- ud_facts.'), nltab,
   write('deletion of clause 9'), nltab,
   write('by unfolding clause 10 w.r.t. atom(s) [1] we get:'), nltab,
   write('12. r(a).'), nltab,
   write('by unfolding clause 11 w.r.t. atom(s) [1] we get:'), nltab,
   write('13. r(a).'), nltab,
   write('deletion of clause 5'), nltab,
   write('| ?- '), nl.