% msg_user.pl
%
% This file contains some messages that may be given during a transformation
% session to the user. They might be simple comments on what the program currently
% does or more specific information to help the user answering to the program's
% requirement. Possible choices of interaction, expected types of entries, etc...

:- use_module(library(system), [ file_exists/1 ]).
:- use_module(library(lists), [ append/3 ]).

msg_help_rules :-
	nl, show_list_v([
'-------------------------- UNFOLDING RULES --------------------------',
' ',
'u/[0,1,2]        < unfold (*)  using initial program >',
'ud/[0,1,2]       < unfold (**) using initial program >',
'ud_p/[0,1,2]     < unfold (**) using current program >',
'ud_i/[0,2,3]     < unfold (**) using any intermediate program >',
'ud_facts/[0,1]   < unfold (**) wrt all instances of unit clauses >',
'ud_eqs/[0,1]     < unfold (**) wrt equalities >',
' ',
'(*)   The unfolding wrt an atom which does unify with no head',
'      of clause fails, (no clause deletion).',
'(**)  The unfolding of C wrt an atom whose predicate never occurs',
'      in the head of a clause fails. Otherwise, if it does unify with ',
'      no head of clause, it produces the deletion of C.',
' ', ' ',
'---------------------------- FOLDING RULES ---------------------------',
' ',
'ts_f/[0,3]       < fold one clause using a definition (Tamaki and Sato) >',
'f/2              < fold several clauses >',
'r_f/2            < fold several clauses with possible body rearragement >',
' ',' ',
'-------------------- DEFINITION INTRODUCTION RULES --------------------',
' ', ' ',
'ts_d   < introduce a new definition (Tamaki and Sato) >',
'd      < introduce a new definition (possibly several clauses) >',
' ',' ',
'---------------------- GOAL REPLACEMENT RULES -------------------------',
' ',
'rep/[0,3]        < goal replacement rule >',
'mv/[0,3]         < move of atoms in a clause body >',
'add_g/[0,1]      < add a goal in a clause body >',
'del_g/[0,1]      < delete a goal from a clause body >',
' ',' ',
'-------------------- CLAUSE REPLACEMENT RULES -------------------------',
' ',
'restr/1          < delete useless clauses wrt a given predicate >',
'del_cl/[0,1]     < delete clause(s) >',
'del_f_cl/[0,1]   < delete clause(s) with failing bodies >',
'add_cl           < add a clause to current program >',
'rep_cl/1         < replacement of a clause by another one >',
'subs/[0,1]       < apply subsumption >',
'split/2          < case split of a clause wrt an atom >',
'i/3              < instantiation of a clause wrt a variable and a type >',
' ',' ',
'------------------------------------------------------------------------',
' ',
'Type man(rule_name) for a detailed help about rule_name.',
' ',
' ']), nl.

msg_help_commands :- 
	   	nl, show_list_v([
'----------------- START AND NAVIGATE THE TRANSFORMATION TREE -------------',
' ',
'st/[0,1]   < start a transformation process >',
'load_s     < restore a session >',
'set/1      < move to a program in the transformation tree >', 
' ',' ',
'-------------------- SELECT AND EDIT REPLACEMENT LAWS --------------------',
' ',
'sel_l          < select a law from a theory and add it to current theory >',
'sel_t          < select a theory and add it to current theory >',
'add_l          < add a law to current theory >',
'add_l/1        < add a law to a theory present in directory Theories/>',
'del_l          < delete law from current theory >',
'del_l/1        < delete law from a theory present in directory Theories/>',
' ',' ',
'----------------------------- SHOW INFORMATION ----------------------------',
' ',
'p1        < show initial program >',
'p/[0,1]   < show program(s) >',
'pp/[0,1]  < show program(s) (new clauses only) >',
'cls/1     < show clause(s) >',
'h/[0,1]   < show history of a program >',
'defs      < show definitions introduced so far>',
'l         < show laws in current theory >',
'dir_p     < show files contained in directory Progs/ >',
'dir_t     < show files contained in directory Theories/ >',
'dir_s     < show files contained in directory Sessions/ >',
'dir_f     < show files contained in directory Final_Progs/ >',
' ',' ',
'----------------------------- SAVE INFORMATION ----------------------------',
' ',
's_p        < save current program in directory Final_Progs/ >',
's_h/[0,1]  < save history of a program in directory Histories/ >',
's_th       < save current theory in directory Theories/ >',
's_sess      < save current session in directory Sessions/  >',
' ',' ',
'----------------------------- EDIT INFORMATION ----------------------------',
' ',
'com_h     < add comments to current history >',
' ',' ',
'--------------------------------- HELPS -----------------------------------',
' ',' ',
'helps     < general help menu >',
'r         < show available transformation rules >',
's         < show available strategies >',
'c         < show other commands >',
' ',' ',
'------------------------------------------------------------------------',
' ',
'Type man(command_name) for a detailed help about command_name.',
' ',
' ']), nl.

msg_help_strategies :-
                 nltab, write('Available strategies: '), nl,
	   	   	show_list_v([ 
'uve             < unnecessary variables elimination >',
'lap             < loop absorption >',
'determ          < reduction of nondeterminism via specialization >' , 
' ',
'Type man(strategy_name) for a detailed help about strategy_name.'
] ) , nl.

msg_failure_no_selection:- nltab, write('No program has been selected.'), fail.

msg_not_a_pnat :- 
   nl, write('{INSTANTIATION ERROR: '),
   write('a natural number is expected}.'), nl.

msg_no_selection :- nltab, write('No selection.'), nl.

msg_no_action :- nltab, write('No action performed.'), nl.

msg_unchanged_after_failing_bodies :-
   nltab, write('Program unchanged.'), nl.

msg_unchanged_after_subsumption :-
   nltab, write('Program unchanged after subsumption.'), nl.

msg_unchanged_after_elimination(P/A) :-	
   nltab, write('No transient clauses for the definition of predicate '),
   write(P/A),
   write('.'), nltab, write('Current program unchanged.'), nl.

msg_unchanged_by_case_split :-
   nltab, write('Case split not possible, current program unchanged.'), nl.

msg_unknown_predicate(Pred) :-
   nltab, write('Warning: unknown predicate '), write(Pred), nl.

msg_unknown_program(P) :-
   nltab, write('Program '), write(P), write(' does not exist.'), nl.

msg_unknown_file(P) :-
   nltab, write('Unknown file: '), write(P), nltab,
   write('Enter "add_prog" to have the list of available programs. '), nl.

msg_unknown_clause(P) :-
   nltab, write('Clause '), write(P), write(' does not exist.'), nl.

msg_not_a_clause_in_current_prog(Clause_Number):-
   write(Clause_Number), write(' is not a clause of the current program.'), nl.

msg_already_used_in_prog(X):-
   write('{Waning: '), write(X),
   write(' is already used in the current program.}'), fail.

msg_already_used_in_def(X):-
   write('{Waning: '), write(X),
   write(' is already used in a definition.}'), fail.

msg_definition_clauses_empty :- nltab, write('No definition !'), nl.

msg_not_one_clause_def_for_ts_fold(Def) :-
	nltab, write(Def), write(' is not a definition clause of current program.'), nl.

msg_must_be_a_list_of_atoms(_) :-
   nltab, write('A list of integers is expected. '), nl.

msg_bad_positions_in_body :-
   nltab, write('Wrong positions for selected atoms ! '), nl.

msg_wrong_var_position(V,N) :-
   nl, write('{INSTANTIATION ERROR: '),
   write(V),
   write(' must be an integer in [1..'),
   write(N), write('].}'), nl, fail.

msg_must_have_same_length(Vs,Ts) :-
   nl, write('{INSTANTIATION ERROR: '),
   nl, write(Vs), write(' and '), write(Ts),
   write(' must be two lists of same length.}'), nl, fail.

msg_impossible_folding :- nltab, write('Impossible folding.'), nl.

msg_impossible_rep:-
   nl,
   write('Replacement not possible (no match), current program unchanged.'), nl.

msg_impossible_folding_1 :- 
   nltab, write('Impossible folding:'), nltab, 
   write('Expected arguments: two lists (clauses and definitions)'),
   write(' of same length.'), nl.

msg_impossible_folding_2(Defs) :- 
   nltab, write('Warning: clauses '), write(Defs),
   write(' must belong to an ancestor Pj'),
   nltab, write('of the current program such that: '), nltab,
   write('    - All heads of clauses '), write(Defs),
   write(' are unifiable to A'),
   nltab, write('    - No other head in Pj is unifiable to A'),	nl, nl,
   write('This condition is not satisfied.'), nltab,
   (yesno('Do you want to force the folding anyway', n) -> true
   ; (msg_no_action, fail)).

msg_neg_lit_not_enough_instantiated(C,Atom_Number) :-
   write('Negated atom at position '),
   write(Atom_Number), write(' in clause '), 
   (integer(C) -> write(C) ; (nl, show_clause(0,C), nl)),
   write(' is not instantiated enough.'),
   fail, !.

msg_show_existing_laws :- nl,
   (yesno('Do you want to see the laws already available', n) -> show_laws 
   ; true), nl.

msg_not_a_law :- write('This is not an expected entry.'), nl, msg_syntax_law.

msg_syntax_law :- 
   nltab, write('A law is a term of the form: A1,...,An -> B1,...,Bm'),
   nltab, write('where A1,...,An, B1,...,Bm are atoms.'), nl.

msg_laws_empty :- nltab, write('There is no law available! '), nl.

msg_no_law_in_theory :- nltab, write('No law in this theory !'), nl.

msg_save_file_laws(File_Theory) :-
   nltab, write(File_Theory), write(' has been updated.'), nltab,
   write('The previous version has been saved in old_'),
   write(File_Theory), nl.

msg_not_a_goal :- write('This is not an expected entry.'), nl.

msg_enter_name_program(File2) :-
   write('New file name :  Final_Progs/'),  get_line(S_File),
   (S_File =[] -> nl, write('Program not saved.'), nl, fail
   ; msg_enter_name_program_ok(S_File,File2)).

msg_enter_name_program_ok(S_File,File3):-
   my_last(S_File,Head_S_File,46), append("Final_Progs/",Head_S_File,File2),
   name(Name_File2,File2),
   (file_exists(Name_File2) -> (msg_choose_another_filename(Name_File2),
   msg_enter_name_program(File3)) ; File2=File3).

msg_choose_another_filename(File) :-
   nl, write('File '), write(File), write(' already exists. '),
   nl, write('Choose another name.').

msg_save_current_program(File) :-
   nltab, write('Program saved in '), write(File).
msg_save_current_theory(File) :- nltab, write('Theory saved in '), write(File).

msg_ask_save_history(File1,File2) :-
   write('The history will be saved in directory Histories in the file '),
   append("Histories/",Filename,File1),   name(Name_File1,Filename),
   write(Name_File1), nl,
   (yesno('Do you agree (enter "no" to give another file name) ', y) -> File1=File2
   ; msg_enter_name_history(File1,File2)).

msg_enter_name_history(File1,File2) :-
    nltab, write('New file name :  Histories/'), get_line(S_File),
    (S_File = [] -> File1=File2, msg_default_name_file 
    ; (my_last(S_File,Head_S_File,46), append("Histories/",Head_S_File,File2))).

msg_default_name_file :- 
    nltab, write('No selection -> default name is taken.'), nl.

msg_save_history(ProgN, File) :-
    nltab, name(FileProgN,ProgN),write('History of program '),
    write(FileProgN), write(' saved in '), write(File), nl.

msg_ask_save_session(File1,File2) :-
   write('The above session will be saved in directory Sessions in the file '),
   append("Sessions/",Filename,File1), name(Name_File1,Filename), 
   write(Name_File1), nl, 
   (yesno('Do you agree (enter "no" to give another name) ', y) -> File1=File2
   ; msg_enter_name_session(File1,File2)).

msg_enter_name_session(File1,File2) :-
    nltab, write('New file name :  Sessions/'), get_line(S_File),
    (S_File = [] -> File1=File2, msg_default_name_file 
    ; (my_last(S_File,Head_S_File,46), append("Sessions/",Head_S_File,File2))).

msg_save_current_session(File) :-
    nltab, write('Current session saved in '), write(File).

msg_not_a_clause :- write('This is not an expected entry.'), nl,
    write('Enter a term of the form "A :- B1,...,Bn.".'), nl.

msg_give_modes(P,N) :-
   nl, write('Give mode for '), write(P/N),
   write(' (+ or - for each argument):'), nl.

msg_no_available_progs :- 
   nltab, write('No available programs (stored in subdirectory Progs/) '), nl.

msg_no_available_final_progs :-
   nltab, write('No programs in subdirectory Final_Progs/) '), nl.

msg_no_available_theories :- 
   nltab,
   write('No available theories (stored in subdirectory Theories/) '), nl.

msg_no_available_sessions :-
   nltab,
   write('No available sessions (stored in subdirectory Sessions/) '), nl.

msg_undefined_predicate(Pred,Prog) :-  
   nl, write('Warning: predicate '), write(Pred), write(' is undefined'), 
   (Prog = 1 ->
       write(' in the initial program.') ; 
       (current_program(Prog) ->
        write(' in current program.') ; write(' in P'), write(Prog))), nl.

msg_undefined_predicate_in_clause(Pred,Prog,C) :-  
   nl, write('Warning: predicate '), write(Pred), write(' in clause '), 
   write(C), write(' is undefined'), 
   (Prog = 1 ->
       write(' in the initial program.') ; 
       (current_program(Prog) ->
        write(' in current program.') ; write(' in P'), write(Prog))), nl.

msg_not_in_ancestor([Cl]) :- !,	nltab,
   write('Clause '), write(Cl),
   write(' is not in an ancestor of current program.'), nl.
msg_not_in_ancestor(Cls) :- nltab,
   write('Clauses '), write(Cls),
   write(' are not in an ancestor of current program.'), nl.

msg_helps :-
        show_list_v([
        'st.            < start a new transformation process >',
        'r.             < show available rules >',
        's.             < show available strategies >',
        'c.             < show available other commands >']).

ask_enter_definition(H,B) :-
     nl, write('enter definition clause: '), nl, read(D2),
     check_clause(D2), rearrange_definition(D2, (H,B)).

ask_enter_first_definition_clause(D,Pred_Head,Arity) :-  nl,
   (yesno('Do you want to check the existing definitions', n) ->
      show_definitions ; true),
   write('enter clauses (use the same new predicate in the heads): '), nl,
   repeat,
   write('first clause: '), read_term(D2, [singletons(S)]), check_clause(D2),
   rearrange_definition(D2,(Head,Body)), D = (Head,Body),
   predname_and_arity(Head,Pred_Head,Arity),
   (check_new_pred(Head,Pred_Head,Arity) -> msg_singleton_variables(S) ;
   msg_choose_another_predicate_name, fail).

msg_singleton_variables([]) :- !.
msg_singleton_variables([A=_|S]) :-
   write('{Warning: ['), write(A), write('] - singleton variable.}'), nl,
   msg_singleton_variables(S).
msg_choose_another_predicate_name :-
   nl, write('Choose another predicate name.'), nl.

ask_enter_next_definition_clause(L,Functor_Head,Arity,Defs) :-
   (yesno('Another clause',n) ->
     (write('clause: '), read_term(D2, [singletons(S)]), check_clause(D2),
    rearrange_definition(D2,(Head,Body)), D = (Head,Body),
      (predname_and_arity(Head,Functor_Head,Arity) ->
        msg_singleton_variables(S), append(L,[D], L2) 
      ; (write('Not accepted definition clause: the head must use predicate '),
         write(Functor_Head/Arity), write('.'), L2=L)),
       nl, ask_enter_next_definition_clause(L2,Functor_Head,Arity,Defs))
      ; L=Defs).

ask_initial_position(Length,Pos) :-
   ask_number('initial position ',1,Length, 1,Pos).

ask_final_position(Length,Pos) :-
   ask_number('final position ',1,Length, 1,Pos).
ask_position_of_new_goal(Length,Pos) :-
   Length1 is Length + 1,
   ask_number('position of new goal',1,Length1, Length1,Pos).

ask_law_number([],_) :- msg_laws_empty, fail.
ask_law_number(Defs,Ans) :-
    ask_oneof('law number ',Defs,no,Ans),
   (Ans=[no] -> msg_failure_no_selection ; true).

ask_definition_number([],_) :- msg_definition_clauses_empty, fail.
ask_definition_number(Defs,Ans) :-
   ask_oneof('definition number ',Defs,no,Ans),
   (Ans=no -> msg_no_action ; true).

ask_enter_goal_to_add(L) :-
   nl, write('Goal: '), read(Goal), nl,
   check_goal(Goal), conj_to_list((Goal),L).

ask_enter_replacing_clause(C) :-
   nl, write('replacing clause: '), nl, read(C), check_clause(C), nl.

ask_enter_clause_to_add(C) :-
    nl, write('enter new clause: '), nl, read(C), check_clause(C), nl.

ask_enter_law(Law) :-
    nl, write('enter new law: '), nl, read(Law), nl.

ask_position_introduction(CNs,Pos) :-
   	ask_oneof('at which position ?',CNs,'at last position',Pos).

ask_clause_number(CNs,Ans) :-
   ask_oneof('clause number ',CNs,no,Ans),
   (Ans=[no] -> msg_failure_no_selection ; true).

ask_position_to_unfold(Length,Pos) :- ask_number('Number ',1,Length, 1,Pos).

ask_program_for_unfolding(PN,Ancestors,Pi) :-
   ask_oneof('Program number',Ancestors,PN,Pi).

ask_positions_in_body(Length,Atom_Numbers) :-
   ask_chars('atom numbers [no]',0,100,Answer),
   (Answer=[] -> msg_no_selection, fail 
   ; recover_positions(Answer,Atom_Numbers),
     check_input_atoms(Length,Atom_Numbers)).

ask_variable_positions(Length,L) :-
   ask_chars('variable number(s)',0,100,Answer),
   (Answer=[] -> L=[]
   ; recover_positions(Answer,Atom_Numbers),
     check_variable_numbers(Length,Atom_Numbers,L)).

check_variable_numbers(_,[],[]).
check_variable_numbers(L,[V|Vars],[V|Vars2]) :-
    integer(V), V > 0, V =< L, !,
    check_variable_numbers(L,Vars,Vars2).
check_variable_numbers(L,[V|Vars],Vars2) :-
    nl, write('bad position: '), write(V), write(' (omitted.)'), nl,
    check_variable_numbers(L,Vars,Vars2).

recover_positions(Answer,Pos) :- recover_positions([],Answer,Pos).
recover_positions([],[32|Answer],Pos) :-!, recover_positions([],Answer,Pos).
recover_positions(Pref,[32|Answer],[Pos|Pos1]) :- !,
   name(Pos,Pref), recover_positions(Answer,Pos1).
recover_positions(Pref,[44|Answer],[Pos|Pos1]) :- !,
   name(Pos,Pref), recover_positions(Answer,Pos1).
recover_positions(Pref,[C|Answer],Pos) :-
   49 =< C, C =< 57, !, append(Pref,[C],Pref2),
   recover_positions(Pref2,Answer,Pos).
recover_positions(Pref,[_|Answer],Pos) :- recover_positions(Pref,Answer,Pos).
recover_positions([],[],[]) :- !.
recover_positions(Pref,[],[Pos]) :- name(Pos,Pref).

help_trace_session :-
     nl, tab(5), write('Available choices:'), nl,
     nl, tab(5), write('p -> show current program'), nl,
     nl, tab(5), write('d -> show current defs'), nl,
     nl, tab(5), write('c -> insert comment in history'), nl,
     nl, tab(5), write('o -> trace session off'), nl.

msg_user_successful_folding(Pi,A):-
   functor(A,NameDef,_), write('Successful folding using definition '),
   write(NameDef), (write(' in P'), write(Pi)).

msg_derivation_stopped :-
   nl, write('Derivation stopped. Depth bound exceeded. '),
   nl,    write('The program derived so far is the following:'), nl, pp.

msg_failure_on_unfolding(E) :-
   nl, write('Unfolding of clause '), write(E),
   write(' fails because the leftmost atom is a failing goal.'), nl.

msg_determ_strat_fails :-
   nl, write('Failure of the strategy. '), nl,
   write('The program derived so far is the following:'), nl, pp.

msg_derivation_fails(E) :-
   msg_no_clause_deletion(E), write('The whole derivation fails.'), fail.

msg_no_clause_deletion(E) :-
   nltab, write('No deletion of clause '), write(E), write('. ').

msg_does_not_contain_unnecessary_var(C) :-
   nl, write('clause '), write(C), write(' has no unnecessary variable.'), nl.
msg_is_not_subject_to_folding(C) :-
   nl, write('clause '), write(C), write(' is not subject to folding.'), nl.
msg_does_not_contain_existential_var(C) :-
   nl, write('clause '), write(C), write(' has no existential variable.'), nl.

% Unfolding and simplification of (in)equalities
new_clauses(unfolding_and_simpl,Cls,CNs2) :-
   new_clauses(CNs2,NewCls),
   nl,  write('by unfolding and simplifying clauses '),
   write(Cls), write(' we get:'), nl, show_numb_clauses(NewCls).


% Tupling
new_clauses(tupling,Cls,Cls2) :-
   new_clauses(Cls2,NewCls),
   nl,  write('by tupling in '),
   write(Cls), write(' we get:'), nl, show_numb_clauses(NewCls).

% Introduction of continuations
new_clauses(introduce_continuations,Cls,Cls2) :-
   new_clauses(Cls2,NewCls),
   nl,  write('by introduction of continuations on '),
   write(Cls), write(' we get:'), nl, show_numb_clauses(NewCls).

% Folding of a conjunction or a disjunction of continuations
new_clauses(fold_continuations,C,Cs2) :-
   new_clauses(Cs2,NewCls),
   nl,  write('by folding continuations in '),
   write(C), write(' we get:'), nl, show_numb_clauses(NewCls).

% Case-split (global analysis)
new_clauses(case_split,Cl,Cls) :- 
   integer(Cl), !, new_clauses(Cls,NewCls), nl,
   write('by case splitting clause '),
   write(Cl), write(' we get:'), nl, show_numb_clauses(NewCls).

% Case-split (global analysis)
new_clauses(case_split,Cls,Ps) :- nl,
   write('by splitting clauses '), write(Cls), write(' we get:'), nl,
   show_packets(Ps).

%unfold/4
new_clauses(unfold,C,A,[]) :- !, nl,
   write('Deletion of clause '), write(C),
   write(' by unfolding it w.r.t. atom '), write(A).

new_clauses(unfold,C,A,CN) :- 
   new_clauses(CN,NewCls), nl, write('by unfolding clause '), write(C),
   write(' w.r.t. atom '), write(A), write(' we get:'), nl,
   show_numb_clauses(NewCls).

new_clauses(unfold_d,C,A,[]) :- !, nl,
   write('Deletion of clause '), write(C),
   write(' by unfolding it w.r.t. atom '), write(A).

new_clauses(unfold_d,C,A,CN) :- 
   new_clauses(CN,NewCls), nl, write('by unfolding clause '), write(C),
   write(' w.r.t. atom '), write(A), write(' we get:'), nl,
   show_numb_clauses(NewCls).

new_clauses(unfold_d_p,C,A,[]) :- !, nl,
   write('Deletion of clause '), write(C),
   write(' by unfolding it w.r.t. atom '), write(A).

new_clauses(unfold_d_p,C,A,CN) :- nl,
   new_clauses(CN,NewCls), write('by unfolding clause '), write(C),
   write(' w.r.t. atom '), write(A), write(' using current program'),
   write(' we get:'), nl, show_numb_clauses(NewCls).

new_clauses(fold,Cs,Def_Numbers,CN) :- 
   CN=[N], clause3(N,Cl,_), nl, write('by folding clause '), write(Cs),
   write(' using '), write(Def_Numbers), write(' we get:'), nl,
   show_numb_clauses([(N,Cl)]).


% Case-split (binary)
new_clauses(case_split,C,C2,Cls) :- 
   new_clauses(Cls,NewCls), nl, write('by case splitting clause '), write(C),
   write(' w.r.t. clause '), write(C2), write(' we get:'), nl,
   show_numb_clauses(NewCls).

%unfold/5
new_clauses(unfold_d_i,Pi,C,A,[]) :- !, nl,
   write('Deletion of clause '), write(C),
   write(' by unfolding it w.r.t. atom '), write(A), nl,
   write(' using program '), write(Pi).

new_clauses(unfold_d_i,Pi,C,A,CN) :- 
   new_clauses(CN,NewCls), nl, write('by unfolding clause '), write(C),
   write(' w.r.t. atom '), write(A), write(' using program '), write(Pi),
   write(' we get:'), nl, show_numb_clauses(NewCls).

new_clauses(instantiation,C,[V,V1|Vs],[T,T1|Ts],Cls) :-  !,
   new_clauses(Cls,NewCls), nl, write('by instantiating clause '), write(C),
   write(' w.r.t. the variables '), 
   write([V,V1|Vs]), write(' with types '), write([T,T1|Ts]), 
   write(' we get:'),
   nl, show_numb_clauses(NewCls).

new_clauses(instantiation,C,[V],[T],Cls) :-  !,
   new_clauses(instantiation,C,V,T,Cls).

new_clauses(instantiation,C,V,T,Cls) :- 
   new_clauses(Cls,NewCls), nl, write('by instantiating clause '), write(C),
   write(' w.r.t. the '), write(V),
   (V=1 -> write('st') ; (V=2 -> write('nd') ; (V=3 -> write('rd') ; 
   write('th')))), write(' variable and the type '), write(T), 
   write(' we get:'),
   nl, show_numb_clauses(NewCls).

% new_clauses(fold,C,As,Def_Number,CN): it 
% shows the new clause CN obtained by a folding step.
new_clauses(fold,C,As,Def,CN) :- 
   CN=[N], clause3(N,Cl,_), nl, write('by folding clause '), write(C),
   write(' w.r.t. atoms '), write(As), write(' using '), write(Def),
   write(' we get:'), nl, show_numb_clauses([(N,Cl)]).


new_clauses([],[]).
new_clauses([N|Ns],[(N,C)|L]) :-  clause3(N,C,_), new_clauses(Ns,L).

new_clause(delclause,C) :- nl, write('deletion of clause '), write(C), nl.
new_clause(deltautology,C) :- nl, write('deletion of clause '), write(C), 
   write(' (tautology)').
new_clause(ignoretautology,C) :- nl, write('clause '), write(C), 
   write(' is a tautology.').
new_clause(delboolsimpl,C) :- nl, 
   write('boolean simplification in the body of clause'), write(C).
new_clause(elimclauses,Cls) :-
   nl, write('elimination of transient clause(s) '), write(Cls).
new_clause(delclause_failure_body,C) :-
   nl,write('deletion of clause '), write(C),
   write(' because of failure in body').
new_clause(addclause,NewCN) :-
   nl, NewCN=[N], write('by adding clause '), write(N),
   write(' we get the new program: '),  nl, show_current_program.
new_clause(moveclause,N) :-
   nl, write('by moving clause '), write(N),
   write(' we get the new program: '),  nl,  show_current_program.

new_clause(repclause,C,C2) :- nl, write('replacement of clause '), write(C),
     write(' by clause '), write(C2).

new_clause(unfold_facts,C,As,NewCN) :-
   NewCN=[N], clause3(N,Cl,_), nl, write('by unfolding clause '), write(C),
   write(' w.r.t. atom(s) '), write(As), write(' we get:'), nl,
   show_numb_clauses([(N,Cl)]).
new_clause(simpl_eq,C,As,NewCN) :-
   NewCN=[N], clause3(N,Cl,_), nl, write('by simplifying (in)equality(ies): '),
   write(As), write(' in clause '), write(C), write(' we get:'),  nl,
   show_numb_clauses([(N,Cl)]).
new_clause(ud_eq,C,As,NewCN) :-
   NewCN=[N], clause3(N,Cl,_), nl, write('by unfolding clause '), write(C),
   write(' w.r.t. (in)equalities): '), write(As), write(' we get:'),
    nl, show_numb_clauses([(N,Cl)]).
new_clause(delgoal,C,As,NewCN) :-
   nl, NewCN=[N], clause3(N,Cl,_), write('by deleting goal '), write(As),
   write(' in clause '), write(C), write(' we get:'), nl,
   show_numb_clauses([(N,Cl)]).

new_clause(addgoal,C,Pos,L,NewCN) :-
   nl, NewCN=[N], clause3(N,Cl,_), write('by adding goal '), numbervars(L,0,_),
   show_goal(L), write(' in clause '), write(C), write(' at position '),
   write(Pos), write(' we get:'), nl, show_numb_clauses([(N,Cl)]).
new_clause(replace_goal,C,OldGoal,Law,CN) :- 
   nl, CN=[N], clause3(N,(H,B),_), law(Law,(L,R)), OldGoal=L,
   numbervars((OldGoal,R),0,_), nl,
   write('by replacing goal '), write(OldGoal), nl, write('by '), write(R), nl,
   write('in clause '), write(C), write(' using '), write('law '), write(Law),
   write(' we get:'), nl, show_numb_clauses([(N,(H,B))]).
new_clause(moveatom,C,Pos1,Pos2,CN) :- 
   CN=[N], clause3(N,Clause,_), nl, write('by moving atom '), write(Pos1),
   write(' to position '), write(Pos2), write(' in clause '), write(C),
   write(' we get:'), nl, show_numb_clauses([(N,Clause)]).




/*-------
write_inf/[1,3] shows the history.
--------*/
write_inf(1) :-  write('P'), write('1: '), file_input(F), write(F).
write_inf(P,(C,(unf,A),_),P1) :-
   write('P'), write(P1), write(': '), write('by unfolding '), write('cl '),
   write(C), 
   write(' in P'), write(P), write(' wrt atom '), write(A), write(' using P1').
write_inf(P,(C,(uunf,A),_),P1) :-
   write('P'), write(P1), write(': '), write('by unfolding '), write('cl '),
   write(C),
   write(' in P'), write(P), write(' wrt atom '), write(A), write(' using P1').
write_inf(P,(C,(uunf_i,P_j,A),_),P1) :-
   write('P'), write(P1), write(': '), write('by unfolding '), write('cl '),
   write(C),
   write(' in P'), write(P), write(' wrt atom '), write(A), write(' using P'), write(P_j).
write_inf(P,(C,(unfold_facts,As),_),P1) :- 
   write('P'), write(P1), write(': '), write('by unfolding '), write('cl '),
   write(C),
   write(' in P'), write(P), write(' wrt atoms: '), write(As), write(' using P1').
write_inf(P,(C,(simpl_eq,As),_),P1) :- 
   write('P'), write(P1), write(': '), write('by simplifying in '),
   write('cl '),
   write(C), write(' in P'), write(P), write(' (in)equalities :'), write(As).
write_inf(P,(C,(ud_eq,As),_),P1) :- 
   write('P'), write(P1), write(': '), write('by unfolding '), write('cl '),
   write(C),
write(' in P'), write(P), write(' wrt (in)equalities: '), write(As).
write_inf(P,(C,(fold,As,D),_),P1) :- 
   write('P'), write(P1), write(': '), write('by folding '), write('cl '),
   write(C),
   write(' in P'), write(P), write(' w.r.t atoms '), write(As),
   write(' using '), write(D).
write_inf(P,([C],(fold,[Def]),_),P1) :- !,
   write('P'), write(P1), write(': '), write('by folding '), write('cl '),
   write(C), write(' in P'), write(P), write(' using cl '), write(Def).
write_inf(P,(Clauses_Number,(fold,Def),_),P1) :- 
   write('P'), write(P1), write(': '), write('by folding '), write('cls '),
   write(Clauses_Number),
   write(' in P'), write(P), write(' using cls '), write(Def).
write_inf(P,(C,(case_split,Cl_Number),_),P1) :-
   write('P'), write(P1), write(': '), write('by case splitting '),
   write('cl '),
   write(C), write(' in P'), write(P), write(' wrt cl '), write(Cl_Number).
write_inf(P,(Cl,case_split,_),P1) :-
   integer(Cl), write('P'), write(P1), write(': '),
   write('by case splitting '),
   write('cl '), write(Cl), write(' in P'), write(P).
write_inf(P,(Cls,case_split,_),P1) :-
   write('P'), write(P1), write(': '), write('by case splitting '),
   write('cls '), write(Cls), write(' in P'), write(P).

write_inf(P,(C,(instantiation,_,_),_),P1) :-
   write('P'), write(P1), write(': '), write('by instantiating cl '), 
   write(C), write(' in P'), write(P).

write_inf(P,(C,(rep,As,Law),_),P1) :-
   write('P'), write(P1), write(': '), write('by replacing '), 
   write('atoms '), write(As), write(' in '), write('cl '), write(C),
   write(' in P'), write(P), write(' using law '), write(Law).
write_inf(P,(def,[NewCN]),P1) :-  
   write('P'), write(P1), write(': '), write('by adding def '),
   write(NewCN), write(' to P'), write(P).
write_inf(P,(gen_def,[NewCN]),P1) :- !, 
   write('P'), write(P1), write(': '), write('by adding def '),
   write(NewCN), write(' to P'), write(P).
write_inf(P,(gen_def, L), P1) :-
   write('P'), write(P1), write(': '), write('by adding defs '),
   write(L), write(' to P'), write(P).
write_inf(_P,(synth, L), P1) :-
   write('P'), write(P1), write(': '), write('by synthesizing clauses '),
   write(L).
write_inf(P,(C,(delgoal,As),_),P1) :-
   write('P'), write(P1), write(': '), write('by deleting goal '),
   write(As), write(' in cl '), write(C), write(' in P'), write(P).
write_inf(P,(C,(addgoal,_,G),_),P1) :-    write('P'), write(P1), write(':
'), write('by adding goal '),    numbervars(G,0,_),
   show_body(G), write(' in cl '), write(C), write(' in P'), write(P).
write_inf(P,(addclause,[N]),P1) :- 
   write('P'), write(P1), write(': '),write('by adding cl '),
   write(N), write(' in P'), write(P). 
write_inf(P,(_,addclause,[N]),P1) :- !,
   write('P'), write(P1), write(': '), write('by adding cl '),
   write(N), write(' in P'), write(P). 
write_inf(P,(_,addclause,Ns),P1) :- 
   write('P'), write(P1), write(': '), write('by adding cls '),
   write(Ns), write(' in P'), write(P). 
write_inf(P,(moveclause,[N]),P1) :- 
   write('P'), write(P1), write(': '), write('by moving cl '),
   write(N), write(' in P'), write(P). 
write_inf(P,(NN,repclause,[N]),P1) :- 
   write('P'), write(P1), write(': '), write('by replacing cl '),
   write(NN), write(' by cl '), write(N), write(' in P'), write(P). 
write_inf(P,(N,loopabs),P1) :- 
   write('P'), write(P1), write(' : '),
   write('by loop absorption strategy from clause '),
   write(N), write(' in P'), write(P). 
write_inf(P,(N,loopabs),P1) :- 
   write('P'), write(P1), write(': '), write('by generalization strategy, to eliminate '),
   nl, write('unnecessary variables in clause '), write(N),
   write(' of P'), write(P). 
write_inf(P,(C,delclause,[]),P1) :-
   write('P'), write(P1), write(': '), write('by deleting cl '),
   write(C), write(' in P'), write(P).
write_inf(P,(C,deltautology,[]),P1) :-
   write('P'), write(P1), write(': '), write('by deleting cl '),
   write(C), write(' (tautology) in P'), write(P).
write_inf(P,(C,delboolsimpl,[]),P1) :-
   write('P'), write(P1), write(': '), write('by deleting cl '),
   write(C), write(' (boolean simplification) in P'), write(P).
write_inf(P,(Cls,elimclauses,[]),P1) :-
   write('P'), write(P1), write(': '), write('by deleting transient clauses '),
   write(Cls), write(' in P'), write(P).
write_inf(P,(C,(moveatom,N,M),_),P1) :-
   write('P'), write(P1), write(': by moving atom from position '), write(N),
   write(' to position '), write(M), write(' in cl '), write(C),
   write(' in P'), write(P).
write_inf(P,(C,fold_continuations,_),P1) :-
   write('P'), write(P1), write(': by folding continuations in the body of cl '), 
   write(C), write(' in P'), write(P).
write_inf(P,(C,tupling,_),P1) :-
   write('P'), write(P1), write(': by folding (tupling) cl '), 
   write(C), write(' in P'), write(P).
write_inf(P,(Cs,unfolding_and_simpl,_),P1) :-
   write('P'), write(P1), write(': '), write('by unfolding* and simplify '), write('cls '),
   write(Cs), 
   write(' in P'), write(P).


