:- op(950,xfy, [or]).
:- op(900,xfy, [and]).
:- dynamic strat_uf/1.


go :- ['src_sics/new_uf_strategy_v2.pl'].
bu :- assert(strat_uf(binary_folding)), assert(strat_uf(tuppling)).
bf :- assert(strat_uf(binary_folding)).
tu :- assert(strat_uf(tuppling)).
no :- retractall(strat_uf(_)).
list :- listing(clauses_of_new_definition/2).

:- dynamic     clauses_of_new_definition/2, strat_prove/1.

prove :-
      statistics(runtime,[_,T1]),
      prove(_),
      statistics(runtime,[_,T2]), nl,
      T is T2-T1, write('Elapsed time '), write(T), write(ms), nl.
prove(Phi) :-
      check_exists_transf_tree,
      retractall(clauses_of_new_definition(_,_)),
      (current_new_def_functor(N) -> PredFormula is N+1 ; PredFormula=1),
      lloyd_topor_and_rearrangement(Phi,Defs),
      reverse(Defs,RDefs),
      apply_uf(RDefs), !,
      append(RDefs,L),
      update_transformation_tree_after_uf_strat(L),
      msg_theorem_proved(PredFormula).
prove_a(Phi) :-
      (current_new_def_functor(N) -> PredFormula is N+1 ; PredFormula=1),
      lloyd_topor_and_rearrangement(Phi,Defs),
      reverse(Defs,RDefs),
      apply_uf(RDefs), !,
      append(RDefs,L),
      update_transformation_tree_after_uf_strat(L),
      %L is a list of clause numbers
      msg_theorem_proved(PredFormula).

apply_uf([S1|Ss]) :-
   apply_uf_stratum(S1),
   apply_uf(Ss).
apply_uf([]).


apply_uf_stratum(Def1) :-
   fetch_clauses(Def1,Cls1),
   assert(clauses_of_new_definition(Def1,Cls1)),
   apply_uf_one_definition(Def1).

apply_uf_one_definition([]) :- !.
apply_uf_one_definition(S1) :-
   unfolding_steps_stratum(S1,U1),
   simplification_steps_stratum(U1,LPackets),
   update_clauses_after_unfolding(S1,LPackets),
   (strat_uf(binary_folding) ->
       (   binary_folding_steps_stratum(LPackets,F1,NewDefs2),
           append(NewDefs2,NewDefs)) 
      ; mult_folding_steps_stratum(LPackets,F1,NewDefs)),
   apply_uf_sub_stratum(NewDefs,F2),
   append(F1,F2,F),
   simplification_steps_after_folding(F,NewF),
   update_clauses_after_simplification(NewF).

apply_uf_one_sub_definition(S1,F) :-
   unfolding_steps_stratum(S1,U1),
   simplification_steps_stratum(U1,LPackets),
   update_clauses_after_unfolding(S1,LPackets),
   (strat_uf(binary_folding) ->
       (   binary_folding_steps_stratum(LPackets,F1,NewDefs2),
           append(NewDefs2,NewDefs)) 
     ; mult_folding_steps_stratum(LPackets,F1,NewDefs)),
   apply_uf_sub_stratum(NewDefs,F2),
   append(F1,F2,F).

apply_uf_sub_stratum([],[]).    
apply_uf_sub_stratum([Defs|LDefs],F) :-
   apply_uf_one_sub_definition(Defs,F1),
   apply_uf_sub_stratum(LDefs,F2),
   append(F1,F2,F).

update_clauses_after_unfolding(S1,LPackets) :-
   retract_clauses(S1,_),
   append(LPackets,UClauses),
   add_clauses(_,UClauses,temp).

update_clauses_after_simplification(NewF) :-
   retractall(clause3(_,_,temp)),   
   add_clauses(_,NewF,uf).


%    FOLDING STRATEGY

binary_folding_steps_stratum([],[],[]).
binary_folding_steps_stratum([P|Ps],Fs,[NewDefs1|NewDefs]) :-
   apply_subsumption_clauses_f(P,[],P2),
   binary_folding_steps_stratum_one_packet(P2,Fs1,NewDefs1),
   binary_folding_steps_stratum(Ps,Fs2,NewDefs),
   append(Fs1,Fs2,Fs).

mult_folding_steps_stratum([],[],[]).
mult_folding_steps_stratum([P|Ps],Fs,[NewDefs1|NewDefs]) :-
   apply_subsumption_clauses_f(P,[],P2),
   mult_folding_steps_stratum_one_packet(P2,Fs1,NewDefs1),
   mult_folding_steps_stratum(Ps,Fs2,NewDefs),
   append(Fs1,Fs2,Fs).

mult_folding_steps_stratum_one_packet([],[],[]) :- !.
mult_folding_steps_stratum_one_packet(P,P,[]) :-
  folding_unnecessary(P), !.
mult_folding_steps_stratum_one_packet(P,NewCls,NewDefs) :-
  (try_to_fold_with_defs2(P,NewClause) -> NewDefs=[] ;
   introduce_new_def_and_fold_one_packet(P,NewDefs,NewClause)),
   NewCls=[NewClause].

apply_subsumption_clauses_f([],L,L).
apply_subsumption_clauses_f([E|L],I,F) :-
   is_subsumed_2(E,L),!,
   apply_subsumption_clauses_f(L,I,F).
apply_subsumption_clauses_f([E|L],I,F) :-
   is_subsumed_2(E,I),!, 
   apply_subsumption_clauses_f(L,I,F).
apply_subsumption_clauses_f([E|L],Inter,F) :-
   append(Inter,[E],Inter2), apply_subsumption_clauses_f(L,Inter2,F).

binary_folding_steps_stratum_one_packet([C],NewCls,[NewDefs]) :- !,
    mult_folding_steps_stratum_one_packet([C],NewCls,NewDefs).
				       
binary_folding_steps_stratum_one_packet(P,NewCls,NewDefs) :-
    binary_folding_steps_stratum_one_packet_one_cl(P,NewP,NewDefs1),
    binary_folding_steps_stratum_one_packet_two_cls(NewP,NewCls,NewDefs2),
    append(NewDefs1,NewDefs2,NewDefs).

binary_folding_steps_stratum_one_packet_two_cls([C],[C],[[]]).
binary_folding_steps_stratum_one_packet_two_cls([C1,C2],Cs,[NewDef1]) :-!,
   mult_folding_steps_stratum_one_packet([C1,C2],Cs,NewDef1).
binary_folding_steps_stratum_one_packet_two_cls([C1,C2|P],Cs,NewDefs) :-
   binary_folding_steps_stratum_one_packet_two_cls(P,NewCs,NewDefs2),
   mult_folding_steps_stratum_one_packet([C1,C2],[C],NewDef1),
   binary_folding_steps_stratum_one_packet_two_cls([C|NewCs],Cs,NewDefs3),
   append(NewDefs2,[NewDef1|NewDefs3],NewDefs).

binary_folding_steps_stratum_one_packet_one_cl([],[],[]).
binary_folding_steps_stratum_one_packet_one_cl([C1|P],[NewCl1|NewCls],[NewDef1|NewDefs]) :-
   mult_folding_steps_stratum_one_packet([C1],[NewCl1],NewDef1),
   binary_folding_steps_stratum_one_packet_one_cl(P,NewCls,NewDefs).

introduce_new_def_and_fold_one_packet(P,NewDefsN,NewClause) :-
   gen_new_name_functor(X), anti_unify_heads(P,Head,U),
   Head =.. [_|L2], new_defs_general_case(U,X,L2,NewDefs2),
   rearrange_definitions(NewDefs2,NewDefs3),
   try_to_drop_arguments(NewDefs3, NewDefs),
   max_clause(N),
   apply_gen_define2_l_f(N,[NewDefs],[NewDefsN]),
   infer_types(NewDefs),
   fold_cond2_whole_bd_f(P,NewDefs,_,_,NewClause), !.

folding_unnecessary([]).
folding_unnecessary([(H,B)]) :-
   contains_no_unnecessary_var2((H,B)),
   is_a_conjunction_of_atoms_with_vars_only(B).

try_to_fold_with_defs2(Clauses,NewCN) :-
    same_length(Clauses,LDefsN),
    clauses_of_new_definition(LDefsN,LDefs),
    try_to_fold_with_defs2(Clauses,NewCN,LDefs).

try_to_fold_with_defs2(Clauses,NewClause,Defs):-
        fold_cond2_whole_bd_f(Clauses,Defs,_,_,NewClause),
        contains_no_unnecessary_var2(NewClause).

fold_cond2_whole_bd_f(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses_f(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   fold_cond2_whole_bd_f_(Clauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses_f(Folded,Units,DefUnits).

fold_cond2_whole_bd_f_(Cls, Defs, Gen_Head, A, Folded):-
    (var(Gen_Head) -> heads_unifiable_f(Cls,_,Gen_Head) ; true),
    (var(A) -> heads_unifiable_f(Defs,_,A) ; true),
    minimal_hd_generalization_clauses_f(Cls,Gen_Head,_,Gen_Cl),
    minimal_hd_generalization_clauses_f(Defs,A,_,Gen_Def),
    recover_folded_clause_whole_bd(Gen_Cl,Gen_Def, Folded).

verify_recover_unit_clauses_f(_,[],[]).
verify_recover_unit_clauses_f(Folded,[CU|Units],[(H,[])|DefUnits]) :-
    copy_term(Folded,Folded2), copy_term(H,H2), unfold_body(Folded2,H2, CU),
    verify_recover_unit_clauses_f(Folded,Units,DefUnits).

minimal_hd_generalization_clauses_f(Cls,Gen_Head,Gen_Cls,Dis_Cl) :-
    Gen_Head=..[_|Args_Gen_Head],
    minimal_hd_generalization_clauses_f(Cls,Args_Gen_Head,Gen_Cls),
    disjunctive_clause(Gen_Cls,Dis_Cl).
minimal_hd_generalization_clauses_f([],_,[]).
minimal_hd_generalization_clauses_f([(Hd,Bd)|P],Gen_Head,[C|Cs]) :-
    Hd=..[F|Args],
    minimal_move_arguments_in_body(Args,Gen_Head,GenArgs,Eqs),
    GenHd=..[F|GenArgs], append(Eqs,Bd,ExtBody),
    C=(GenHd,ExtBody),
    minimal_hd_generalization_clauses_f(P,Gen_Head,Cs).

heads_unifiable_f([],H,H).
heads_unifiable_f([(Hd_C,_)|L],H,H3) :-
   (var(H) -> H2 = Hd_C ; anti_unify(Hd_C,H,H2)), heads_unifiable_f(L,H2,H3).

omit_first_unit_clauses_f([],[],[],[],[],[]).
omit_first_unit_clauses_f([C|Cls],[D|Defs],[C|Units],[D|DefUnits],Cls2,Defs2) :-
    D=(_,[]), !,
    omit_first_unit_clauses_f(Cls,Defs,Units,DefUnits,Cls2,Defs2).
omit_first_unit_clauses_f([C|Cls],[D|Defs],Units,DefUnits,[C|Cls2],[D|Defs2]) :-
    omit_first_unit_clauses_f(Cls,Defs,Units,DefUnits,Cls2,Defs2).

%    SIMPLIFICATION STRATEGY

simplification_steps_after_folding(F_i2,NewF_i_R) :-
   reverse(F_i2,F_i2_R),
   (strat_prove(apply_partial_instantiation) ->
    simplification_steps_after_folding_non_mut_ex(F_i2_R,NewF_i) ;
    simplification_steps_after_folding_mut_ex(F_i2_R,NewF_i)),
   reverse(NewF_i,NewF_i_R).

simplification_steps_after_folding_non_mut_ex(Cls,NewF_i) :-
   member_rest((Hd,Bd),Cls,Rest),
   ((Bd=[X], X==Hd) ->
       ((member((Hd2,_),Rest), \+ \+ unify_with_occurs_check(Hd,Hd2)) ->
             simplification_steps_after_folding_non_mut_ex(Rest,NewF_i) ;
           (propagate_false_in_clause_bodies(Hd,Rest,Cls2), 
            simplification_steps_after_folding_non_mut_ex(Cls2,NewF_i2),
            append([(Hd,[false])],NewF_i2,NewF_i))) ;
   (atom(Hd) ->
      (Bd = [] ->
	   ( propagate_true_in_clause_bodies(Hd,Rest,Cls2), 
            simplification_steps_after_folding_non_mut_ex(Cls2,NewF_i2),
            append([(Hd,[])],NewF_i2,NewF_i))
    ; (Bd=[false] ->
	  (  member((Hd,_),Rest) ->
             simplification_steps_after_folding_non_mut_ex(Rest,NewF_i) ;
           (propagate_false_in_clause_bodies(Hd,Rest,Cls2), 
            simplification_steps_after_folding_non_mut_ex(Cls2,NewF_i2),
            append([(Hd,[false])],NewF_i2,NewF_i))))))).

simplification_steps_after_folding_non_mut_ex(Cls,Cls).

simplification_steps_after_folding_mut_ex(Cls,NewF_i) :-
   member_rest((Hd,Bd),Cls,Rest),
   ((Bd=[X], X==Hd) -> 
    (propagate_false_in_clause_bodies(Hd,Rest,Cls2), 
    simplification_steps_after_folding_mut_ex(Cls2,NewF_i2),
    append([(Hd,[false])],NewF_i2,NewF_i)) ;
   (atom(Hd) ->
      (Bd = [] -> propagate_true_in_clause_bodies(Hd,Rest,Cls2) 
    ; (Bd=[false] -> propagate_false_in_clause_bodies(Hd,Rest,Cls2))),
    simplification_steps_after_folding_mut_ex(Cls2,NewF_i2),
    append([(Hd,Bd)],NewF_i2,NewF_i))).

simplification_steps_after_folding_mut_ex(Cls,Cls).

has_a_propositional_head(C,Hd,Bd) :-
   clause_number_to_clause(C,(Hd,Bd)), atom(Hd).

propagate_true_in_clause_bodies(_,[],[]) .
propagate_true_in_clause_bodies(Hd,[(H,B)|Cls],[C2|Cls2]) :-
   (contains_atom_in_a_negative_lit(Hd,B,_) ->
                            C2=(H,[false])
    ; (bagof(X, pos(Hd,B,X), Poss) -> apply_delete_goal_((H,B),Poss,C2) 
    ; C2=(H,B))),
   propagate_true_in_clause_bodies(Hd,Cls,Cls2).

contains_atom_in_a_negative_lit(Hd,B,Positions) :-
    bagof(X, (NegLit^Atom^(pos1(NegLit,B,X),
             is_a_negative_literal(NegLit,Atom), Atom==Hd)), Positions).

propagate_false_in_clause_bodies(_,[],[]) .
propagate_false_in_clause_bodies(Hd,[(H,B)|Cls],[C2|Cls2]) :-
   (belongs(Hd,B) -> C2=(H,[false])
   ; (contains_atom_in_a_negative_lit(Hd,B,Poss) ->
         apply_delete_goal_((H,B),Poss,C2) 
    ; C2=(H,B))),
   propagate_false_in_clause_bodies(Hd,Cls,Cls2).

simplification_steps_stratum(L,L_Packets) :-
   (strat_prove(apply_partial_instantiation) ->
       simplification_steps_no_packet(L,[],L_Packets) ;
       (split_into_packets_of_equal_heads_f(L,LPackets2),
        simplification_steps_packets(LPackets2,L_Packets))).

split_into_packets_of_equal_heads_f([],[]).
split_into_packets_of_equal_heads_f([F|Fs],[LF|L2]) :-
   split_into_packets_of_equal_heads_f(Fs,F,[F],LF,[],LRest),
   split_into_packets_of_equal_heads_f(LRest,L2).
split_into_packets_of_equal_heads_f([],_,Irev,I,Lrev,L) :- 
   reverse(Irev,I), reverse(Lrev,L).
split_into_packets_of_equal_heads_f([(H2,B2)|Fs],(H,B),LFI,LF,LRestI,LRest) :-
   variant(H2,H), !,
   split_into_packets_of_equal_heads_f(Fs,(H,B),[(H2,B2)|LFI],LF,LRestI,LRest).
split_into_packets_of_equal_heads_f([F2|Fs],(H,B),LFI,LF,LRestI,LRest) :-
   split_into_packets_of_equal_heads_f(Fs,(H,B),LFI,LF,[F2|LRestI],LRest).


simplification_steps_no_packet([],L,L).
simplification_steps_no_packet([(Hd,Bd)|L],L_I,L_Packets) :-
   (has_a_body_equiv_to_false((Hd,Bd)); Bd==[Hd] ;
           is_subsumed_2((Hd,Bd),L) ;   is_subsumed_2((Hd,Bd),L_I)), !,
    ((L=[_|_];L_I=[_|_]) -> simplification_steps_no_packet(L,L_I,L_Packets)
          ;	L_Packets=[[(Hd,[false])]]).
simplification_steps_no_packet([C|L],L_I,L2) :-
   simplification_steps_no_packet(L,[[C]|L_I],L2).
	
simplification_steps_packets([],[]).
simplification_steps_packets([P|Ps],[SP|SPs]) :-
   simplification_steps_one_packet(P,0,SP),
   simplification_steps_packets(Ps,SPs).

simplification_steps_one_packet([],_,[]).
simplification_steps_one_packet([(Hd,Bd)|L],Num,L2) :-
   (has_a_body_equiv_to_false((Hd,Bd)); Bd==[Hd]), !,
    (L=[_|_] -> simplification_steps_one_packet(L,Num,L2) ;
	(Num=0 -> L2=[(Hd,[false])] ; L2=[])).

simplification_steps_one_packet([(Hd,Bd)|L],_,[S|L2]) :-
    (setof(Pos, boolean_simplification_goals(Bd,Pos), LPos) ->
     apply_delete_goal_((Hd,Bd),LPos,S) ; S=(Hd,Bd)),
   simplification_steps_one_packet(L,1,L2).

%simplification_steps_one_packet([C|L],[C|L2]) :-
%   simplification_steps_one_packet(L,L2).

has_a_body_equiv_to_false((_,Bd)) :-
   member(false,Bd),!.
has_a_body_equiv_to_false((_,Bd)) :-
   member(E,Bd),
   member(E2,Bd), opposite_literals(E,E2), !.

apply_put_body_to_false(C,NewC) :-
   clause_number_to_clause(C,(H,Bd)),
  (Bd==[false] -> true ;
   (add_clauses(NewCs,[(H,[false])],temp), 
   update_programs((C,repclause,NewCs),_), NewCs=[NewC],
   new_clause(repclause,C,NewC))).

boolean_simplification_clause(F,S) :-
   body_of_clause(F,B),
   setof(Pos, boolean_simplification_goals(B,Pos), L),
   apply_delete_goal2(F,L,S).

boolean_simplification_goals(B,Pos) :-
   append(L,[E|R],B), pos(E,R,Pos2), length(L,N), Pos is N+Pos2+1.

%    UNFOLDING STRATEGY

unfolding_steps_stratum([],[]).
unfolding_steps_stratum([CN1|CNs],Us) :-
   unfolding_first_step_stratum_cl(CN1,Us1),
   unfolding_steps_stratum(CNs,Us2), append(Us1,Us2,Us).

unfolding_first_step_stratum_cl(CN,Us) :-
   length_body_clause(CN,L),    clause_number_to_clause(CN,C),
   full_instantiation_if_necessary(C,Cs),
   apply_unfold_lit_uf_strat_list(Cs,L,Us).

partial_instantiation_if_necessary(C,Cs) :-
   apply_partial_instantiation_cl(C,Cs).

apply_partial_instantiation_cl((H,B),Cs) :-
   setof(ArgVar, is_a_variable_to_be_instantiated((H,B),ArgVar), ArgVars),!,
   H =..[PredH|ArgsH],
   my_free_variables((H,B),AllVars),
   infer_pos_and_types_vars(ArgVars,PredH,ArgsH,B,AllVars,Ps,Types),
   apply_instantiate_clause2((H,B),Ps,Types,Cs).
apply_partial_instantiation_cl(C,[C]).

is_a_variable_to_be_instantiated((H,B),X) :-
   is_an_head_only_variable((H,B),X),
   strat_prove(apply_minimal_instantiation).

is_a_variable_to_be_instantiated((_,B),X) :-
   is_a_body_variable_to_be_instantiated(B,X).

is_an_head_only_variable((H,B),X) :-
   H=..[_|ArgsH],
   (member(X,ArgsH), var(X)),
   (check_is_a_variable_argument(B,X) -> fail ; true), !.

is_a_body_variable_to_be_instantiated([Lit|_],ArgVar) :-
   is_a_negative_literal(Lit,Atom),
   Atom=..[_|ArgsH],
   clause3(_,(Hd,_),_), Hd=..[_|ArgsHd],
   subsumes_chk(Atom,Hd),
   pos1(ArgVar,ArgsH,Pos),
   pos1(ArgHd,ArgsHd,Pos), 
   (nonvar(ArgHd) ;    (occurrences_of_var(ArgHd,ArgsHd,X), X > 1)).

is_a_body_variable_to_be_instantiated([_|Lits],ArgVar) :-
   is_a_body_variable_to_be_instantiated(Lits,ArgVar).

full_instantiation_if_necessary(C,Cs) :-
   check_necessary_instantiation(C), !,
   apply_full_instantiation_cl(C,Cs).
full_instantiation_if_necessary(C,[C]).

check_necessary_instantiation(C) :-
   C = (H,B),
   H=..[_|ArgsH],
   (member(X,ArgsH), var(X)),
   (check_is_a_variable_argument(B,X) -> fail ; true), !.

check_necessary_instantiation((H,Bd)) :-
   member(Lit,Bd),
   (is_a_negative_literal(Lit,Atom) ->   (\+ ground((H,Atom))) ;
    (Lit = [_|ArgsLit], H =.. [_|ArgsH],
     member(X,ArgsLit), member(Y,ArgsH), strict_subterm(Y,X))), !.

apply_full_instantiation_cl((H,B),Cs) :-
   setof(ArgVar, is_a_variable_argument([H|B],ArgVar), ArgVars),
   H =..[PredH|ArgsH],
   my_free_variables((H,B),AllVars),
   infer_pos_and_types_vars(ArgVars,PredH,ArgsH,B,AllVars,Ps,Types),
   apply_instantiate_clause2((H,B),Ps,Types,Cs).

infer_pos_and_types_vars([],_,_,_,_,[],[]).
infer_pos_and_types_vars([V|ArgVars],PredH,ArgsH,B,AllVars,[Pos|Ps],[T|Ts]) :-
   pos(V,AllVars,Pos),
   (pos(V,ArgsH,PosH) -> (same_length(ArgsH,Types),
                          type_definition(PredH,_,Types),
                          pos1(T,Types,PosH)) ;
   (member(Lit,B),
    is_a_negative_literal(Lit,Atom),
    Atom=..[PredA|ArgsB],
    pos(V,ArgsB,PosB),
    same_length(ArgsB,Types),
    type_definition(PredA,_,Types),
    pos1(T,Types,PosB))), !,
   infer_pos_and_types_vars(ArgVars,PredH,ArgsH,B,AllVars,Ps,Ts).
infer_pos_and_types_vars([_|ArgVars],PredH,ArgsH,B,AllVars,Ps,Ts) :-
   infer_pos_and_types_vars(ArgVars,PredH,ArgsH,B,AllVars,Ps,Ts).

apply_unfold_lit_uf_strat_list(L,0,L)  :- !.
apply_unfold_lit_uf_strat_list([],_,[]).
apply_unfold_lit_uf_strat_list([C|Cs],AN,NewCs) :-
    apply_unfold_lit_uf_strat(C,AN,NewCs1),
    (NewCs1=[(_,[false])] -> NewCs2=NewCs1 ;
          (AN2 is AN-1,
          apply_unfold_lit_uf_strat_list(NewCs1,AN2,NewCs2))),
    apply_unfold_lit_uf_strat_list(Cs,AN,NewCs3),
    append(NewCs2,NewCs3,NewCs).

apply_unfold_lit_uf_strat((Head,Body1),Pos,Cls) :- 
    (strat_uf(order_before_unfolding(X)) ->
           split_order(Body1,Pos,X,Before,Lit,After) ;
           split(Body1,Pos,Before,Lit,After)),
    (is_a_negative_literal(Lit,Atom) -> 
         apply_unfold_neg_uf_strat(Head,Body1,Before,Atom,After,Pos,Cls)
       ; apply_unfold_pos_uf_strat((Head,Body1),Pos,Cls)).

split_order([Lit|Body1],1,_,[],Lit,Body1) :-!.
split_order(Body1,Pos,XX,NewBefore,Lit,After) :-
	X is Pos-1, length1(Before,X), append(Before,[E|After],Body1),
	weight_lit(E,WeightE),
	(WeightE=< XX -> (NewBefore=Before, Lit=E) ;
             	   choose_smaller_weight_lit(Before,E,WeightE,NewBefore,Lit)).

choose_smaller_weight_lit([],E,_,[],E).
choose_smaller_weight_lit([F|Before],E,WeightE,NewBefore,Lit) :-
        weight_lit(F,N),
	(N=< 1 -> (NewBefore=[E|Before], Lit=F) ;
	    	(N >= WeightE -> (choose_smaller_weight_lit(Before,E,WeightE,NewBefore2,Lit),
				     NewBefore=[F|NewBefore2]) ;
	    	(choose_smaller_weight_lit(Before,F,N,NewBefore2,Lit),
				     NewBefore=[E|NewBefore2]))).

weight_lit(F,W) :-
       (is_a_negative_literal(F,Atom) -> true ; F=Atom),
       free_variables(Atom,[],[],Vars),
       form_goal(Vars, B^P^clause3(N,(Atom,B),P), G),
       ( bagof(N, G, L) ->
           length(L,W) ; W = 0).

form_goal([], G,G).
form_goal([V|Vars], G1, G2) :-
   form_goal(Vars, V^G1, G2).

apply_unfold_pos_uf_strat(C, Pos,Cls) :-
    bagof((Hd,Bd), one_clause_unf(C,Pos,(Hd,Bd)), Cls), !.
apply_unfold_pos_uf_strat((Hd,_),_,[]) :- add_clauses_f(_,[(Hd,[false])],temp).


one_clause_unf(C,Atom_Number,(Head,Body)) :-
    copy_term(C,(Head,Body1)),
    split(Body1,Atom_Number,Before,Atom,After),
    (Atom==true -> BodyD=[] ;
        (same_functor(Atom,HeadD),
	clause3(_,(HeadD,BodyD),_),
         unify_with_occurs_check(HeadD,Atom),
        (BodyD=[false] -> fail ; append(Before,BodyD,After,Body)))).

apply_unfold_neg_uf_strat(Head,Body1,Before,Atom,After,Pos,Cls) :-
    bagof((Hd,Bd), clause_with_head_unifiable_with_f(Atom,Hd,Bd), L), !,
    (check_instantiation_head_before_unfold_neg((Head,Body1),
						Atom,Pos,L,Bodies) ->
         (apply_unfold_neg(Head,Before,After,Bodies,Cls2),
          (Cls2=[] -> Cls=[(Head,[false])] ; Cls2=Cls)) ;
          (apply_interactive_instantiation((Head,Body1),NewCs),
           apply_unfold_lit_uf_strat_list(NewCs,Pos,Cls))).

apply_unfold_neg_uf_strat(Head,_,_,true,_,_,[(Head,[false])]) :- !.
apply_unfold_neg_uf_strat(Head,_,Before,_,After,_,[(Head,Bd)]) :-
    append(Before,After,Bd).

apply_interactive_instantiation(C,NewCs) :-
    my_free_variables(C,Vars),
    length(Vars,VarNumber),
    ask_variable_positions(VarNumber,L),
    (L=[] -> apply_interactive_instantiation(CN,NewCs) ;
       (ask_type_of_variables(1,L,Types),
        check_input_var_numbers_and_types(CN,L,Types),
        apply_instantiate_clause2(C,L,Types,NewCs))).

clause_with_head_unifiable_with_f(Atom,Hd,Bd) :-
    same_functor(Atom,Hd),
    clause3(_,(Hd,Bd),_),
    unifiable(Hd,Atom).

% INSTANTIATION STRATEGY

apply_interactive_instantiation(CN,AN,NewCNs) :-
    clause_number_to_clause(CN,C),
    my_free_variables(C,Vars),
    length(Vars,VarNumber),
    ask_variable_positions(VarNumber,L),
    (L=[] -> apply_interactive_instantiation(CN,AN,NewCNs) ;
     ask_type_of_variables(1,L,Types),
     check_input_var_numbers_and_types(CN,L,Types),
     apply_instantiate_clause(C,L,Types,Clauses),
     add_clauses(NewCNs,Clauses,temp),
     update_programs((CN,(instantiation,L,Types), NewCNs), _), 
     new_clauses(instantiation,CN,L,Types,NewCNs)).

retract_clauses([],[]).
retract_clauses([CN|CNs],[C|Cs]) :-
	retract(clause3(CN,C,_)), retract_clauses(CNs,Cs).

instantiate_clause([],C,_,[C]).
instantiate_clause([C|Cs],CI,Atom,NewCs) :-
   clause_number_to_clause(C,(H,_)),
   ((subsumes_chk(Atom,H), \+ variant(Atom,H)) ->
      (infer_patterns(Atom,H,Atoms),
       apply_instantiate_clause_(CI,Atom,Atoms,NewCs)) 
    ; instantiate_clause(Cs,CI,Atom,NewCs)).

infer_patterns(Atom,H,Atoms) :-
    functor(Atom,F,N),
    type_definition(F,N,Type),
    infer_patterns(Atom,H,Type,N,Atoms).

infer_patterns(Atom,H,Type,N,Atoms) :- N > 0,
    arg(N,Atom,ArgNAtom),
    arg(N,H,ArgNH),
   (subsumes_chk(ArgNAtom,ArgNH), \+ variant(ArgNAtom,ArgNH)) ->
    (arg(N,Type,ArgNType),
    current_clauses(Cls),
    type_pattern(Cls,ArgNType,Patterns),
    apply_instantiate_clause_(Atom,ArgNAtom,Patterns,Atoms)) ;
    (N2 is N - 1, infer_patterns(Atom,H,Type,N2,Atoms)).

rearrangement_after_lloyd_topor(Ss,NewSs) :-
   rearrangement_after_lloyd_topor(Ss,[],_,NewSs).

rearrangement_after_lloyd_topor([],I,I,[]).
rearrangement_after_lloyd_topor([S|Ss],NewDefsI,NewDefsI2,NewSs) :-
   rearrangement_after_lloyd_topor(Ss,NewDefsI,NewDefsI3,S2s),
   boolean_simplification_after_lloyd_topor(S,S2),
%   subsumption_after_lloyd_topor(S2,S3),
   rearrangement_after_lloyd_topor_one_stratum(S2,S4,NewDefsI3,NewCls1),
   append(NewDefsI3,NewCls1,NewDefsI2),
   (NewCls1==[] -> NewSs=[S4|S2s] ; NewSs=[S4,NewCls1|S2s]).

boolean_simplification_after_lloyd_topor([],[]) :- !.
boolean_simplification_after_lloyd_topor([(_,B)|Cs],Cs2) :-
   (belongs(false,B) ; (member(E,B), member(E2,B), opposite_literals(E,E2))),
   !,
   boolean_simplification_after_lloyd_topor(Cs,Cs2).
boolean_simplification_after_lloyd_topor([(H,B)|Cs],[(H,B3)|Cs2]) :-
   remove_dupplicates(B,B2),
   del_3(true,B2,B3),
   boolean_simplification_after_lloyd_topor(Cs,Cs2).
   


rearrangement_after_lloyd_topor_one_stratum([],[],_,[]).
rearrangement_after_lloyd_topor_one_stratum([C|Cs],[C2|Cs2],NewDefsI,L) :-
   rearrangement_after_lloyd_topor_one_clause(C,C2,NewDefsI,LC),
   append(NewDefsI,LC,NewDefsI2),
   rearrangement_after_lloyd_topor_one_stratum(Cs,Cs2,NewDefsI2,LCs),
   append(LCs,LC,L).



rearrangement_after_lloyd_topor_one_clause((H,B),(H,B),_,[]) :-
    is_a_negative_literal_with_vars_only(B), !.

rearrangement_after_lloyd_topor_one_clause((H,B),C,NewDefsI,NewCs) :-
   append(L,[E|R],B),
   (is_a_negative_literal(E,A) ;
       (   contains_a_non_variable_argument(E), E=A)), !,
   ( (member((HDef,BDef),NewDefsI), variant(BDef,[E])) ->
       (functor(HDef,X,_), free_variables(A,[],[],Vars),
        NewH =..[X|Vars], append(L,[NewH|R],B2), LNewC=[], NewDefsI2=NewDefsI)
    ;  ( gen_new_name_functor(X), 
         free_variables(A,[],[],Vars),
         NewH =..[X|Vars], append(L,[NewH|R],B2),
         NewB = [E], copy_term((NewH,NewB), NewC), NewDefsI2=[NewC|NewDefsI], LNewC=[NewC])),
      rearrangement_after_lloyd_topor_one_clause((H,B2),C,NewDefsI2,L2), append(L2, LNewC, NewCs).

rearrangement_after_lloyd_topor_one_clause((H,B),C,NewDefsI,NewCs) :-
   strat_uf(tuppling), int_var(H,B,Vars), !,
   tuppling_after_lloyd_topor_one_clause((H,B),Vars,C,NewDefsI,NewCs).

rearrangement_after_lloyd_topor_one_clause((H,B),(H,B),_,[]).


tuppling_after_lloyd_topor_one_clause((H,B),_,(H,B),_,[]) :-
	length(B,L), L =< 2, !.

tuppling_after_lloyd_topor_one_clause((H,[A1,A2|B]),[],C,NewDefsI,NewCs) :-
    ((member((HDef,BDef),NewDefsI), variant(BDef,[A1,A2])) ->
       (functor(HDef,X,_), free_variables([A1,A2],[],[],Vars),
        NewH =..[X|Vars], B2=[NewH|B], LNewC=[], NewDefsI2=NewDefsI)
    ;  ( gen_new_name_functor(X), 
         free_variables([A1,A2],[],[],Vars),
         NewH =..[X|Vars], B2=[NewH|B],
         NewB = [A1,A2], copy_term((NewH,NewB), NewC), NewDefsI2=[NewC|NewDefsI], LNewC=[NewC])),
      tuppling_after_lloyd_topor_one_clause((H,B2),[],C,NewDefsI2,L2), append(L2, LNewC, NewCs).

tuppling_after_lloyd_topor_one_clause((H,B),[_|_],C,NewDefsI,NewCs) :-
	(( subseq(B, [A1,A2], RestB), free_variables(A1,H,[],VarsIntA1),
          free_variables(A2,H,[],VarsIntA2), intersection_vars_non_empty(VarsIntA1,VarsIntA2)) ;
	    B=[A1,A2|RestB]), !,
       free_variables([A1,A2],[],[],VarsFoldedPart),
    linking_vars_folded_part(H,[],RestB,VarsFoldedPart,LV),
    ((member((HDef,BDef),NewDefsI), HDef=[X|ArgsDefs], variant((ArgsDefs,BDef),(LV,[A1,A2]))) ->
       (NewH =..[X|LV], B2=[NewH|RestB], LNewC=[], NewDefsI2=NewDefsI)
    ;  ( gen_new_name_functor(X), 
         NewH =..[X|LV], B2=[NewH|RestB],
         NewB = [A1,A2], copy_term((NewH,NewB), NewC), NewDefsI2=[NewC|NewDefsI], LNewC=[NewC])),
      free_variables(B2,H,[],VarsInt),
      tuppling_after_lloyd_topor_one_clause((H,B2),VarsInt,C,NewDefsI2,L2), append(L2, LNewC, NewCs).

	

subsumption_after_lloyd_topor(S2,S3) :-
   subsumption_after_lloyd_topor(S2,[],S3).

subsumption_after_lloyd_topor([],L,L).
subsumption_after_lloyd_topor([E|L],I,F) :-
   is_subsumed_2(E,L), !, 
   subsumption_after_lloyd_topor(L,I,F).
subsumption_after_lloyd_topor([E|L],I,F) :-
   is_subsumed_2(E,I), !,
   subsumption_after_lloyd_topor(L,I,F).
subsumption_after_lloyd_topor([E|L],Inter,F) :-
    append(Inter,[E],Inter2), subsumption_after_lloyd_topor(L,Inter2,F).

is_subsumed_2((H1,B1),[(H2,B2)|_]):-
   same_length(X,B2),
   (X=[];subseq(X,B1)), subsumes_chk([H2|B2],[H1|X]).
is_subsumed_2(E,[_|L]):- is_subsumed_2(E,L).

has_the_same_head_as_another_clause((H1,_),[(H2,_)|_]) :-
   variant(H1,H2), !.
has_the_same_head_as_another_clause(F,[_|Fs]) :-
   has_the_same_head_as_another_clause(F,Fs).

try_to_remove_dupplicated_clauses_in_defs([],[]).
try_to_remove_dupplicated_clauses_in_defs([Def|Defs],[Def2|Defs2]) :-
   try_to_remove_dupplicated_clauses_in_def(Def,Def2),
   try_to_remove_dupplicated_clauses_in_defs(Defs,Defs2).

try_to_remove_dupplicated_clauses_in_def([C],[C]) :- !.
try_to_remove_dupplicated_clauses_in_def([C|Defs],Defs2) :- 
   try_to_remove_dupplicated_clauses_in_def(Defs,Defs3),
   ((member(D,Defs3), variant(C,D)) -> Defs2=Defs3; Defs2=[C|Defs3]).

try_to_remove_synonymous_def(Strata1,Strata2) :-
    try_to_remove_synonymous_def([],Strata1,Strata2).

try_to_remove_synonymous_def(Defs,[],Defs).
try_to_remove_synonymous_def(Defs,[S|Defs1],Defs2) :-
    exists_variant_definition(S,Newi,Defs1,Newj), !,
    rename_higher_strata(Defs,Newi,Newj,DefsR), append(DefsR,Defs1,NewDefs),
    try_to_remove_synonymous_def([],NewDefs,Defs2).
try_to_remove_synonymous_def(Defs,[S|Defs1],Defs2) :-
    append(Defs,[S],Defs3),
    try_to_remove_synonymous_def(Defs3,Defs1,Defs2).

exists_variant_definition(S,Newi,Defs1,Newj) :-
    length(S,LengthS),
    member(S2,Defs1),
    length(S2,LengthS),
    variant_stratum(S,Newi,S2,Newj).

variant_stratum([],_,[],_).
variant_stratum([(Hd,Bd)|L1],Newi,[(Hd2,Bd2)|L2],Newj) :-
    Hd=..[Newi|ArgsHd],     Hd2=..[Newj|ArgsHd2],
    append(ArgsHd,Bd,X1),
    append(ArgsHd2,Bd2,X2),
    variant(X1,X2),
    variant_stratum(L1,Newi,L2,Newj).

rename_higher_strata([],_,_,[]).
rename_higher_strata([S1|Strata],Newi,Newj,[S1R|StrataR]) :-
    rename_higher_stratum(S1,Newi,Newj,S1R),
    rename_higher_strata(Strata,Newi,Newj,StrataR).

rename_higher_stratum([],_,_,[]).
rename_higher_stratum([(H,B)|S1],Newi,Newj,[(H,BR)|S1R]) :-
    rename_def_in_body_clause(B,Newi,Newj,BR),
    rename_higher_stratum(S1,Newi,Newj,S1R).
rename_def_in_body_clause([],_,_,[]).
rename_def_in_body_clause([\+(Atom)|B],Newi,Newj,[\+(Atom2)|BR]) :-
     Atom=..[Newi|Args],!,
     Atom2=..[Newj|Args],
     rename_def_in_body_clause(B,Newi,Newj,BR).
rename_def_in_body_clause([not(Atom)|B],Newi,Newj,[not(Atom2)|BR]) :-
     Atom=..[Newi|Args],!,
     Atom2=..[Newj|Args],
     rename_def_in_body_clause(B,Newi,Newj,BR).
rename_def_in_body_clause([Atom|B],Newi,Newj,[Atom2|BR]) :-
     Atom=..[Newi|Args],!,
     Atom2=..[Newj|Args],
     rename_def_in_body_clause(B,Newi,Newj,BR).
rename_def_in_body_clause([Atom|B],Newi,Newj,[Atom|BR]) :-
         rename_def_in_body_clause(B,Newi,Newj,BR).


msg_after_lloyd_topor_transformation(LDefs) :-
   nl, write('After Lloyd Topor transformation: '), nl,
   show_clauses(LDefs).

msg_after_transformation_of_stratum(_,NumStratum,NewS1,NewDefs) :-
   msg_after_transformation_of_one_def(PredDef1,NewS1,NewDefs),
   write('End of transformation of stratum '), write(NumStratum), nl.

one_definition_in_stratum([C1,C2|S1],[C1|Def1],PredDef1,RestS1) :-
   C1=(Hd1,_), C2=(Hd2,_),
   predname_and_arity(Hd1,PredDef1,Arity),
   predname_and_arity(Hd2,PredDef1,Arity),!,
   one_definition_in_stratum([C2|S1],Def1,PredDef1,RestS1).
one_definition_in_stratum([C1|S1],[C1],PredDef1,S1) :-
    (var(PredDef1)->    ( %clause_number_to_clause(C1,(Hd1,_)), 
                         C1=(Hd1,_),
                         predname_and_arity(Hd1,PredDef1,_)) ; true).

update_transformation_tree_after_lloyd_topor(Defs) :- flatten(Defs,LDefs),
   update_programs((after_lloyd_topor,LDefs), _),
   msg_after_lloyd_topor_transformation(LDefs).

update_transformation_tree_after_one_stratum(Def1,PredDef1,NumStratum,
					     NewDef1,NewDefs) :-
   update_programs((Def1,strategy_one_stratum,NewDef1), _),
   msg_after_transformation_of_stratum(PredDef1,NumStratum,NewDef1,NewDefs).

update_transformation_tree_after_one_def(Def1,PredDef1,NewDef1,_) :-
   update_programs((Def1,strategy_one_stratum,NewDef1), _).

update_transformation_tree_after_uf_strat(L) :-
   bagof(Cl, C^(clause3(Cl,C,uf), current_clauses(P), \+ member(Cl,P)), L2), update_programs((L,after_uf,L2),_).

