%:- module(specialize_ctlstar, [specialize_ctlstar/5]).

:- use_module(library(terms)).
:- use_module(library(ordsets)).
:- use_module(basic_operations).

specialize_ctlstar_strategy(Cls1,Prog1,Defs1,Prog2,Defs2) :-
	specialize_ctlstar_cls(Cls1,Prog1,Defs1,Cls2,Defs2),
	hd_preds(Cls2,Preds),
	cls_to_defs(Preds,Cls2,Prog2).

% specialize_ctlstar_cls :- the main loop of the specialize_ctlstar strategy

specialize_ctlstar_cls([],_Prog,Defs,[],Defs).
specialize_ctlstar_cls([Cl|Cls],Prog_in,Defs_in,Cls_out,Defs_out) :-
	write('.'),
	ctlstar_unfold(Cl,Prog_in,UnfoldedCls),
         (derivation_trace_on ->
          (write(' -unfold- '), nl, show_cls(UnfoldedCls), nl);
          true),
 	simplify(UnfoldedCls,SimplifiedCls),
         (derivation_trace_on ->
          (write(' -simplify- '), nl, show_cls(SimplifiedCls), nl);
          true),
 	ctlstar_define_fold(SimplifiedCls,Defs_in,FoldedCls,NewDefs,Defs1),
         (derivation_trace_on ->
          (write(' -define- '), nl, show_cls(NewDefs),  nl,
	   write(' -fold- '), nl, show_cls(FoldedCls), nl);
          true),
	append(NewDefs,Cls,Cls1),
	append(FoldedCls,Cls_out1,Cls_out),
	specialize_ctlstar_cls(Cls1,Prog_in,Defs1,Cls_out1,Defs_out).


% UNFOLD STRATEGY

ctlstar_unfold(Cl,Prog,UCls) :-
	unfold_clause(leftmost,Cl,Prog,Cls),
	repeat_unfold(ctlstar,Cls,Prog,UCls).

repeat_unfold(_Uselection,[],_Prog,[]) :- !.
repeat_unfold(Uselection,[Cl|Cls_in],Prog,Cls_out) :-
	unfold_clause(Uselection,Cl,Prog,UCls),!,
	append(UCls,Cls_in,Cls_in1),
	repeat_unfold(Uselection,Cls_in1,Prog,Cls_out).
repeat_unfold(Uselection,[Cl|Cls_in],Prog,Cls_out) :-
	Cls_out=[Cl|Cls_out1],
	repeat_unfold(Uselection,Cls_in,Prog,Cls_out1).

unfold_clause(Uselection,Cl,Prog,Cls) :-
	unfolding_selection(Uselection,Cl,Cl1),
	pos_unfold(Cl1,Prog,Cls).

unfolding_selection(leftmost,(H,bd(Cns,[A|Lits])),(H,bd(Cns,[],[A|Lits]))) :- !,
	\+ A=(\+ _).
unfolding_selection(ctlstar,(H,bd(Cns,[A|Lits])),(H,bd(Cns,LLits,[S|RLits]))) :-
	 unfoldable(ctlstar,A), !,
	(LLits=[], S=A, RLits=Lits).
unfolding_selection(ctlstar,(H,bd(Cns,[A|Lits])),(H,bd(Cns,LLits,[S|RLits]))) :-
	LLits=[A|LLits1],
 	unfolding_selection(ctlstar,(H,bd(Cns,Lits)),(H,bd(Cns,LLits1,[S|RLits]))).

unfoldable(ctlstar,t(_,_)).
unfoldable(ctlstar,elem(_,_)).
unfoldable(ctlstar,sats(_,_)).
unfoldable(ctlstar,satp(_,F)) :- \+ F=u(_,_), \+ F=f(_).

%/*
pos_unfold((H,bd(C,G,[A|Ls])),Prog,Cls) :-
	literal_pred(A,P),
	def(P,Prog,DefP),
	resolvents((H,bd(C,G,[A|Ls])),DefP,Cls).


resolvents(_Cl,[],[]).
resolvents(Cl1,[Cl2|Cls],Res) :-
	copy_term(Cl1,(H1,bd(C1,G1,[A1|Ls1]))),
	copy_term(Cl2,(H2,bd(C2,G2))),
	A1=H2,
	append(C1,C2,C3),
	term_variables((H1,G1,G2,Ls1),V),
	solve(C3,V,C4),
	satisfiable(C4), !,
	append(G1,G2,G1G2),
	append(G1G2,Ls1,G1G2Ls1),
	Res=[(H1,bd(C4,G1G2Ls1))|Res1],
	resolvents(Cl1,Cls,Res1).
resolvents(Cl1,[_Cl2|Cls],Res) :-
	resolvents(Cl1,Cls,Res).
%*/

/*
pos_unfold(Cl,Prog,Cls) :-
	copy_term(Cl,(H,bd(C,G,[A|Ls]))),
	literal_pred(A,P),
	def(P,Prog,DefP),
	copy_term(DefP,DefP1),
	findall(D,
		(member((A,bd(C1,G1)),DefP1),
		 append(C,C1,C2),
	         term_variables((H,G,G1,Ls),V),
		 solve(C2,V,C3),
		 satisfiable(C3),
		 append(G,G1,GG1),
		 append(GG1,Ls,GG1Ls),
		 copy_term((H,bd(C3,GG1Ls)),D)),
		Cls).
*/

% SIMPLIFY STRATEGY

% Constraint replacement is done during unfolding

simplify(Cls,Cls).


% DEFINE_FOLD STRATEGY

ctlstar_define_fold(Cls,Defs_in,FoldedCls,NewDefs,Defs_out) :-
	repeat_define_fold(ctlstar,Cls,Defs_in,FoldedCls,NewDefs,Defs_out).

repeat_define_fold(_Fselection,[],Defs,[],[],Defs).
repeat_define_fold(Fselection,[Cl|Cls],Defs_in,FoldedCls,NewDefs,Defs_out) :-
	define_fold_clause(Fselection,Cl,Defs_in,FCl,NewDs),!,
	Cls1=[FCl|Cls],
	append(NewDs,Defs_in,Defs1),
        append(NewDs,NewDefs1,NewDefs),
	repeat_define_fold(Fselection,Cls1,Defs1,FoldedCls,NewDefs1,Defs_out).
repeat_define_fold(Fselection,[Cl|Cls],Defs_in,FoldedCls,NewDefs,Defs_out) :-
	FoldedCls=[Cl|FoldedCls1],
	repeat_define_fold(Fselection,Cls,Defs_in,FoldedCls1,NewDefs,Defs_out).

define_fold_clause(Fselection,Cl,Defs,FoldedCl,NewDs) :-
	folding_selection(Fselection,Cl,Cl1),
	define_clause(Cl1,Defs,Def,NewDs),
	fold_clause(Cl1,Def,FoldedCl).

folding_selection(ctlstar,(H,bd(Cns,[L|Lits])),(H,bd(Cns,LLits,[SL|RLits]))) :-
	foldable(ctlstar,L),!,
	LLits=[], SL=L, RLits=Lits.
folding_selection(ctlstar,(H,bd(Cns,[L|Lits])),(H,bd(Cns,LLits,[SL|RLits]))) :-
	LLits=[L|LLits1],
	folding_selection(ctlstar,(H,bd(Cns,Lits)),(H,bd(Cns,LLits1,[SL|RLits]))).

foldable(ctlstar,sats(_,_)).
foldable(ctlstar,satp(_,_)).
foldable(ctlstar,(\+ sats(_,_))).
foldable(ctlstar,(\+ satp(_,_))).

define_clause(Cl,Defs,Def,NewDs) :-
	copy_term(Cl,(_,bd(Cns,_,[L|_]))),
	term_variables(L,LV),
	candidate_def(LV,Cns,L,CD),
	introduce_def(Defs,CD,Def,NewDs).

introduce_def(Defs,CD,Def,NewDs):-
% either an old definition can be used for folding
	member(D,Defs),
	copy_term(D,D1),
	equiv_cl(D1,CD),!,
	Def=D1,
	NewDs=[].
introduce_def(_Defs,CD,Def,NewDs):-
% or else a new definition is introduced
	Def=CD,
	NewDs=[Def].

candidate_def(V,Cns,L,CD) :-
	gen_new_name_functor(Pred),
	term_variables(L,Vars),
	solve(Cns,Vars,Cns1),
	NewH=..[Pred|V],
	atom(L,A),
	CD=(NewH,bd(Cns1,[A])).

equiv_cl((H1,bd(Cns1,A1)),(H2,bd(Cns2,A2))) :-
	variant(A1,A2),
	A1=A2,
	entails(Cns2,Cns1),
	entails(Cns1,Cns2),
	term_variables(H1,Vars1),
	term_variables(H2,Vars2),
	subset_chk(Vars1,Vars2),
	subset_chk(Vars2,Vars1).

subset_chk([],_).
subset_chk([T1|Ts1],Ts2) :-
	member_chk(T1,Ts2),
	subset_chk(Ts1,Ts2).

fold_clause(Cl,Def,FoldedCl) :-
	copy_term(Cl,(H,bd(Cns,LLits,[L|RLits]))),
	copy_term(Def,(H1,bd(_,[A1]))),
	replace_literal(L,A1,H1,L1),
	append(LLits,[L1|RLits],NewLits),
	FoldedCl=(H,bd(Cns,NewLits)).

replace_literal(L,A1,H1,L1) :-
	L=A1, !, L1=H1.
replace_literal(L,A1,H1,L1) :-
	L= (\+ A1), L1= (\+ H1).


/*
linking_variables(Ctx,G,LinkingVars) :-
        setof(X,(
        term_variables(G,VG),
        term_variables(Ctx,VCtx),
        member(X,VG),
        member_chk(X,VCtx)), LV) -> LinkingVars=LV ; LinkingVars=[].

add_cls_to_prog([],Prog,Prog).
add_cls_to_prog([Cl|Cls],Prog1,Prog3) :-
	add_cl_to_prog(Cl,Prog1,Prog2),
	add_cls_to_prog(Cls,Prog2,Prog3).

add_cl_to_prog(Cl,Prog1,Prog2) :-
	hd_pred(Cl,P/N),
	add_cl_to_def_in_prog(P/N,Cl,Prog1,Prog2).

add_cl_to_def_in_prog(P,Cl,[],[(P,[Cl])]).
add_cl_to_def_in_prog(P0,Cl,[(P1,Def1)|Prog1],[(P1,Def2)|Prog2]) :-
	P0=P1 ->
	(Def2=[Cl|Def1], Prog2=Prog1);
	(Def2=Def1, add_cl_to_def_in_prog(P0,Cl,Prog1,Prog2)).
*/





