:- use_module(library(ordsets), [is_ordset/1, list_to_ord_set/2]).

msg_info_strata(Strata,DefsN) :-
   make_new_defs_strata(Strata,Defs),
   try_to_remove_synonymous_def(Defs,Defs3),
   add_new_defs_strata(Defs3,DefsN).

add_new_defs_strata(LL1,LL2) :-
   max_clause(N),    
   apply_gen_define_l_f(N,LL1,LL2).

make_new_defs_strata([],[]).
make_new_defs_strata([S|Strata],Defs) :-
    make_new_defs_strata(Strata,DefsN),
    make_new_defs_stratum(S,Defs1N), append(Defs1N,DefsN,Defs).

make_new_defs_stratum([],[]).
make_new_defs_stratum([C|S],[Defs1|Defs2]) :-
   make_new_defs_stratum(C,S,[],S2,Defs1),
   make_new_defs_stratum(S2,Defs2).% append(Defs1,Defs2,Defs).

make_new_defs_stratum(C,[],D,[],[C|D]) :- !.
make_new_defs_stratum((Hd,Bd),[(Hd2,Bd2)|S], D,S2,DefsN) :-
   predname_and_arity(Hd,F,A),
   predname_and_arity(Hd2,F,A), !,
   append(D,[(Hd2,Bd2)],D2),
   make_new_defs_stratum((Hd,Bd),S,D2,S2,DefsN).
make_new_defs_stratum(C,S,D,S,[C|D]).

apply_gen_define_l_f(N,[],[]) :-
   retract(max_clause(_)), assert(max_clause(N)).
%apply_gen_define_l_f(N,[Def1|Defs],[Def1N|DefsN]) :-
%   gen_X(Def1,N,X), add_clauses_f(N,N2,Def1N, Def1, [gen_def|X]),
%   assert(clauses_of_new_definition(X,Def1)),
%   apply_gen_define_l_f(N2,Defs,DefsN).
apply_gen_define_l_f(N,[Def1|Defs],[Def11N|DefsN]) :-
   one_definition_in_stratum(Def1,Def11,_,RestDef1),
   gen_X(Def11,N,X), add_clauses_f(N,N2,Def11N, Def11, [gen_def|X]),
%   assert(clauses_of_new_definition(X,Def11)),
    (RestDef1=[] -> apply_gen_define_l_f(N2,Defs,DefsN) ;
           apply_gen_define_l_f(N2,[RestDef1|Defs],DefsN)).

apply_gen_define2_l_f(N,[],[]) :-
   retract(max_clause(_)), assert(max_clause(N)).
apply_gen_define2_l_f(N,[Def1|Defs],[Def11N|DefsN]) :-
   one_definition_in_stratum(Def1,Def11,_,RestDef1),
   gen_X(Def11,N,X), add_clauses_f(N,N2,Def11N, Def11, [gen_def|X]),
   assert(clauses_of_new_definition(X,Def11)),
    (RestDef1=[] -> apply_gen_define2_l_f(N2,Defs,DefsN) ;
           apply_gen_define2_l_f(N2,[RestDef1|Defs],DefsN)).


add_clauses_f(N,N,[],[],_).
add_clauses_f(N2,N3,[M|Ms],[C|Cs], Aux) :-
     M is N2 + 1, assert(clause3(M,C,Aux)),
     add_clauses_f(M,N3,Ms,Cs, Aux).

add_clauses_f(LN,L,Aux) :-
   max_clause(N), add_clauses_f(N,N2,LN,L,Aux),
   retract(max_clause(N)), assert(max_clause(N2)).

msg_new_types_and_modes([]). 
msg_new_types_and_modes([[(H,_)|_]|S]) :-
   predname_and_arity(H,P,A),
   clause(type_definition(P,A,T),true), clause(mode_definition(P,A,M),true),
   write(P/A), tab(5), write(T), tab(5), write(M), nl,
   msg_new_types_and_modes(S).

ask_type_of_variables(_,[],[]).
ask_type_of_variables(N,[_|Vars],[T|Types]) :-
    write('type of variable '), write(N), write(': '), read(T),
    N2 is N + 1,
    ask_type_of_variables(N2,Vars,Types).

not_ordered(Vs) :- is_ordset(Vs), !, fail.
not_ordered(_).

reorder(Vs,Ts,Vs2,Ts2) :-
   two_lists_in_one(Vs,Ts,L),
   list_to_ord_set(L,Set),
   two_lists_in_one(Vs2,Ts2,Set).

two_lists_in_one([],[],[]).
two_lists_in_one([V|Vs],[T|Ts],[pair(V,T)|Pairs]) :-
   two_lists_in_one(Vs,Ts,Pairs).

opposite_literals(E,F) :- 
   ((is_a_negative_literal(E,A), A==F) ;
    (is_a_negative_literal(F,A), E==A)), !.


has_a_false_body(C) :-
   body_of_clause(C,B),
   member(false,B), !.

is_a_conjunction_of_atoms_with_vars_only([]).
is_a_conjunction_of_atoms_with_vars_only([E|B]) :-
   E =.. [_|L],
   is_a_list_of_vars(L),
   is_a_conjunction_of_atoms_with_vars_only(B).


contains_a_non_variable_argument(E) :-
   E =.. [_|L],
   is_a_list_of_vars(L), !, fail.
contains_a_non_variable_argument(_).

is_a_negative_literal_with_vars_only([E]) :-
    (E =.. [not,_|L] ;     E =.. [\+,_|L]),
   is_a_list_of_vars(L).

is_a_list_of_vars([]).
is_a_list_of_vars([V|Vs]) :-
   var(V),
   is_a_list_of_vars(Vs).

contains_no_unnecessary_var(C) :-
   clause_number_to_clause(C,(H,B)),
   contains_no_unnecessary_var2((H,B)).
contains_no_unnecessary_var2((H,B)) :-
   int_var(H,B,[]),
   my_all_free_variables(B,Vars),
   no_dupplicates(Vars), !.

contains_no_existential_var(C) :-
   clause_number_to_clause(C,(H,B)),
   int_var(H,B,[]).
contains_no_dupplicate_var_in_bd(C) :-
   body_of_clause(C,B),
   my_all_free_variables(B,Vars),
   no_dupplicates(Vars), !.

split_into_strata_([],[]) :- !.
split_into_strata_(L,[S1|Ss]) :-
   first_stratum(L,S1,L2),
   split_into_strata_(L2,Ss).
first_stratum([],[],[]).
first_stratum([E|L1],S1,L3) :-
   is_a_list(E), !,
   first_stratum(L1,S1,L2), append(E,L2,L3).
first_stratum([E|L1],[E|S1],L2) :-
   first_stratum(L1,S1,L2).


is_a_list([]).
is_a_list([_|_]).

infer_types_strata(Cs) :-
      flatten(Cs,LCs),
      infer_types(LCs).

infer_types(L) :- reverse(L,Linv), infer_types2(Linv).
infer_types2([]).
infer_types2([(Hd,Bd)|Cs]) :-
   functor(Hd,F,N),
   (clause(type_definition(F,N,Type),true) -> 
   (my_ground(Type) -> true ; infer_one_type(Hd,F,N,N,Bd,Type)) ;
    infer_one_type(Hd,F,N,Bd)),
    infer_types2(Cs).

infer_one_type(Hd,F,N,Bd) :- 
   length(Type,N),
   infer_one_type(Hd,F,N,N,Bd,Type).
infer_one_type(_,F,N,0,_,Type) :-
    (clause(type_definition(F,N,_),true) ->
       retractall(type_definition(F,N,_)) ; true),
    assert(type_definition(F,N,Type)).

infer_one_type(Hd,F,N,N2,Bd,Type) :- N2 > 0,
   arg(N2,Hd,ArgHd),
   pos1(ArgType,Type,N2),
   (my_ground(ArgType) -> true ; infer_type_arguments(ArgHd,Bd,ArgType)), 
   N3 is N2 - 1,
   infer_one_type(Hd,F,N,N3,Bd,Type).

infer_type_arguments(ArgHd,Bd,ArgType) :-
    member(Lit,Bd),
    (is_a_negative_literal(Lit,A) ; Lit=A),
   functor(A,F,N),
   A=..[_|L],
    (pos(ArgHd,L,Pos) -> (  type_definition(F,N,Type) ->
			                  pos1(ArgType,Type,Pos) ;
                                     (N=1,F=ArgType));
    ((pos1(Arg,L,Pos), subterm(ArgHd,Arg)) ->
                         (  type_definition(F,N,Type), pos1(T,Type,Pos),
                            infer_sub_type(ArgHd,Arg,T,ArgType)))), !.
infer_type_arguments(_,_,_).

infer_sub_type(ArgHd,Arg,T,T) :- ArgHd==Arg, !.
infer_sub_type(ArgHd,Arg,T,ArgType) :- 
   Hd=..[T,Arg], clause3(_,(Hd,Bd),prog),
   infer_type_arguments(ArgHd,Bd,ArgType).

enter_property(Phi,A,Phi_L) :-
   (var(Phi) -> (write('enter a property: '), nl, read(Phi), nl) ; true),
   gen_new_name_functor(Pred),
   vars(Phi,[],L),
   A =.. [Pred|L], conj_to_list(Phi,Phi_L).

lloyd_topor_and_rearrangement(A,Phi_L,Defs) :-
      lloyd_topor((A,Phi_L),Cs),
      split_into_strata_(Cs,Strata),
%      apply_subsumption_in_strata(Strata,Strata_S),
      rearrangement_after_lloyd_topor(Strata,Strata2),
      msg_info_strata(Strata2,Defs),
      infer_types_strata(Strata2),
      update_transformation_tree_after_lloyd_topor(Defs).

lloyd_topor_and_rearrangement(Phi,Defs) :-
      enter_property(Phi,A,Phi_L),
      lloyd_topor_and_rearrangement(A,Phi_L,Defs).




apply_subsumption_in_strata([],[]).
apply_subsumption_in_strata([S|Strata],[S_S|Strata_S]) :-
      apply_subsumption_clauses_f(S,[],S_S),
      apply_subsumption_in_strata(Strata,Strata_S).
apply_subsumption_in_stratum([],[]).
	


split_into_strata(Cs,Defs) :-
      split_into_strata_(Cs,Strata),
      msg_info_strata(Strata,Defs),
      msg_info_types(Strata).

msg_theorem_proved(PredNum) :-
     name(PredNum,Name), name(NamePredNum,[110,101,119|Name]),
     (clause3(_, (NamePredNum,[]), _) -> (nl, write(NamePredNum), write(' is true !'), nl) ;
     (clause3(_, (NamePredNum,[false]), _) -> (nl, write(NamePredNum), write(' is false !'), nl) ; true)).

vars(a(L, Phi), L2, L3) :- !,
   add_new(L,L2,L4),
   vars(Phi,L4,L3).
vars(e(L, Phi), L2, L3) :- !,
   add_new(L,L2,L4),
   vars(Phi,L4,L3).
vars(not(Phi), L, L2) :- !,
   vars(Phi,L,L2).
vars(\+(Phi), L, L2) :- !,
   vars(Phi,L,L2).
vars(and(Phi,Phi_2), L, L4) :- !,
   vars(Phi,L,L2),
   vars(Phi_2,L,L3),
   add_new(L2,L3,L4).
vars((Phi,Phi_2), L, L4) :- !,
   vars(Phi,L,L2),
   vars(Phi_2,L,L3),
   add_new(L2,L3,L4).
vars(or(Phi,Phi_2), L, L4) :- !,
   vars(Phi,L,L2),
   vars(Phi_2,L,L3),
   add_new(L2,L3,L4).
vars(->(Phi,Phi_2), L, L4) :- !,
   vars(Phi,L,L2),
   vars(Phi_2,L,L3),
   add_new(L2,L3,L4).
vars(Phi,L,L2) :- !,
   free_variables(Phi,L,[],L2).

my_ground(X) :- free_variables(X,[],[],[]).


/* abbreviations */

v2:- ['src_sics/uf_strategy_v2.pl'], ['src_sics/uf_strategy_util_v2.pl'].

ld(cls) :- ['src_sics/clause_replacement_rules.pl'].
/*
a( [F1], e([F2], a([X], or( (m(X,F1), \+(m(X,F2))), (\+(m(X,F1)), m(X,F2)))))).
*/
l_t :- lloyd_topor(_).
lloyd_topor :- lloyd_topor(_).

unf_steps(Cs) :- unfolding_steps_stratum(Cs,_).
fold_steps(Cs) :- folding_steps_stratum(Cs,_,_).
%simpl_steps(Cs) :- simplification_steps_stratum(Cs).
simpl_steps(Cs) :- remove_clauses_stratum(Cs,_).
