%%% WIDENING

clause_widening(D1,D2,D3) :-
	copy_term(D1,(H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(H1,V1),
	term_variables(H2,V2),
	V1=V2,
	B1=B2,
	primitives(C1,I1),
	primitives(C2,I2),
	cns_widening(I1,I2,C3),
	D3=(H2,bd(C3,B2)),
	nl, write(w), nl,
	show_cls([D1,D2,D3]).

cns_widening(C1,C2,C3) :-
	keep_entailed(C1,C2,E1),
	keep_leq_cns(C2,C1,E2),
	append(E1,E2,E3),
	term_variables(E3,V3),
	solve(E3,V3,C3).

%%% keep_entailed(C1,C2,C3)
%%% C3 is the subset of the constraints in C1 entailed by C2

keep_entailed([],_C2,[]).
keep_entailed(C1,C2,E) :-
	C1=[PC1|PC1s],
	(entails(C2,[PC1]) ->
	    E=[PC1|E1],
	    keep_entailed(PC1s,C2,E1)
	;
	    keep_entailed(PC1s,C2,E)
	).

%%% keep_leq_cns(C1,C2,C3)
%%% C3 is the subset of constraints in C1 which are leq than some constraint in C2

keep_leq_cns([],_C2,[]).
keep_leq_cns([PC1|PC1s],C2,E) :-
	member(PC2,C2), 
	leq_cns(PC1,PC2), !,
	E=[PC1|E1],
	keep_leq_cns(PC1s,C2,E1).
%keep_leq_cns([PC1|PC1s],C2,E) :-
%	\+ has_leq_vars(PC1,C2), !,
%	E=[PC1|E1],
%	keep_leq_cns(PC1s,C2,E1	).
keep_leq_cns([_|PC1s],C2,E) :-
	keep_leq_cns(PC1s,C2,E).

has_leq_vars(PC,C) :-	
	term_variables(PC,V1),
	member(C2,C),
	term_variables(C2,V2),
	length(V1,L1),
	length(V2,L2),
	L1=<L2.

leq_cns(C1,C2) :-	
	max_coeff(C1,M1),
	max_coeff(C2,M2),
	init_prog_max_coeff(Maxcoeff),
	{M1=<max(M2,Maxcoeff)}.


:- use_module(hull_clpq).

convex_hull_cl(D1,D2,D3) :-
	copy_term(D1,(_H1,bd(C1,B1))),
	copy_term(D2,(H2,bd(C2,B2))),
	term_variables(B1,V1),
	term_variables(B2,V2),
	convex_hull(V1,C1,V2,C2,V3,C3),
	int_coeff_transf(C3,C4),
%	append(C2,C4,C5),
	V2=V3,
	D3=(H2,bd(C4,B2)),
	nl, write(ch), nl,
	show_cls([D1,D2,D3]).

int_coeff_transf([],[]).
int_coeff_transf([C|Cs],[D|Ds]) :-
	transf_constraint(C,D),
	int_coeff_transf(Cs,Ds).

transf_constraint(C,D) :-
	den_list(C,Dens),
	lcm(Dens,M),
	multiply(M,C,D).
%	term_variables(C1,Vs),
%	solve([C1],Vs,[D]).

den_list(C,Dens) :- 
	C=..[_,Lhs,Rhs],
	d_list(Lhs,LDens),
	d_list(Rhs,RDens),
	append(LDens,RDens,Dens).

d_list(V,[D]) :- 
	var(V), !, 
	D=1.
d_list(rat(_,Y),[D]) :-
	!,
	D=Y.
d_list(rat(_,Y)*V,[D]) :-
	var(V), !, 
	D=Y.
d_list(P1+P2,Dens) :- !,
	d_list(P1,Dens1),
	d_list(P2,Dens2),
	append(Dens1,Dens2,Dens).
d_list(-V,[D]) :- 
	var(V), !, 
	D=1.
d_list(-rat(_,Y),[D]) :-
	!,
	D=Y.
d_list(-rat(_,Y)*V,[D]) :-
	var(V), !, 
	D=Y.
d_list(P1-P2,Dens) :-
	d_list(P1,Dens1),
	d_list(P2,Dens2),
	append(Dens1,Dens2,Dens).

lcm([N],L) :- L=N.
lcm([N1,N2|Ns],L) :-
	N is ((N1*N2)/gcd(N1,N2)),
	lcm([N|Ns],L).

multiply(M,C1,C2) :- 
	C1=..[R,Lhs1,Rhs1],
	multiply_poly(M,Lhs1,Lhs2),
	multiply_poly(M,Rhs1,Rhs2),
	C2=..[R,Lhs2,Rhs2].

multiply_poly(M,V,C) :- 
	var(V), !, 
	{Coeff=M*1},
	({Coeff=1} -> C=V ; C=Coeff*V).
multiply_poly(M,rat(X,Y),C) :-
	!,
	{N=:=M*rat(X,Y)},
	C=N.
multiply_poly(M,rat(X,Y)*V,C) :-
	var(V), !, 
	{Coeff=:=M*rat(X,Y)},
	({Coeff=1} -> C=V ; C=Coeff*V).
multiply_poly(M,P1+P2,Q1+Q2) :-
	multiply_poly(M,P1,Q1),
	multiply_poly(M,P2,Q2).
multiply_poly(M,-V,C) :- 
	var(V), !, 
	{Coeff=M*1},
	({Coeff=1} -> C=(-V) ; C=(-Coeff)*V).
multiply_poly(M,-rat(X,Y),C) :-
	!,
	{N=:=M*rat(X,Y)},
	C=(-N).
multiply_poly(M,-rat(X,Y)*V,C) :-
	var(V), !, 
	{Coeff=:=M*rat(X,Y)},
	({Coeff=1} -> C=(-V) ; C=(-Coeff)*V).
multiply_poly(M,P1-P2,Q1-Q2) :-
	multiply_poly(M,P1,Q1),
	multiply_poly(M,P2,Q2).
